/*	FAST GENERATOR OF PSEUDO-RANDOM UNIT NORMAL VARIATES

		C.S.Wallace, Monash University, 1994

To use this code, files needing to call the generator should #include the
file "FastNorm.h" and be linked with the maths library (-lm)
	FastNorm.h contains declaration of the initialization routine
'initnorm()', definition of a macro 'FastGauss' used to generate variates,
and three variables used in the macro.
	Read below for calling conventions.

THIS CODE ASSUMES TWO'S-COMPLEMENT 32-BIT INTEGER ARITHMATIC.  IT ALSO
ASSUMES THE 'C' COMPILER COMPILES THE LEFT-SHIFT OPERATOR "<<" AS A LOGICAL
SHIFT, DISCARDING THE SIGN DIGIT AND SHIFTING IN ZEROS ON THE RIGHT, SO
" X << 1" IS EQUIVALENT TO " X+X ".   IT ALSO ASSUMES THE RIGHT-SHIFT
OPERATOR ">>" IS SIGN-PRESERVING, SO ( -2 >> 1) = -1,  ( -1>>1) = -1.
*/


/*	Shuffle version	*/
#include <math.h>
/*
	A fast generator of pseudo-random variates from the unit Normal
distribution. It keeps a pool of about 1000 variates, and generates new
ones by picking 4 from the pool, rotating the 4-vector with these as its
components, and replacing the old variates with the components of the
rotated vector.


	The program should initialize the generator by calling initnorm(seed)
with seed a long integer seed value. Different seed values will give
different sequences of Normals.
	Then, wherever the program needs a new Normal variate, it should
use the macro FastGauss, e.g. in statements like:
	x = FastGauss;  (Sets x to a random Normal value)
*/

#define Scale ((double) 20000000.0)
#define Rscale ((double) (1.0 / Scale))
#define Rcons ((double) (1.0 / (2.0 * 1024.0 * 1024.0 * 1024.0)))
#define LEN  281   /* This is a quarter of the pool size. Must be odd */
#define TLEN  (4*LEN)
long gaussfaze, *gausssave;
double GScale;
/*	GScale, gaussfaze and gausssave must be visible to callers   */
static long vec1 [TLEN], vec2 [TLEN];
static long *u, *v;
static long nslew;
static long irs, lseed;


/*	--------------------------------------------------------     */

/*	Initinorm is called with an integer seed to initialize the Normal
	generator	*/

void initnorm (long seedvalue)
{
	lseed = seedvalue;
	irs = seedvalue;
	gaussfaze = 1;
	nslew = 0;
	GScale = Rscale;
	u = vec1;
	return;
}

/*	-----------------------------------------------------   */


double fastnorm ()
{
	long i;
	long p, q, r, s, c;
	long *pa, *pb, *pc, *pd;
	double ts, tr, tx, ty, tz;

/*	See if time to make a new set of 'original' deviates  */
	if (! nslew) goto renormalize;

startpass:
/*	Count passes	*/
	nslew ++;
/*	Reset index into Saved values	*/
	gaussfaze = TLEN - 1;	/* We will steal the last one	*/
/*	Set counter  */
	i = LEN;
/*	Have just finished the set in vector 'u', so mix from u to other */
/*	Also point u at other	*/
	v = u;
	if (u == vec1)  u = vec2;  else  u = vec1;
	gausssave = u;
/*	So will move from v to u	*/
	pa = u;  pb = u + LEN;  pc = pb + LEN;  pd = pc + LEN;
/*	Update pseudo-random irs and use to choose type of rotation  */
	irs = (irs <= 0) ? ((irs << 1) ^ 333556017) : (irs << 1);

	if (irs < 0) goto matrix2;

matrix1:
	p = *v++;  q = *v++;  r = *v++;  s = *v++;
	c = (p + q + r + s) >> 1;
	*pa++ = c-r;
	*pb++ = c-q;
	*pc++ = p-c;
	*pd++ = s-c;
	if (--i) goto matrix1;   goto endpass;

matrix2:
	p = *v++;  q = *v++;  r = *v++;  s = *v++;
	c = (p + q + r + s) >> 1;
	*pa++ = q-c;
	*pb++ = r-c;
	*pc++ = c-p;
	*pd++ = c-s;
	if (--i) goto matrix2;   goto endpass;

endpass:
/*	Choose a value for GScale which will make the sum-of-squares have
	the variance of Chi-Sq (TLEN), i.e., 2*TLEN.  Choose a value from
	Chi-Sq (TLEN) by choosing from Normal (TLEN, 2*TLEN), using the unit
	Normal in gausssave[TLEN-1], which will not be used by the caller.
	*/
	ts = GScale * u [TLEN-1] * sqrt (2.0 / TLEN);
	GScale = Rscale * sqrt (1.0 + ts);
	return (GScale * gausssave[0]);

renormalize:
/*	Here, replace the whole pool with conventional Normal variates  */
	nslew = -2000;
	ts = 0.0;
	p = 0;
nextpair:
	lseed = 69069 * lseed + 33331;
	irs = (irs <= 0) ? ((irs << 1) ^ 333556017):(irs << 1);
	r = irs + lseed;
	tx = Rcons * r;
	lseed = 69069 * lseed + 33331;
	irs = (irs <= 0) ? ((irs << 1) ^ 333556017):(irs << 1);
	r = irs + lseed;
	ty = Rcons * r;
	tr = tx * tx + ty * ty;
	if ((tr > 1.0) || (tr < 0.1)) goto nextpair;
	lseed = 69069 * lseed + 33331;
	irs = (irs <= 0) ? ((irs << 1) ^ 333556017):(irs << 1);
	r = irs + lseed;
	if (r < 0) r = ~r;
	tz = -2.0 * log ((r + 0.5) * Rcons);   /* Sum of squares */
	ts += tz;
	tz = sqrt ( tz / tr );
	u [p++] = (long) (Scale *  tx * tz);   
        u [p++] = (long) (Scale *  ty * tz);
	if (p < TLEN) goto nextpair;
/*	Horrid, but good enough	*/
/*	Calc correction factor to make sum of squares = TLEN	*/
	ts = TLEN / ts;  /* Should be close to 1.0  */
	tr = sqrt (ts);
	for (p = 0; p < TLEN; p++)	{
		ts = u [p] * tr;
		u [p] = (long) ((ts < 0.0) ? (ts - 0.5) : (ts + 0.5));
		}
	goto startpass;

}

