#include "dataclasses.h"
#include "tubparser.h"
#include "mathftn.h"

int ModelData::NoModels = 0;

void StateData::SetPropertiesTheSameAs(const StateData &S)
{
    int i;
    for (i=0; i<MAXLINKS; i++)   SiteIndex[i]=S.SiteIndex[i]; 
    for (i=0; i<MAXSPHERES; i++) Colours[i]=S.Colours[i]; 
}

void StateData::print(int N) 
{
    cout << "  State: " << Name << " ("<<N<<") "; 
    int i=0;
    cout << "\n    Links: ";
    while (i<MAXLINKS)
    {
        if (SiteIndex[i]!=-1) cout<<i<<": "<<SiteIndex[i]<<" "; 
        i++;
    }
    i=0;
    cout << "\n    Sphere Colours: "; 
    while (i<MAXSPHERES) 
    {
        if (Colours[i]!=-1) cout<<i<<": "<<Colours[i]<<" "; 
        i++;
    }
    cout << "\n\n";
}


void LinkData::print(int N) 
{
    cout << "  Link ("<<N<<") from site("<<SiteNo<<") and state(" <<StateNo <<")";
    cout << " at <"<<pos.x<<" "<<pos.y;
    cout <<" "<<pos.z<<"> dirtn: <"<<dirtn.x<<" "<<dirtn.y<<" "<<dirtn.z<<">";
    if (twist!=0) cout << " (twist "<<twist<<") ";
    if (twistVect!=0) cout << " (twist Vector " << twistVect << ") ";
    /*cout << "\n    break/bind probs: \n";
    for (int i=0; i<MAXLINKS; i++) 
    {
        if (BreakProb[i]>=0)
            cout <<i<<": [ " << BreakProb[i] << "/" << BindProb[i] << " ] ";
    }
    */
    cout << " ID 0x"<< hex << TypeId << " binds to: 0x" << hex << BindsTo << dec;
    cout << "\n";
}
            
ModelData::ModelData(TubParser* P, char *S) : BaseData(NULL, S) 
{
    Parser = P;
    int i;
    for (i=0; i<MAXSPHERES;   SphereNo[i++] = NULL);
    for (i=0; i<MAXSITES;     SiteNo[i++]   = NULL);
    for (i=0; i<MAXSTATES;    StateNo[i++]  = NULL);
    for (i=0; i<MAXEVENTS;    EventNo[i++]  = NULL);
    for (i=0; i<MAXLINKS;     LinkNo[i++]   = NULL);

    NoSpheres=0; NoSites=0; NoEvents=0; NoStates=0; NoLinks=0; 
    BoundRad = 0;
    Diffusion = -1;
    ModelNo = NoModels++;
}


int ModelData::GetObjectNo(ShortString* ObName, BaseData** ObArray, int NoObjects)
{
    if (ObName == NULL) {cerr<<"Object Name Null in GetObjectNo call\n"; return -1;}
    if (ObArray==NULL)  {cerr<<"ObArray Null: GetObjectNo-"<<ObName<<"\n";return -1;} 
    for (int i=0; i<NoObjects; i++) 
    {
        if (ObArray[i]==NULL) 
        {
            cerr<<"ObArray["<<i<<"] null in GetObjectNo: "<<ObName<<"\n";
            return -1;
        }
        if (ObArray[i]->Name==ObName) 
            return i;
    }
    return -1;
}
/*
 *  Try to calculate MOI, centroids etc. for objects.  A lot of this
 *  is inaccurate if the objects aren't defined with their co-ordinates
 *  the same as the MOI axes, since it doesn't handle the calculations
 *  properly - but heck, it's an approximation anyway...
 *
 *  It also pretty much expects the thing to be centered on zero, although
 *  it doesn't *have* to be...
 */

void ModelData::CalculateDetails(float Viscosity, float Temperature, float TimeStep)
{
    // *** Calculate Bounding Radius of all spheres ***

    float temp;
    for (int i=0; i<NoSpheres; i++)
    {
        temp = pt::abs(SphereNo[i]->center) + SphereNo[i]->radius;
        if (temp>BoundRad) BoundRad = temp;
    }
    BoundRad *= 1.1;  // add 10% for good luck (well, to avoid comparison errors)

    // *** Calculate Centroid & Mass ***
   
    Mass     = SphereNo[0]->mass;
    Centroid = SphereNo[0]->center * Mass;

    for (int i=1; i<NoSpheres; i++)
    {
        Centroid += SphereNo[i]->center*SphereNo[i]->mass;
        Mass     += SphereNo[i]->mass;
    }
    Centroid /= Mass;

cout << "centroid = " << Centroid << " mass = " << Mass << "\n";

    // *** Calculate Moment Of Inertia ***
       
    MOI = pt(0); 

    // find total M.O.I. by recursively adding spheres using parallel axis theorem
    for (int i=0; i<NoSpheres; i++)
    {
        pt dist = SphereNo[i]->center - Centroid;
        float dist_sq = pt::sqr(dist);
        pt dist_coord_sq = pt(dist_sq,dist_sq,dist_sq) - dist.element_mult(dist);
 
        MOI = MOI + pt(1,1,1)*SphereNo[i]->MOI;

        //pt Components_squared = SphereNo[i]->center.element_mult(SphereNo[i]->center);
        //MOI = MOI + (pt(dist_sq,dist_sq,dist_sq)-Components_squared)*SphereNo[i]->mass; 

        MOI = MOI + dist_coord_sq * SphereNo[i]->mass;
    }

cout << "moi = " << MOI << "\n";
   
    if (NoSpheres == 1)
        Diffusion = MathFtn::CalculateDiffusion(SphereNo[0]->radius, Viscosity, Temperature);
    else 
        Diffusion = MathFtn::CalculateDiffusion(MOI, Viscosity, Mass, Temperature); 

    rms_delta_pos = sqrt(2*Diffusion*TimeStep);

cout << " RMS MOVEMENT:\n\n *** " << rms_delta_pos << " \n\n";
}

void ModelData::CalculateProbMatrix()
{
    // We use the Bounding Radius, on the basis that this is the
    // effective collision radius of the molecule, due to rapid
    // rotation.  It also makes things much simpler. 

    /*
     *  This function calculates probabilities by reducing the problem
     *  of two spheres with different radii and different r.m.s. movement
     *  values to the problem of one sphere colliding with a stationary
     *  point.  It then reads the appropriate values from a general
     *  table read in by the parser.
     */

//TODO: Make sure distance measured from centers
//TODO: Possibly this should be moved to the TubParser class.

    float t = Parser->TimeScale;

    for (int i=0; i<Parser->NoModels; i++)
    {
        // The effective radius is the sum of the colliding object radii. 
        float RadiusSum = (BoundRad + Parser->ModelNo[i]->BoundRad) * 1e-9;  

        // could have used already calculated rms values, but this is just the same.
        float EffSigma = sqrt( 2*t*(Diffusion + Parser->ModelNo[i]->Diffusion) );

        float EffRadius = RadiusSum / EffSigma;

        for (int r=0; r<MAXINTERACTIONRANGE; r++)
        {
            if (EffSigma==0) // stationary objects
            {
cout << "Warning Stationary Objects\n";
                CollisionProbMatrix[i][r] = 0;  
            }
            else
            {
                float EffDistance = ((((float)r)*1e-9)/EffSigma ) - EffRadius;

                if (EffDistance<0) 
                    CollisionProbMatrix[i][r] = 1;  // spheres are overlapping.
                else
{
                    CollisionProbMatrix[i][r] = Parser->CollProb(EffDistance, EffRadius);
}
            }
        }        
    }
}

void ModelData::print(int N) 
{
    cout << "**MODEL**\n{\n  Name = "<<Name << " ("<<N<<")\n\n";
    cout << "  Frequency : " << Freq << "\n";
    cout << "  Mass : "<<Mass<<" Centroid : "<<Centroid<<" M.O.I. : "<<MOI<<"\n";
    cout << "  Diffusion : "<<Diffusion<<" [M*M/s]\n";
    cout << "  Bounding Sphere Radius " << BoundRad << "\n";
    cout << "  Spheres: "<<NoSpheres<<" States: ";
    cout <<NoStates<<" Events: "<<NoEvents<<" Sites: "<<NoSites<<" \n\n";

    int i;
    cout << "Spheres (" << NoSpheres << ")\n";
    i=0; while (i<NoSpheres && SphereNo[i]!=NULL) SphereNo[i]->print(i++); 
    cout << "Sites   (" << NoSites <<")\n";
    i=0; while (i<NoSites && SiteNo[i]!=NULL) SiteNo[i]->print(i++);
    cout << "States  (" << NoStates <<")\n";
    i=0; while (i<NoStates && StateNo[i]!=NULL) StateNo[i]->print(i++);
    cout << "Events  (" << NoEvents <<")\n";
    i=0; while (i<NoEvents && EventNo[i]!=NULL) EventNo[i]->print(i++);

    cout << "Links   \n";
    i=0; while (i<MAXLINKS && LinkNo[i]!=NULL) LinkNo[i]->print(i++);

    cout << "\n}\n";
}
			
