/*
 *   A parser class - reads a file in, and extracts yummy useful info...
 *
 */

#ifndef DATACLASSES
#define DATACLASSES

#include "parser.h"
#include "hashdefines.h"
#include "assert.h"

#define DEGTORAD (3.141593/180.0)

/*
 *   A whole buncha little classes to hold data from the parser in. 
 */

class TubParser;
class ParseWord;
class SiteData;
class SphereData;
class LinkData;
class StateData;
class EventData;
class ModelData;

/*
 *   A little base class.  Might come in useful, and cuts down on typing.
 */

class BaseData
{
    public:
 
    ShortString Name;
    ModelData*  Owner;

    BaseData(ModelData* O, ShortString S) {Owner = O; Name = S;}
    BaseData(ModelData* O) {Owner = O; Name = "";}

    virtual void print(int N) {cout << N;};
};


/*
 *   Our "Virtual Proteins" (or whatever) are modelled as multiple little
 *   spheres.  A sphere is little more than a position and a radius -
 *   but a colour and a name get spliced on as well.
 */

class SphereData : public BaseData
{
    public:
    
    pt    center;
    float radius;  // radius in nanometers...
    int   colour;
    float mass;
    float MOI;    // moment of inertia (in any axis - hence a float, not a vector)
 
    SphereData(ModelData* O,ShortString N,pt c,float r,int Col,float M=0) : BaseData(O,N) 
        { center = c; radius = r; colour=Col; 
          mass=(M!=0)?M:4.18879*r*r*r*PROTEIN_DENSITY;   // sphere assumed constant density 
          MOI = 0.4*mass*r*r;}                           

    void print(int N) 
    {
         cout <<"  Sphere "<<Name<<" ("<<N<<") <"<<center.x << " "<<center.y;
         cout <<" "<<center.z<<"> r: "<<radius<<" col: "<<colour<<" \n";
         cout <<"  Mass: " << mass << "   M.O.I. " << MOI << "\n\n";}
};


/*
 *  An event is something which changes a virtual proteins state from one
 *  state to another.
 */

class EventData : public BaseData      // Data for a single Event Type
{
    public:


    int   Site;
    int   LinkedSite;
    int   Type;
    float RandomVal;
    int   OldState;
    int   NewState;

    EventData(ModelData* M, int T, int O, int N, int S1, int S2 ) : BaseData(M)
        {Type=T; OldState=O; NewState=N; Site=S1; LinkedSite=S2; RandomVal=0;}
    EventData(ModelData* M, int T, int O, int N, float R) : BaseData(M)
        {Type=T; OldState=O; NewState=N; RandomVal=R; Site=NONE; LinkedSite=NONE;}

    void printsite(int site) 
    {
        if (site==NONE) cout << " none ";
        else if (site==ALL) cout << " all ";
        else if (site==UNKNOWN) cout << " unknown ";
        else cout << " "<<site<<" ";
    }

    void print(int N) 
    {
        cout << "  Event: ("<<N<<") ";
        if (Type==RANDOMEVENT)
            cout<<"  Random: "<<RandomVal;
        else if (Type==BINDEVENT)
        {
            cout << "  Bind "; printsite(Site); printsite(LinkedSite);
        }
        else if (Type==BREAKEVENT)
        {
            cout << "  Break "; printsite(Site); printsite(LinkedSite);
        }
        cout << " state " << OldState << " -> " << NewState << "\n\n";
    }
};


/*
 *  The state a V.P. is in: may change the colours of things, and
 *  which links are active. 
 */

class StateData : public BaseData
{
    public:

    int SiteIndex[MAXLINKS];       // index = site no. val = link no.
    int Colours[MAXSPHERES];       // index = sphere no. val = colour val.

    int FirstState;

    StateData(ModelData* O, char *N, int First=FALSE) : BaseData(O,N) 
    { 
        FirstState = First;
        for (int i=0; i<MAXLINKS;   SiteIndex[i++]=-1);
        for (int i=0; i<MAXSPHERES; Colours[i++]=-1);
    }

    int SetLink(int No, int Val)
        {if (No>=MAXLINKS) return -1;   SiteIndex[No]=Val; return 0;}
    int SetColour(int No, int Val) 
        {if (No>=MAXSPHERES) return -1; Colours[No]=Val; return 0;}

    void SetPropertiesTheSameAs(const StateData &S);

    void print(int N); 
};


/*
 *  A name to attach links to.
 */

class SiteData : public BaseData
{
    public:

    SiteData(ModelData* O, char *S) : BaseData (O,S) {;} // No = NoSites++; }

    void print(int N) {cout << "  Site: " << Name << " ("<<N<<")\n\n";}
};


/*
 *  A link lives at a site, and knows how attractive other links are.
 *  It also has a position, a direction, and possibly a twist.
 */

class LinkData : public BaseData
{
    public: 

    int   LinkNo;    // The global link no - used in BindProb[] and BreakProb[]

    int   SiteNo;    // the (Model) Index of the current site
    int   StateNo;   // the (Model) Index of the first state this link appears in
    pt    pos;       // position of the link in local co-ords
    pt    dirtn;     // the unit vect direction of the point in local co-ords
    float twist;     // angle of twist; used instead of...
    pt    twistVect; // the twist vector - a vector at right angles to the dirtn used to
                     // align links with a difficult geometry.
    float BindProb[MAXLINKS];
    float BreakProb[MAXLINKS];


// could use a "long int" here, but doing it this way may allow easier porting to
// PCs...
    long TypeId;         // boolean Id as power of two (low 32 bits)
    long TypeIdHigh;    // boolean Id as power of two (high 32 bits)
    long BindsTo;       // boolean template giving possible binding combinations
    long BindsToHigh;   // same, high 32 bits...

    LinkData(ModelData* O,ShortString N,int S,int St,pt P,pt D,float T) : BaseData(O,N)
    {
        TypeId = 0; TypeIdHigh = 0;
        SiteNo = S; StateNo = St; pos = P; dirtn = ~D; twist = T*DEGTORAD;
        for (int i=0;i<MAXLINKS;i++) {BindProb[i]=-1.1; BreakProb[i]=-1.1; twistVect = pt(0,0,0);}
    }

    LinkData(ModelData* O,ShortString N,int S,int St,pt P,pt D,pt T) : BaseData(O,N)
    {
        TypeId = 0; TypeIdHigh = 0;
        SiteNo = S; StateNo = St; pos = P; dirtn = ~D; twist = 0;
        for (int i=0;i<MAXLINKS;i++) {BindProb[i]=-1.1; BreakProb[i]=-1.1; twistVect = ~T; } 
    }

    LinkData(LinkData* Old, int NewState) : BaseData(Old->Owner,Old->Name)
    {
        TypeId = 0; TypeIdHigh = 0;
        SiteNo = Old->SiteNo; pos = Old->pos; dirtn = ~Old->dirtn; 
        twist = Old->twist; twistVect = Old->twistVect;
        for (int i=0;i<MAXLINKS;i++) 
        {
             BindProb[i]=Old->BindProb[i]; 
             BreakProb[i]=Old->BreakProb[i];
        }

        StateNo = NewState; 
    }

    void SetBinding(LinkData* Bindee)
    {
        BindsTo += Bindee->TypeId;
        BindsToHigh += Bindee->TypeIdHigh; 
    }

    int CanBind(LinkData* Bindee)
    {
        return ((BindsTo&Bindee->TypeId)||(BindsToHigh&Bindee->TypeIdHigh));
    }

    int SetBindProb(int No, float P)
        {if (No>=MAXLINKS) return -1; BindProb[No]=P; return 0;}
    int SetBreakProb(int No, float P)
        {if (No>=MAXLINKS) return -1; BreakProb[No]=P; return 0;}
    void SetNo(int N) {LinkNo = N;}

    float GetBindProb(int linkNo)
    {
       if (linkNo<MAXLINKS) return BindProb[linkNo];
       cout << "bad link no "<<linkNo<<" passed to GetBindProb\n";
       return -1; 
    }

    float GetBreakProb(int linkNo)
    {
       if (linkNo<MAXLINKS) return BreakProb[linkNo];
       cout << "bad link no "<<linkNo<<" passed to GetBreakProb\n";
       return -1; 
    }


    void print(int N); 
            
};


/*
 *   A model describes an entire V.P. - (except for the links, which are kept
 *   separately.)
 */

class ModelData : public BaseData
{
    public:

    SphereData *SphereNo[MAXSPHERES];
    SiteData   *SiteNo[MAXLINKS];
    StateData  *StateNo[MAXSTATES];
    EventData  *EventNo[MAXEVENTS]; 
    LinkData   *LinkNo[MAXLINKS];

    int NoSpheres,NoSites,NoStates,NoEvents,NoLinks;

    float  MaxDist[MAXMODELS];
    float  CollisionProbMatrix[MAXMODELS][MAXINTERACTIONRANGE];

    float  Diffusion;              // Diffusion Constant
    float  Freq;                   // Amount of this model in total sim.
    float  BoundRad;               // Smallest distance enclosing component spheres
    pt     Centroid;               // Centorid (Center of mass) of object
    pt     MOI;                    // Moment of Inertia of object.
    float  Mass;                   // Total Mass of object.
    float  FrictionCF;             // Coefficient of Friction (Theoretical)
    float  rms_delta_pos;          // standard deviation of motion / time step

    TubParser* Parser;             // The parser that produced this model

    int ModelNo;
    static int NoModels;

public:

    ModelData(TubParser* P, char *S); 

    int  GetColour(int Sphere_no, int State_no)
    {
        return StateNo[State_no]->Colours[Sphere_no];
    }

    void SetDiffusionConst(float D) {Diffusion = D;}

    LinkData* GetLink(int stateno, int siteno)
    { 
        LinkData* ret = LinkNo[ StateNo[stateno]->SiteIndex[siteno] ]; 
        assert (ret != NULL);
        assert (ret->SiteNo  == siteno);
        return ret;
    }


    SphereData* GetSphereByNo(int N) {return SphereNo[N];}
    SiteData*   GetSiteByNo(int N)   {return SiteNo[N];}
    StateData*  GetStateByNo(int N)  {return StateNo[N];}
    EventData*  GetEventByNo(int N)  {return EventNo[N];}
    LinkData*   GetLinkByNo(int N)   {return LinkNo[N];}

    int AddSphere(SphereData *S)
       {if (NoSpheres>=MAXSPHERES) return -1; SphereNo[NoSpheres++]=S; return 0;}
    int AddSite(SiteData *S)
       {if (NoSites>=MAXSITES) return -1; SiteNo[NoSites++]=S; return 0;}
    int AddEvent(EventData *E)
       {if (NoEvents>=MAXEVENTS) return -1; EventNo[NoEvents++]=E; return 0;}
    int AddState(StateData *S)
       {if (NoStates>=MAXSTATES) return -1; StateNo[NoStates++]=S; return 0;}

    int AddLink(LinkData *S, int LinkPos)
    { 
        if (NoLinks>=MAXLINKS) return -1; LinkNo[LinkPos]=S; 
        return 0;
    }


    void SetFrequency(float F) { Freq = F; }

    float GetFrequency() { return Freq; }

    int GetSiteNo(ShortString* Name) 
    {
        return GetObjectNo(Name, (BaseData **) SiteNo, NoSites); 
    }

    ShortString GetSiteName(int No) 
    {
        if (No<0)
        {
            switch (No)
            {
                case UNKNOWN: return ShortString(" unknown ");
                case NONE: return ShortString(" none ");
                case ALL: return ShortString(" all ");
                default: return ShortString(" error:bad site const ");
            }
        }
        else 
            return (No<NoSites)?SiteNo[No]->Name:ShortString(" error ");
    }
    ShortString GetSphereName(int No) 
    {
        return (No<NoSpheres)?SphereNo[No]->Name:ShortString(" error ");
    }
    ShortString GetStateName(int No) 
    {
        return (No<NoStates)?StateNo[No]->Name:ShortString(" error ");
    }

    SiteData* GetSiteByName(ShortString* Name)
    {
        int No = GetObjectNo(Name, (BaseData**) SiteNo, NoSites);
        SiteData* Bloop = NULL;
        return (No>=0)?SiteNo[No]:Bloop;
    }

    StateData* GetStateByName(ShortString* Name)
    {
        int No = GetObjectNo(Name, (BaseData**) StateNo, NoStates);
        StateData* Bloop = NULL;
        return (No>=0)?StateNo[No]:Bloop;
    }

    int GetStateNo(ShortString* Name)
    {
        return GetObjectNo(Name, (BaseData**) StateNo, NoStates);
    }

    int GetSphereNo(ShortString* Name)
    {
        return GetObjectNo(Name, (BaseData**) SphereNo, NoSpheres);
    }

    int GetObjectNo(ShortString* ObName, BaseData** ObArray, int NoObjects);

    void CalculateDetails(float Viscosity, float Temperature, float TimeStep);


    float CollisionProb(ModelData *B, float Dist)
    {
        if (Dist > MAXINTERACTIONRANGE)
            return 0;
        else 
            return CollisionProbMatrix[B->ModelNo][ (int) Dist];
    }

    void CalculateProbMatrix();
 
    void print(int N); 

};
			
#endif
