/* The Dimer class
 *
 * The Dimer class handles the little dimer objects that
 * are doing the combining.  There are methods here to
 * jump the little fellas around, and generally 
 * describe their movement and interaction with other
 * dimers - such as sticking together, dephosphorelating
 * etc.
 *
 * SCALE: one `unit' in the simulator corresponds to one  nanometer.
 * 
 */

//#define VEM_HACK   // emergency hack for virtual microscopy stuff... :-)
//#define EVIL_VEM_HACK   // an especially evil emergency hack for virtual microscopy stuff... :-)

#define DEBUGLINKAGE FALSE
#include "assert.h"
#include "tubular.h"
//#include "link.h"
#include "dimer.h"
#include "microtubule.h"
#include "matrix.h"

/* dimer static data variables.  I want dynamic linking. */

Dimer *Dimer::Dimers;
int    Dimer::NoObjects;
int    Dimer::Field_Size;
int    Dimer::Half_Field_Size;
long   Dimer::NoCollisions = 0;

int    debugEvents = 0;

/*
 *   And now, some dimer functions...
 */  

ostream& operator << (ostream &o, Dimer *D) 
{
    o << "\nDimer   : " << setw(25) << setiosflags(ios::left) 
      << D->TagNo << " center:" << D->center 
      << "\n  state : " << setw(25) << setiosflags(ios::left)
      << "state:  " << D->Model->GetStateName(D->State); 

    if (D->Owner != NULL) o << " owner = " << D->Owner->TagNo;

    o << "\n"; 

    if (D->GetBound(true)) 
    {
        for (int i=0; i<D->NoSites; i++)
            if (D->LinkNo[i].to!=NULL) 
            {
                 o << "  link  : " << setw(25) << setiosflags(ios::left)
                   << D->Model->GetSiteName(i) << " to " 
                   << setw(6) << D->LinkNo[i].to->TagNo << "      pos:"
                   << "   " << D->LinkNo[i].true_pos << "\n";
            }
        o << "\n";
    }
    else
        o << " *not bound* \n";
    return o;
}

ostream& operator << (ostream &o, Dimer &D) 
{
    o << &D;
    return o;
}


/********************************************************************
 *
 *    ***   START OF DIMER CLASS   ***
 *
 *******************************************************************/

// Constructor.  Does stuff all.  Real work done by "SetData()" below
// (Reason: screwed initialisation of array objects in C++)
Dimer::Dimer()
{
    TagNo = -1;  // Flag for an uninitialised dimer.
}


/*
 *   Effectively, the constructor.  Gives dimers a random, unattached position,
 *   and sets up the data internals.
 *
 */

void Dimer::SetData(ModelData* M)
{
    assert ( M != NULL ); 
    assert ( M->Parser != NULL );

    static int counter=0;

    // Set up the local co-ordinate frame

    Basis = CBMatrix::Identity; 

    Model = M;

    NoSpheres = M->NoSpheres;

    NoSites = M->NoSites;

//    SetTwistMatrices();

    State = 0;                              // Objects start in base state 

    SetSpheres();                           // Set up spheres for this state

    SetLinks();                             // Set up links for this state

    Owner=NULL;                             // Dimer starts on its own.

    TagNo = counter++;                      // Give the dimer a name

    LocalMomentOfInertia = M->MOI; 

    center    = pt(RpmS, RpmS, RpmS);       // give dimer random position

    BoundRad = M->BoundRad;
 
    rms_dp   = M->rms_delta_pos * 1e9;      // rms movement vector / prog cycle
                                            // (scaled to 1unit = 1 nanometer)

    spin();                                 // give dimer random orientation
}

void Dimer::SetData(ModelData* M, PosData *P)
{
    assert (M != NULL);
    assert (P != NULL);


    if (M==NULL || P==NULL)
    {
        cout << " Error: Null ModelData/PosData pointers passed to Init\n";
        if (M!=NULL) cout << "   (Model was : " << M->Name << ")\n";
        if (P!=NULL) cout << "   (PosData o.k. though) \n";
        return;
    }
    SetData(M);
    center = P->center;

    RecalculateDetails();
}

void Dimer::SetSpheres()   // *ahem*.  Currently stateless...
{
    assert (Model != NULL);

    for (int i=0; i<NoSpheres; i++)
    {
        SpherePos[i]=Model->SphereNo[i]->center;  // init positions to local positions
        SphereRad[i]=Model->SphereNo[i]->radius;  // redundant copy of sphere radii
    }
}

void Dimer::SetLinks()
{
     
     for (int SiteNo=0; SiteNo<Model->NoSites; SiteNo++)
     {
         assert (Model != NULL);
         LinkNo[SiteNo].UpdateType(Model->GetLink(State, SiteNo));
         assert (LinkNo[SiteNo].Type != NULL);

         if (LinkNo[SiteNo].Bound()==TRUE)
         {
             Link* Other = LinkNo[SiteNo].to->FindLink(this);
             if (Other == NULL)
             {
                 cout << " Error twixt two links -> dimer " << TagNo;
                 cout << " not really linked to " << LinkNo[SiteNo].to->TagNo <<"\n";
                 if (Owner != NULL) 
                     cout << TagNo << " owned by " << Owner->TagNo <<"\n"; 
                 if (LinkNo[SiteNo].to->Owner != NULL) 
                     cout << LinkNo[SiteNo].to->TagNo << " owned by " << LinkNo[SiteNo].to->Owner->TagNo <<"\n"; 
cout <<  " Attempting unlink: " << TagNo << "," << LinkNo[SiteNo].to->TagNo <<"\n";
                 UnLink(this,LinkNo[SiteNo].to);
                 //clear();
                 return;
             }
             LinkNo[SiteNo].SetBreakProb(Other);
             Other->SetBreakProb(&LinkNo[SiteNo]);
         }
     }
}


/*
 *   Destructor - should never be called
 */

Dimer::~Dimer()
{
    cerr << "OOPS - BAD THINGS! BAD THINGS!  Evil Monster is eating Dimers!\n";
}


/*
 *   This initialises the binding "twist" matrices for the different Dimers.
 *
 *   CONSIDER -> All these should be read from a file
 */

void Dimer::SetTwistMatrices()
{
/*
    Link::matrix_no[TOP]    = CBMatrix::Identity;
    Link::matrix_no[BOTTOM] = CBMatrix::Identity;
    Link::matrix_no[LEFT]   = CBMatrix::Identity;
    Link::matrix_no[RIGHT]  = CBMatrix( 0.88545603, 0.46472317, 0, 
                                       -0.46472317, 0.88545603, 0, 
                                        0,          0,          1); 
*/
}


/*
 *  spins the dimer to a random orientation
 */

void Dimer::spin()
{
//TODO:  Speed Up.

    // This may prove prohibatively slow.  If so, maybe switch to random
    // access into big positional table??

    pt blah = RAND_U_VECT;
    Basis.SetZ(blah);
    pt temp = RAND_U_VECT;
    temp = ~(temp * blah);
    Basis.SetY(temp);
    temp = temp * blah;
    Basis.SetX(temp);
    Basis.is_identity = FALSE;


#ifdef SAFE
    float D = Basis.det();
    if (D>1.01&&D<.99)
    {
        cout << "Dimer " << TagNo << " Bad Basis post Spin: " << Basis;
        Basis.normalise();
    }
#endif

    RecalculateDetails();
}


/*
 *  Returns true if the dimer is linked to another dimer.
 */

int Dimer::GetBound(int ignore_errors) // ignore_errors used for debug printing
{
    for (int i=0; i<NoSites; i++) 
        if (LinkNo[i].Bound()) 
        {
            if (Owner!=NULL) 
                return TRUE;
            else
            {
#ifdef VEM_HACK
                if (rms_dp==0) return FALSE;
#endif
                cerr << " Bad Dimer Found. ("<<TagNo<<")Clearing.\n";

                if (ignore_errors == FALSE)  // during mt merges a dimer may be ownerless,
                {                            // but still have links.
                    assert (Owner != NULL); 
                    free();
                }
                return FALSE;
            }
        }

    return FALSE;
}


/*
 *  Returns the link from this dimer to the parameter Dimer
 */

Link* Dimer::FindLink(Dimer *D)
{
    int i;
    for (i=0; i<NoSites; i++)
    {
        if (LinkNo[i].to==D) 
             return (&LinkNo[i]);
    }
    return NULL;
}


/*
 *   Checks if the linked dimers are in the correct position...
 */

int  Dimer::CheckLinks()
{
    for (int i=0; i<NoSites; i++)
    {
        if (LinkNo[i].to!=NULL)
        {
            Link* BackLink = FindLink(LinkNo[i].to);

            if (BackLink == NULL)
            {
                cout << " Linkage Error: Dimer " << TagNo << " linked to ";
                cout << LinkNo[i].to->TagNo << " but is not bound back\n";
            }

            // Both links should be joined, i.e. at the same spatial position...

//            if (pt::sqr(BackLink->true_pos,LinkNo[i].true_pos)>.01)
            if (TheProgram.EqualWithWrap(BackLink->true_pos,LinkNo[i].true_pos)==false)
            {
                cout <<"Error: dimers "<<TagNo<<" & "<<LinkNo[i].to->TagNo;
                cout <<" badly linked\n" << "A: " << this << "\n link: ";
                cout << LinkNo[i] << "\nB: " << LinkNo[i].to << "\n link: " << BackLink;
                return ERROR;
            }
        }
        if (LinkNo[i].Type == NULL)
            cout << " Dimer " << TagNo << " has null link type for link : " << i << "\n";
     }
     return OK;
}


/*
 *  Sets the Owning Microtubule 
 */

void Dimer::SetOwner(Microtubule *M)
{
    Owner = M;
}



/*
 *  move a dimer linearly by the offset;
 */

void Dimer::add(pt offset)
{
#ifdef VEM_HACK
    if (rms_dp==0) return;
#endif

    assert(isnan(offset.x)==FALSE);

    center += offset;

    assert(isnan(center.x)==FALSE);

    for (int i=0; i<NoSites; i++) LinkNo[i] += offset; 

    for (int i=0; i<NoSpheres; i++) SpherePos[i] += offset;

/*    if (TheProgram.OutOfBounds(center))
    {
        cout<<"\nWierdness in add - Dimer "<<this<<"\n(offset = " << offset << ")\n\n";
    }
*/
}


/*
 *  Moves the dimer in a random direction, a random distance (up to Dist)
 *  - but *DOES NOT* rotate the dimer.
 */

void Dimer::jump()
{
   pt jmp = TheProgram.RandomNormalPt()*rms_dp; 
   assert(isnan(rms_dp)==FALSE);
   assert(isnan(jmp.x)==FALSE);

   pt offset = TheProgram.BoundsOffset(center+jmp);
   jmp += offset;

   add(jmp); 
}


/*
 *  Moves a dimer a random distance and direction (up to Dist), 
 *  and gives it a random orientation. 
 */


inline void Dimer::bump()
{
    if (rms_dp!=0) // if not stationary object (part of static structure)
    {
        jump();
        spin();
    }

#ifdef EVIL_VEM_HACK
    if (rms_dp !=0)
    {
        if (center.x < 0)
        {
            center.x = - center.x;  // le hack brutal...
            RecalculateDetails();
        }
   
        if (center.x > (Half_Field_Size - 10))
        {
            center.x = center.x - 10; 
            RecalculateDetails();
        }
    }
#endif
}


/*
 *   finds the square of the shortest distance between the 
 *   surfaces of the component spheres of two objects 
 *
 *   Nb: Returns -ve if spheres are overlapping.
 *   NbII: Adjusted for wrap-around field.
 */

float Dimer::min_contact_dist_sq(Dimer *a, Dimer *b)
{
    float distance_sq;
    float sum_radii, dist;
   
    float ReturnVal=9999;
 
    for (int i=0; i<a->NoSpheres; i++)
    {
        for (int j=0; j<b->NoSpheres; j++)
        {
            pt sphereA = TheProgram.MoveIntoBounds(a->SpherePos[i]); // adjust for wrap-around
            pt sphereB = TheProgram.MoveIntoBounds(b->SpherePos[j]); // ""
            distance_sq = pt::sqr(sphereA, sphereB);
            sum_radii = a->SphereRad[i] + b->SphereRad[j]; 
            dist = distance_sq - (sum_radii*sum_radii); 
            if (dist<ReturnVal) 
            {
                ReturnVal = dist;
            }
        }
    }

    return ReturnVal;
} 


/*
 *   finds the shortest distance between two dimers,
 *   measuring from the alpha and beta centers.
 */

inline float Dimer::min_contact_dist(Dimer *a, Dimer *b)//, pt *Diff, int GetDiff)
{
    float F = min_contact_dist_sq(a,b); //, Diff, GetDiff);
    if (F<0) F *=-1;   // F may be -ve if objects are overlapping...
    return sqrt(F);
} 



/* 
 *   Removes a dimer from any fixed relationship 
 *   with it's neighbouring dimers.
 *
 *   Some elaborate checking goes on to make sure that
 *   other, correctly linked dimers, aren't accidently
 *   affected when this dimer breaks its links.
 *
 *   BUG:  A rare (but possible) condition arises where
 *         a single dimer, forming a bridge between two
 *         otherwise disconnected sections of a microtubule
 *         is deleted.  The program *should* break the 
 *         microtubule in to two sections - it does not.
 */

void Dimer::unbind()
{
    if (TheProgram.verbose)
    {
        cout << TheProgram.NoCyclesCompleted << " unbinding dimer "<<TagNo;
       if (Owner!=NULL) cout <<" from " << Owner->TagNo ;
       cout << "\n";
    }

    for (int i=0; i<NoSites; i++)
    {
        if (LinkNo[i].Bound())
        {
            UnLink(this,LinkNo[i].to); 
        }
    }

    if (GetBound() == TRUE) 
    {
        cerr << "\nERROR in unbind " << TagNo<< "\n";
        for (int i=0; i<NoSites; i++)
        {
            if (LinkNo[i].Bound()) cout << "link " << i << " bound \n";
        }
        printlinks();
    }
}

/*  
 *   This frees a dimer from an MT without *any* consideration of other
 *   dimers.  Used mainly when an MT self-destructs.
 */

void Dimer::FreeBySelf()
{
    Owner = NULL;

    int BrokenLink[NoSites];

    for (int i=0; i<NoSites; i++)         // first, break *all* links...
    {
        BrokenLink[i] = FALSE; 
        if (LinkNo[i].Bound())
        {
            LinkNo[i].to=NULL;            // break the link.
            BrokenLink[i] = TRUE; 
        }
    }

    for (int i=0; i<NoSites; i++)         // .. then, check for any relevant
    {                                     // state changes.  *Don't* check for 
        if (BrokenLink[i]==TRUE)          // state changes while dimer is half-unlinked.
            CheckStateOnBreak(&LinkNo[i]);
    }
}

/*
 *  Go through ALL dimers, checking distance between them.  This   
 *  is very comp. expensive, and is only used to double check
 *  while debugging (The error should never be triggered, if the
 *  rest of the program is working correctly.)
 */

void Dimer::Verify_spacing()  //Brute force search for debugging
{
    for (int i=0; i<NoObjects; i++)
    {
        for (int j=i+1; j<NoObjects; j++)
        {
            if (min_contact_dist_sq(&Dimers[i], &Dimers[j])<-0.01)
            {
               cerr << "\nSPACING ERROR Dimer["<<i<<"] & ["<<j<<"] ";
               cerr << " & " <<  Dimers[j] << " Dist: " << min_contact_dist_sq(&Dimers[i], &Dimers[j])<<")\n";
            }
        }
    }
}


/*
 * This links a dimer to another dimer, and adjusts the linkages
 * accordingly, as well as positioning the linked to dimer.
 *
 * D     - dimer being linked
 * L1    - "this" dimer's active link
 * LD    - the dimer "D" 's active link
 *
 *
 *  ALGORITHM:
 *
 *    a)  make co-ordinate basis of dimer to be linked (*D) the same as the 
 *        linking dimer (*this).  Make origins of co-ord systems identical.
 *    
 *    b)  rotate  the co-ord basis of the second dimer, depending
 *        on the directions and degree of twist specified by the links.
 *        HERE BE MAGIC!  
 *            IF there is a "twist" in radians: 
 *                The dimers are rotated so that the links 
 *        align (IN OPPOSITE DIRECTIONS), by taking the *shortest* 
 *        rotational arc that will 
 *        accomplish this.  If there are multiple shortest arcs (i.e.
 *        the link directions are identical) the arc is chosen which
 *        has the *Z-AXIS* as the arcs axis (i.e. the dimer is twisted
 *        around the z-axis until the linkage sites are aligned).   
 *            IF there is a "twist Vector" for a link:
 *                The dimers are rotated so that the directions are
 *        aligned (opposite directions) and the "twist Vectors" are
 *        co-incident.  I.e. the the second dimer is aligned, and 
 *        then rotated until it's "twist vector" is identical to the
 *        first dimer.
 *
 *    c)  Finally, move the linked dimer (*D) to a new position, calculated
 *        by co-locating the position of the two links. (i.e. moving the
 *        dimer (*D) so that the link (*LD) is at the same global position
 *        as the link (*L1)
 *
 */

int Dimer::Align(Dimer *A, Dimer *B, Link *LA, Link *LB)
{
    // Sanity Checks

    if ((LA->CanBind(LB)==false) || (LB->CanBind(LA)==false))
    {
        cout << "naughty dimers = " << A->Model->Name << " & " << LA->Type->Name;
        cout << " linking " << B->Model->Name << " & " << LB->Type->Name<<"\n";
        if (LA->CanBind(LB)==false) cout << " First can't link second ";
        if (LB->CanBind(LA)==false) cout << " Second can't link first";
    }

    assert (LA->CanBind(LB) == TRUE);             // ... dimers must be able to bind each other
    assert (LB->CanBind(LA) == TRUE);             // ... dimers must be able to bind each other



    A->SetToMyCoords(B);  // make both dimers basis co-ords identical

    /*
     *   The next step is to align the direction vectors of the two links
     *   we are joining together.
     */

    pt dirtnA = LA->Type->dirtn; 
    pt dirtnB = LB->Type->dirtn*-1;

    if (pt::aboutequal(dirtnA,dirtnB))
    {
          // cout << "dimers already aligned!!!\n";
          // don't bother recalculating if directions already matching:
          // (i.e. pointing in exactly opposite directions).
    }   
    else
    {
        pt axis;
        float angle;

        if (pt::aboutequal(dirtnA, LB->Type->dirtn)) // both pointing in the same dirtn. 
        {                                            // -> spin arbitrarily
            if (pt::aboutequal(dirtnA, pt(0,0,1)))
                axis = pt(0,1,0);
            else
                axis = ~(dirtnA*pt(0,0,1)); 
            angle = 3.14159; 
        } 
        else        
        {
            float DotProd = dirtnA^dirtnB;

            axis = ~(dirtnA*dirtnB);

//sanity checks
assert ((dirtnA^axis)<FPERROR);
assert ((dirtnB^axis)<FPERROR);
assert ((dirtnA^axis)>-FPERROR);
assert ((dirtnB^axis)>-FPERROR);

            angle = acos(DotProd);
        }
        // to align, rotate by -ve the angular difference...
        B->Basis = A->Basis * CBMatrix::RotationMatrix(axis, angle);
  
        B->RecalculateDetails();    // recalc global positions of links etc.

        // make sure that directions were properly calculated!
        pt new_dirtnB = CBMatrix::RotationMatrix(axis, angle) * dirtnB;

        // debug sanity check
        if (new_dirtnB != dirtnA)
        {
            cout << "link directions not co-linear: \ndirtnA: " << dirtnA << "\ndirtB: ";
            cout << dirtnB << "\nnew dirtn B = "<<new_dirtnB << "\n";
            cout << "axis = " << axis << "    - angle = " << (angle*180.0/PI) << "\n";
            CBMatrix temp = CBMatrix::RotationMatrix(axis,-angle);
            cout << "rot matrix = \n" << temp << "\n";
            pt trueA = A->Basis * LA->Type->dirtn;
            pt trueB = B->Basis * LB->Type->dirtn;
            cout << "\ntA: " << trueA << "\ntB: " << trueB << "\n " << (trueA^trueB); 
        }
        // sanity check
        assert (new_dirtnB == dirtnA);
    }


    float twistAngle = 0;

    if (LA->Type->twist!=0 || LB->Type->twist!=0)
    /*
     *  We have a "twist angle".  After aligning the dimers using the 
     *  a) align basis matrices
     *  b) make smallest rotation so that links are colinear
     *  - then do c) rotate by the combined twist angle, around the
     *    direction vector of dimer A (which is -ve dirtn of dimer B) 
     */
    {
        twistAngle = LA->Type->twist + LB->Type->twist; 
        if (twistAngle < 0.0001) twistAngle = 0;
    }

    if (LA->Type->twistVect != 0 || LB->Type->twistVect != 0)
    /*
     *  We have two "twist vectors".  After aligning the direction
     *  vectors as above, rotate dimer B so that it's twist vector
     *  is identical to dimer A.  (Think of the twist vectors as
     *  locking poles that are tied colinearly together when the dimers join.)
     */
    {
        if (LA->Type->twistVect == 0 || LB->Type->twistVect == 0)
        {
            cerr << " error -> if linking dimers with twist vectors, *both* ";
            cerr << " dimers must have a defined twist vector\n";
            return ERROR;
        }

        // Find angle between twist vectors...
        pt twistA = A->Basis * LA->Type->twistVect;
        pt twistB = B->Basis * LB->Type->twistVect;

        // (note that twistVectors are unit vectors)
        double dot_prod = twistA^twistB;
        if (dot_prod>1.0) dot_prod = 1.0;    // correct floating point error.
        if (dot_prod<-1.0) dot_prod = -1.0;  // correct f.p. error.

        twistAngle = acos(dot_prod);
        if (twistAngle < 0.0001)    // dimers already correctly aligned!
        {                           // (depending on geometry this can be common)
            twistAngle = 0;
        }
        else if (twistAngle > 3.1415) // margin to correct f.p. error
        {
            twistAngle = PI;
        }
        else  // work out sign of angle...
        { 
// *** dirtnA & B are now made ABSOLUTE direction in global co-ords, not local co-ords! ***

            dirtnA = A->Basis * LA->Type->dirtn;
            dirtnB = B->Basis * LB->Type->dirtn;

//cout << "dA " << dirtnA <<"\ndB " << dirtnB << "\ntA " << twistA << "\ntB " << twistB <<"\n";
assert ((dirtnA^twistA) < FPERROR);
assert ((dirtnA^twistA) > -FPERROR);
assert ((dirtnB^twistB) < FPERROR);
assert ((dirtnB^twistB) > -FPERROR);
assert ((dirtnA^twistB) < FPERROR);
assert ((dirtnA^twistB) > -FPERROR);
assert ((dirtnA^dirtnB) < -0.9999 );

            pt orthog = ~(twistA*twistB);  // should be colinear with both dirtn vects.
            if ((orthog^dirtnA) < 0.9999 && (orthog^dirtnA) > -0.9999) 
            {

                cout << "error -> orthog twist vector != link direction " << orthog << ", "<< dirtnA << "\n";
                cout << "raw twistAngle  = " << twistAngle << "; " << twistA << ","<<twistB << "\n";
                cout << "magic dot product = " << (orthog^dirtnA) << "\n";
                return ERROR;
            }

            if ((orthog^dirtnA) > 0) twistAngle = -twistAngle;  

            twistA = A->Basis * LA->Type->twistVect;
            pt axisTemp = LB->Type->dirtn;  
            twistB = B->Basis * CBMatrix::RotationMatrix(axisTemp, twistAngle) * LB->Type->twistVect;
            pt twistB2 = B->Basis * CBMatrix::RotationMatrix(axisTemp, -twistAngle) * LB->Type->twistVect;
            if (twistA != twistB) 
            {
                cout << "dirtnA: " << dirtnA << " , dirtnB: " << dirtnB << "\n";
                cout << "twists: " << twistA << " , " << twistB << " , " << twistB2 << "\n";
                dirtnB = B->Basis * CBMatrix::RotationMatrix(dirtnA, -twistAngle) * LB->Type->dirtn;
                cout << "new dirtnB I " << dirtnB << "\n";
                pt dirtnB2 = (B->Basis * CBMatrix::RotationMatrix(dirtnA, -twistAngle)) * LB->Type->dirtn;
                cout << "new dirtnB II " << dirtnB2 << "\n";
                pt dirtnB3 = (B->Basis * CBMatrix::RotationMatrix(dirtnA, 0.0001)) * LB->Type->dirtn;
                cout << "new dirtnB III " << dirtnB3 << "\n";
            }

        }
    }

    // having found the twist vector, rotate the "B" dimer basis matrix to line up the twist vectors.

    if (twistAngle != 0)
    {
        pt oldBLinkPos = LB->true_pos;

        pt axis = LB->Type->dirtn;  

        B->Basis = B->Basis * CBMatrix::RotationMatrix(axis, twistAngle);  
        B->RecalculateDetails();

        if ((A->Basis*LA->Type->dirtn)!=(B->Basis*LB->Type->dirtn*-1))
        {
            pt bloop1 = (A->Basis*LA->Type->dirtn);
            pt bloop2 = (B->Basis*LB->Type->dirtn*-1);
            cout << "oops.  dirnt twist bad on dimer join.\n";
            cout << "dirtn A: " << bloop1 << "\n";
            cout << "dirtn B: " << bloop2  << "\n";
        }

        assert (LB->true_pos == oldBLinkPos);


        if (LA->Type->twistVect != 0 )
        { 
            pt twistA = A->Basis * LA->Type->twistVect;
            pt twistB = B->Basis * LB->Type->twistVect;

            // fp error can become quite bad at this stage, several basis rotations later.
            // ... so we settle for a fairly crude check...

            assert (pt::aboutequal(twistA,twistB));
/*
            if (twistA != twistB)
            {
                cout << " twist A : " << twistA << "\n twist B : " << twistB << "\n";
                pt temp = twistA - twistB;
                cout << " diff = " << temp.x << ","<<temp.y<<","<<temp.z << " ... " << 1e-5 << "\n";
                exit(-1);
            }
*/
        }
    }

    pt offset = LA->true_pos - LB->true_pos;  // the 'center' pos.s fall out of eqn.
    assert(isnan(offset.x)==FALSE);
    if (offset == 0)
    {
        cerr << "Fatal error for dimers " << A << "\n" << B << "\n";
        cerr << " links are: " << LA << "\n" << LB << "\n";
        long bloop = (LA->Type->BindsTo & LB->Type->TypeId);
        long bloopII = (LA->Type->BindsToHigh & LB->Type->TypeIdHigh);

        cerr << " binding: " << bloop << "," << bloopII << " t: " << TRUE << " f: " << FALSE << "  " << (hex) << bloop << " \n"; 
    assert (LA->CanBind(LB) == TRUE);             // ... dimers must be able to bind each other
    assert (LB->CanBind(LA) == TRUE);             // ... dimers must be able to bind each other
    }

    B->add(offset);                           // linearly move bound dimer to final pos

    assert (A->center != B->center);

    return OK;
}

int Dimer::CoupleLinks(Dimer *A, Dimer *B, Link *LA, Link *LB)
{
   //assert(LA->to == NULL);  // assertion false when mts merging.
   //assert(LB->to == NULL);

   LA->to = B;             // assign dimers to links
   LB->to = A;

    // This may be changed in "SetState()" if the dimers turn out to have changed state

    LA->SetBreakProb(LB);     // set the break probability
    LB->SetBreakProb(LA);
  
    return TRUE;
}

/*
 *  Links two dimers together, adding the second dimer to the microtubule of
 *  the first.
 *
 *  ARGUMENTS:
 *  Dimer B must be unowned, and the Links must be null...
 */

int Dimer::LinkUp(Dimer *A, Dimer *B, Link *LA, Link *LB)
{
    assert ((A->Owner == NULL && B->Owner != NULL) == FALSE); // if only one dimer owned, it must be A 
    assert (LA->CanBind(LB) == TRUE);           // ... dimer links must be able to bind each other

    if ((A->Owner!=NULL)&&(B->Owner!=NULL))
    {
        if (A->Owner->NoObjects<B->Owner->NoObjects)
        {
            // a tweak to make large mts eat small mt.s, not vica-versa
            // (reason: the dimers in the added mt are set free if they cannot
            // be easily added to the original mt.) 
        }
    }

    if (TheProgram.verbose)
        cout << " linking dimers " << A->TagNo << "," << B->TagNo << "\n";

    pt OldPosB = B->center;                    // physically align the dimers
    CBMatrix OldBasisB = B->Basis;

    if (Align(A, B, LA, LB)==ERROR) 
    {
        if (B->Owner != NULL)
        {    
            cout << "resetting dimer pos: " << B->TagNo << " of " << B->Owner->TagNo << "\n";
            B->SetToCoords(OldBasisB, OldPosB);  // reset to old position 
        }
        return ERROR;
    }

    // check against trying to put it somewhere already occupied
    if (A->Owner!=NULL && A->Owner->CheckDimerPos(B) == ERROR)   
    {
        UnLink(A,B);
        assert (LA->to != B && LB->to != A); 
        if (B->Owner != NULL)
        {    
            cout << "resetting dimer pos: " << B->TagNo << " of " << B->Owner->TagNo << "\n";
            B->SetToCoords(OldBasisB, OldPosB);  // reset to old position 
        }
        return (ERROR);  
    }

    // is this necessary?
    if (B->Owner!=NULL && B->Owner->CheckDimerPos(A) == ERROR)   
    {
        UnLink(A,B);
        assert (LA->to != B && LB->to != A); 
        if (B->Owner != NULL)
        {    
            cout << "resetting dimer pos: " << B->TagNo << " of " << B->Owner->TagNo << "\n";
            B->SetToCoords(OldBasisB, OldPosB);  // reset to old position 
        }
        return (ERROR);  
    }

    CoupleLinks(A, B, LA, LB);      // set the dimers to link to each other

    if (Aggregate(A,B) == FAIL)                 // create or modify mt.s appropriately
    {
        if (B->Owner != NULL)
        {    
            cout << "resetting dimer pos: " << B->TagNo << " of " << B->Owner->TagNo << "\n";
            B->SetToCoords(OldBasisB, OldPosB);  // reset to old position 
        }
        return (ERROR);
    }
    else

    assert(A->Owner->IntegrityCheck("End of Link Up")==OK);

    return OK;                 // create or modify mt.s appropriately
}

int Dimer::Aggregate(Dimer *A, Dimer *B)
{
    if ((A->Owner==NULL)&&(B->Owner==NULL))
    {
        Microtubule::CreateNewMicrotubule(A,B);
    }
    else if (A->Owner==NULL)
    {
        if (TheProgram.verbose)
            cout <<"adding dimer"<<A->TagNo<<" to mt: "<<B->Owner->TagNo<<"\n";
        B->Owner->Add(A,B);
    }
    else if (B->Owner==NULL)
    {
        if (TheProgram.verbose)
            cout <<"adding dimer"<<B->TagNo<<" to mt: "<<A->Owner->TagNo<<"\n";
        A->Owner->Add(B,A);
    }
    else
    {
        if (A->Owner == B->Owner)  // we are cross linking within an aggregate
        {
            if (TheProgram.verbose)
                cout <<"crosslinking dimers "<<A->TagNo<<","<<B->TagNo<< " in mt: "<<A->Owner->TagNo<<"\n";
            // which doesn't actually require us to do anything...
        }
        else 
        {
assert(A->Owner->IntegrityCheck("fnord A")==OK);
            if (Microtubule::Meld(A,B) == OK)
                cout << "Melded MTs with dimers " << A->TagNo <<" & " << B->TagNo<<"\n";
            else
            {
                if (TheProgram.verbose) 
                    cout << "Didn't link mts: "<< A->TagNo << "," << B->TagNo << "\n" ;
                Dimer::UnLink(A,B);
                return FAIL;
            }
        }
    }
    return OK;
}


void Dimer::UnLinkAll()
{
    if (TheProgram.verbose)
        cout << "UnLinkAll from dimer: " << TagNo << "\n";
    for (int i=0; i<NoSites; i++)
        if (LinkNo[i].to!=NULL) 
            UnLink(this, LinkNo[i].to);
}

int Dimer::UnLink(Dimer *A, Dimer *B)
{
    if (TheProgram.verbose)
    {
        cout << "unlink " << A->TagNo; 
      if (A->Owner!=NULL) cout << " (mt: " << A->Owner->TagNo << ") "; 
        cout << " and " << B->TagNo;
      if (B->Owner!=NULL) cout << " (mt: " << B->Owner->TagNo << ") "; 
      cout << "\n";
    }

    Link *LA = A->FindLink(B);
    Link *LB = B->FindLink(A);

    if (LA == NULL) 
       if (TheProgram.verbose)
           cout << " dimer " << A->TagNo << " not linked to " << B->TagNo << "\n";  
    if (LB == NULL) 
       if (TheProgram.verbose)
           cout << " dimer " << B->TagNo << " not linked to " << A->TagNo << "\n";  

    if (LA != NULL)
    {
        assert (LA->to == B);    
        LA->to = NULL;
    }

    if (LB != NULL)
    {
        assert (LB->to == A);    
        LB->to = NULL;
    }

    if (A->Owner != NULL && A->GetBound() == FALSE)
    {
        if (TheProgram.verbose)
            cout << "Unlink deleteing " << A->TagNo << " from mt: " << A->Owner->TagNo<<"\n";


        A->Owner->DeleteDimer(A);
        assert (A->Owner==NULL);
        assert (A->GetBound()==FALSE);

    }

    // nb - program rarely executes next bit of code - if mt was a 2 dimer mt,
    //      by now it has been destroyed.  If not, then B is still bound.
    //      rarely, A may have left from the middle of the mt, and B *might*
    //      be stranded.  Theoretically possible.

    if (B->Owner != NULL && B->GetBound() == FALSE)
    {
        if (TheProgram.verbose)
            cout << "rare unlinking of 2nd dimer " << B->TagNo << " in UnLink()\n";
        B->Owner->DeleteDimer(B);
        assert (B->Owner==NULL);
        assert (B->GetBound()==FALSE);

    }
    else
    {
        //cout << " dimer " << B->TagNo << " - someone got to it first...\n";
    }


    A->CheckStateOnBreak(LA);
    B->CheckStateOnBreak(LB);



    return OK;
}


void Dimer::CheckLinkTypes(char* Msg)
{
    for (int j=0; j<NoSites; j++)
    {
        if (LinkNo[j].Type == NULL)
        {
            cout << "NULL LINK ("<<Msg<<"): "<<j<<" Dimer = " << TagNo << "\n" << this << "\n";
        } 
        assert (LinkNo[j].Type != NULL);
    }
}




/*
 *   This sets the position and the orientation of another dimer
 *   to be identical to the current dimer (a transitory state used
 *   in linkage)
 */

void Dimer::SetToMyCoords(Dimer *Clone)
{
    Clone->SetToCoords(Basis, center);
}

void Dimer::SetToCoords(CBMatrix B, pt Center)
{
    assert(isnan(Center.x)==FALSE);
    Basis  = B;
    center = Center;
    RecalculateDetails();
}



/*
 *   This recalculates the position of the alpha and beta subunits,
 *   and the global positions of all the links, using the basis
 *   vectors and the current center position
 */

void Dimer::RecalculateDetails()
{
    for (int i=0; i<NoSites; i++)
    {
        if (LinkNo[i].Type == NULL)
        {
            cout << "NULL LINK (in recalc): "<<i<<" Dimer = " << TagNo << "\n" << this << "\n";
            cout << " Link = " << LinkNo[i] << "\n";
        } 
        assert (LinkNo[i].Type != NULL);
        LinkNo[i] = center + Basis * LinkNo[i].Type->pos;
    }

    for (int i=0; i<NoSpheres; i++) 
    {
        SpherePos[i] = center + Basis * Model->SphereNo[i]->center;
    }
}


/*
 *   Handles (Potential) Dimer collisions
 *
 *   ... a dimer collision is defined as an overlap between two dimers
 */

int Dimer::Collision(Dimer *A, Dimer *B)
{
     int Ret_value = FALSE;

#ifdef VEM_HACK
     if (A->rms_dp == 0 && B->rms_dp == 0) return Ret_value;  // stationary objects can overlap.

     if (A->rms_dp == 0 && B->GetBound() == TRUE) return Ret_value;
     if (B->rms_dp == 0 && A->GetBound() == TRUE) return Ret_value;
#endif

     //nb: "fast_dist_low" is a very inaccurate fast distance routine that always returns low.

     if (pt::fast_dist_low(A->center, B->center) < (A->BoundRad + B->BoundRad))
     {
       // "expensive" accurate check (returns -ve if overlapping)
       // (use -0.01 to avoid rounding error)

         if (Dimer::min_contact_dist_sq(A, B)<-GROSSERROR) 
         {
             // dimers from same mt shouln't collide.
             if ((A->Owner!=NULL)&&(A->Owner==B->Owner))  
             {
                 cerr << "serious error - internal collision between " <<A->TagNo <<",";
                 cerr << B->TagNo << " in mt: " << A->Owner << "\n";
                 cerr << "   dimers: " << A << "\n" << B << "\n";

                 cerr << " Attempting Rescue\n";
//cout << "min dist was : " << Dimer::min_contact_dist_sq(A, B) << "\n";
                 B->free();
                 B->clear();
                 if (A->Owner != NULL) A->Owner->RecalculateDimerPositions();
//cout << "min dist now: " << Dimer::min_contact_dist_sq(A, B) << "\n";
                 assert(Dimer::min_contact_dist_sq(A, B)>-GROSSERROR);
             }
             else 
                 Dimer::Repulse(A, B);             // move dimers apart

             Ret_value = TRUE;                
         }
     }
     return Ret_value;
}


/*
 *   We use a "wrap-around" simulation field, where one side is
 *   supposed to be attached seamlessly to its opposite side.
 *   (a 4D torus.)
 *
 *   For the purpose of comparing things on either side of the
 *   "seam" we may need to move "edge" dimers temporarily to new 
 *   positions outside the field boundaries.
 */

void Dimer::FalsifyPosition(pt offset)
{
    if (offset!=0)
    {

if (pt::abs(offset)>(TheProgram.field_size*2)) cout << "Huge Offset Found: " << offset << "\n";
        assert(isnan(offset.x)==FALSE);
        add(offset);
    }
}


/*
 *  Sometimes, for the purposes of calculation, an "edge" dimer is 
 *  moved to a position outside the boundaries of the simulation 
 *  field. 
 *  This returns the dimer to  a legal, in-bounds position if necessary.
 */

void Dimer::DeFalsifyPosition()
{
    Normalise();       // slow but sure
}


/*
 *  This forces a dimer that is within one field_size of the
 *  simulation field back into bounds.
 *
 *  nb: If the dimer is *more* than one field_size it screams.
 *  this shouldn't happen -> but it's interesting if it does.
 */

void Dimer::Normalise()
{
    pt offset = TheProgram.BoundsOffset(center);
    assert(isnan(offset.x)==FALSE);

    if (offset != 0)
    {
        pt p = center;   // temp for debugging
        
        add(offset);
        if (TheProgram.OutOfBounds(center))        // Double Check that everything is now nice
        {
            cerr<<"Whoops! Dimer loose cycle( "<<TheProgram.NoCyclesCompleted<<"):\n ";
            cerr<<"Originally: " << p << "\n"; 
            cerr<<"Finally:    " << center <<"\n";
            cerr<<"offset: "<<offset<<"\n";
            cerr<<"field size = " << TheProgram.field_size << " : " << TheProgram.int_half_field_size << "\n";

            unbind();
            clear();                    // if not, make sane by brute-force
        }
    }
}



/*
 *  This takes a binding strength, and a distance, and rolls dice to
 *  see whether there is binding. 
 *
 *  The function assumes the interaction is between two unit spheres
 *  moving in a gaussian fashion with a standard deviation of 1.
 *
 *  Strength = the strength of the interaction (i.e. the probability
 *             that if right binding sites *did* collide, the would
 *             bind together)
 *  Dist            = Distance between the surface of the objects
 *  Colliding Types = An Index into the types of objects that 
 *                    may be colliding.  
 *
 */

int Dimer::probftn(float Strength, float DistSq, Dimer* A, Dimer* B) 
{
    if (Strength==0) return 0;  // no chance of binding, so don't bother going
                                // on.

    // Fast Method for free objects...
    float Prob;

    int BoundA = A->GetBound();
    int BoundB = B->GetBound();
 

    if ((BoundA==FALSE) && (BoundB==FALSE))
    {
        float Dist = TheProgram.fastsqrt(DistSq); 
        Prob = A->Model->CollisionProb(B->Model,Dist) * Strength;
//cout << "Prob Normal @ dist ("<<Dist<<") = " << A->Model->CollisionProb(B->Model,Dist) << " * " << Strength << " = " << Prob << "\n";
    }
    else
    // Slow method for bound objects
    {
        /*
         *   One or more of the particles is bound.  This reduces its
         *   effective rms movement -> which changes the equation for
         *   collision probabilities.  Rather than using rms, we'll use
         *   Variances for a while and convert at the end.
         *
         *   EXPLANATION: What we are doing here is changeing the *scale*
         *   of the problem.  The slower moving bound particles have
         *   smaller rms movements, hence distances between them are
         *   "effectively" larger, and their radii, as a proportion of
         *   rms movement, are "effecively" bigger.
         *
         *   TODO:  Speed up?
         */
      
        if ((BoundA&&A->Owner==NULL) || (BoundB&&B->Owner==NULL))
        {
            cout << "AWOOGA ->BadOwner!\n"; return 0;  //error - bail
        }

        // Calculate movement variance changes for A & B (may be none if unbound) 
        // variance of movement of the two objects
        float rmsA = (BoundA) ? A->Owner->rms_dp : A->rms_dp; 
        float rmsB = (BoundB) ? B->Owner->rms_dp : B->rms_dp; 

        float var_mov = (rmsA*rmsA + rmsB*rmsB);  // variance of movement

        float EffDist = sqrt(DistSq/var_mov);
        float EffRadiusSum = ((A->BoundRad + B->BoundRad) / sqrt(var_mov));        
        EffDist -= EffRadiusSum;  // we want distance between sphere surfaces. 


assert (isnan(EffDist)==FALSE); 
assert (isnan(EffRadiusSum)==FALSE); 

        Prob = A->Model->Parser->CollProb(EffDist, EffRadiusSum) * Strength; 
//cout << "Prob for Binding 2 objects, one bound, = " << Prob << "\n";
    }
     
    return ((Prob)>R1);
}

/*
 *  When two objects interact, they may link together.
 *  This procedure checks to see if this is completely
 *  impossible.  (Most of the conditions here would be 
 *  picked up later anyway, but this saves time and is
 *  safer.)
 */

int Dimer::CheckInteraction(Dimer *A, Dimer *B)
{
#ifdef SAFE     // Error catching code - should never trigger (:-) )
    if (  ((A->GetBound()==FALSE) || (B->GetBound()==FALSE)) && 
          ((A->Owner!=NULL) && (B->Owner!=NULL))  )
    {
        if (A->GetBound()==FALSE)
        {
            cerr << "Dimer Interact ERROR ";
            cerr << A->TagNo<<" is really owned by "<<A->Owner->TagNo<<"\n";
//            if (A->Owner->IntegrityCheck("Dimer::CheckInteraction A ")==ERROR)
                delete(A->Owner);
        }
        if (B->GetBound()==FALSE)
        {
            cerr << "Dimer Interact ERROR ";
            cerr << B->TagNo<<" is really owned by "<<B->Owner->TagNo<<"\n";
//	    if (B->Owner->IntegrityCheck("Dimer::CheckInteraction B ")==ERROR)
		delete(B->Owner);
        }
    }
#endif

    // Checks to see whether interaction is impossible...

    if (A->GetFullyBound()==TRUE)                // dimer A has no free links
        return (DONOTHING); 

    if (B->GetFullyBound()==TRUE)                // dimer B has no free links
        return DONOTHING;

     if ((A->Owner!=NULL)&&(B->Owner!=NULL)&&(A->Owner==B->Owner)) // they're from the same aggregate
        return DONOTHING;

    return (OK);
}


/*
 *   utility ftn: returns TRUE if every link on the dimer is already bound
 */ 

int Dimer::GetFullyBound()
{
    for (int i=0; i<NoSites; i++)
         if (LinkNo[i].to==NULL) return FALSE;
    return TRUE;
}


/*
 *  Checks whether the two links that are about to be bound are both
 *  free.
 */

int Dimer::LinksAlreadyBound(Link *Alink, Link *Blink)
{
//    return ((Alink==NULL)||(Blink==NULL)||(Alink->to!=NULL)||(Blink->to!=NULL));
    if ((Alink==NULL)||(Blink==NULL))
    {
        cout << "Null links sent to 'LinksAlreadyBound()'\n";
        return true;
    }

    if ((Alink->to!=NULL)||(Blink->to!=NULL))
    {
//        cout << "Links already bound\n";
        return true;
    }

//    cout << "returning unbound\n";
    return false;
}


/*  
 *  We need to know which two matching links are closest
 *
 *  This returns the two closest links and the square of their distance
 *
 */

void Dimer::GetLinksAndDistSq(Dimer *A, Dimer *B, Link **Alink, Link **Blink, float &Dist)
{
    int i,j;
    float D;
    
    Dist = 9999; 
    for (i=0; i<A->NoSites; i++)
    {
        for (j=0; j<B->NoSites; j++)
        {
            if (A->LinkNo[i].CanBind(&B->LinkNo[j]))
            {
                D = pt::sqr(A->LinkNo[i].true_pos,B->LinkNo[j].true_pos); 
                if (D<Dist)
                {
                    Dist = D; 
                    *Alink = &(A->LinkNo[i]);
                    *Blink = &(B->LinkNo[j]);
                }
            }
        }
    }
    if (Dist == 9999)
    {
//        cout << "unable to find bindable links for ("<<A->TagNo<<") ";
//        cout << A->Model->Name << " and ("<< B->TagNo<<") "<< B->Model->Name << "\n";
 
//   Actually, it's quite common for objects to collide which can't link...
       
        Alink = NULL; 
        Blink = NULL; 
    }
}


/*
 *   Calculate what the basic (unmodified by distance) chance of two
 *   objects binding together is. 
 */ 

float Dimer::GetBindingStrength(Link *Alink, Link *Blink)
{
    if (Alink->CanBind(Blink))
    {
        float temp = Alink->Type->GetBindProb(Blink->Type->LinkNo);
        return temp;
    }
    else
        return 0; 
}


void Dimer::printlinks(int PrintAllLinks)
{
    for (int i=0; i<NoSites; i++)
    {
        if (PrintAllLinks || (LinkNo[i].to != NULL))
        {
            cout << "link no " << i << " ("<<LinkNo[i].Type->LinkNo<<") " << LinkNo[i].Type->pos;
            if (LinkNo[i].to != NULL)
            {
                cout << " linked to dimer #" << LinkNo[i].to->TagNo;
            }
            cout << "\n       global pos " << LinkNo[i].true_pos;
            cout << "\n";
        }
    }
}

/*
 *   Change to a new State, and modify links and sphere positions
 *   appropriately. (Colours are handled seperately in the display 
 *   file "surfette.cc")
 */

void Dimer::SetState(int NewState)
{
    State = NewState;
    SetLinks();
    // eventually... SetSpheres();

    // SPEED UP POSSIBILITY:
    //  - if this is too slow, stick flags in seperate data struct.

    // Check a flag if there is a random event defined for this state...
    RandomEventFlag = FALSE;
    for (int i=0; i<Model->NoEvents; i++)
    {
        if (Model->EventNo[i]->Type == RANDOMEVENT)
        {
            if (Model->EventNo[i]->OldState == State)
            {
                RandomEventFlag = TRUE;
            }
        }
    }
}


/*
 *  There is a possibility that the object will undergo a random state change
 */

void Dimer::CheckStateOnRandomEvent()
{
    for (int i=0; i<Model->NoEvents; i++)
    {
        if (Model->EventNo[i]->Type == RANDOMEVENT)
        {
            if (Model->EventNo[i]->OldState == State)
            {
                if (R1 < Model->EventNo[i]->RandomVal)
                {
//                    cout << TheProgram.NoCyclesCompleted << " RANDOM EVENT (of " << TagNo << ") CHANGING STATE FROM " << State;
//                    cout << " TO " << Model->EventNo[i]->NewState<<" !!!\n";

                    SetState(Model->EventNo[i]->NewState);



                    break;                           // EXIT
                }
            }
        }
    }
}


/*
 *   See if a Dimers State has changed due to binding
 */

void Dimer::CheckStateOnBreak(Link* from)
{
   if (from == NULL) return;  // A dimer may break off during binding, only partially
                              // linked with its neighbour.

   for (int i=0; i<Model->NoEvents; i++)
    {
        if (Model->EventNo[i]->Type == BREAKEVENT)
        {

            if (Model->EventNo[i]->OldState == State)
            {
// cout << " comparing state " << State << ":" << Model->EventNo[i]->Site << " with " << from->Type->SiteNo << "\n"; 

                /*
                 *  See whether this particular link breaking is marked for special
                 *  treatment, or whether this is the *last* link broken, and there
                 *  is an "ALL" links broken state change.
                 */

                if (Model->EventNo[i]->Site == from->Type->SiteNo ||
                    (GetBound()==FALSE && Model->EventNo[i]->Site == ALL) )
                {
                   if (debugEvents) 
                       DATALOG5 << TheProgram.NoCyclesCompleted << ": " 
                       << "changing state of #" << TagNo << " on break of " 
                       << Model->GetSiteName(Model->EventNo[i]->Site) << " from " 
                       << Model->GetStateName(Model->EventNo[i]->OldState) << " to " 
                       << Model->GetStateName(Model->EventNo[i]->NewState) << "\n";

                   SetState(Model->EventNo[i]->NewState);
                   break;                           // EXIT
                }
            }
        }
    }
}


/*
 *   See if a Dimers State has changed due to binding
 */

void Dimer::CheckStateOnBind(Link* from, Link* to)
{
    if (debugEvents)
         DATALOG5 << TheProgram.NoCyclesCompleted << ": binding " << TagNo << " in state " 
                  << Model->GetStateName(State) << " at site "
                  << Model->GetSiteName(from->Type->SiteNo) << "\n";
    if (from==NULL||to==NULL) return;

    for (int i=0; i<Model->NoEvents; i++)              // work through all known events
    {
        if (Model->EventNo[i]->Type == BINDEVENT)      // aha! found a bind event.  Now see...
        {
            if (Model->EventNo[i]->OldState == State)  // if it's a bind *from* the state the dimer is now in
            {                                          // ... and between the correct sites
//TODO:  make this work for "all" site links...
                if (((Model->EventNo[i]->Site == from->Type->SiteNo ) || (Model->EventNo[i]->Site == ALL)) &&
                    ((Model->EventNo[i]->LinkedSite == to->Type->SiteNo) || (Model->EventNo[i]->LinkedSite == ALL)))
                {
                   if (debugEvents) 
                       DATALOG5 << TheProgram.NoCyclesCompleted << ": "
                       << "changing state of #" << TagNo << " on bind of " 
                       << Model->GetSiteName(Model->EventNo[i]->Site) << " to site " 
                       << Model->GetSiteName(Model->EventNo[i]->LinkedSite) << " from "  
                       << Model->GetStateName(State) << " to " 
                       << Model->GetStateName(Model->EventNo[i]->NewState) << "\n";

                   SetState(Model->EventNo[i]->NewState);
                   break;                           // EXIT 
                }
            }
        }
    }
}
/*
 *  Check to see whether the dimers link together...
 */

int Dimer::Interact(Dimer *A, Dimer *B)
{
if (A->Owner != NULL) A->Owner->IntegrityCheck("start interact()");
if (B->Owner != NULL) B->Owner->IntegrityCheck("start interact()");


    if (TheProgram.collision_test == TRUE)   // collision test is a special mode where we
    {                                        // check how often dimer '0' is collided with...
        if (A->TagNo != 0 && B->TagNo != 0) 
            return DONOTHING;
    }

    //-----------Set up the ground work----------------------------------------------


    int BindingMTs = FALSE;                   // Are we binding aggregates?

    if (CheckInteraction(A,B)!=OK)            // If they can't interact, don't try.
        return (DONOTHING);
    
    if ((A->Owner!=NULL) && (B->Owner!=NULL)) // set a flag, warning us that we
    {                                         // are binding two aggregates. 
        BindingMTs = TRUE;                    // nb: "CheckInteraction()" has
    }                                         // already made sure they're different
 
    if ((A->GetBound()==FALSE) && (B->GetBound()==TRUE)) // saves code... if only one dimer
    {                                                    // is bound, make it dimer 'A'
        Dimer *temp = A; A = B; B = temp;
    }


#ifdef VEM_HACK
    if (A->rms_dp==0 && B->rms_dp==0)
        return (DONOTHING);

    if (B->rms_dp==0)
    {
        if (A->GetBound())
        {
            cout << " Evil VEM HACK Wierdness. Pretending not to notice.\n";
            return (DONOTHING);
        }
        Dimer *temp = A; A = B; B = temp;
    }
#endif 

    //----Find the two closest matched links, and their seperation---------------------

    Link *Alink = NULL;
    Link *Blink = NULL;
    float distance_sq  = 0;
    GetLinksAndDistSq(A,B,&Alink,&Blink,distance_sq);      // modifies Alink/Blink/distance

    if (Alink == NULL || Blink == NULL)  //objects can't bind: not an error, happens quite often.
        return DONOTHING;

    if (LinksAlreadyBound(Alink, Blink))      // check links aren't already used
        return (DONOTHING);

    if (Alink->CanBind(Blink) == FALSE)
    {
        cerr << "bad links slipped through GetLinksAndDistSq...\n";
        return (DONOTHING);
    }

    /*------------------------------------------------------------------------------- 
     *    Everything is now in readiness!  Roll the dice, and see if the
     *    thingumies bind together.
     *
     *--------------------------------------------------------------------------------*/

    // otherwise...

    if (TheProgram.collision_test == TRUE)
    {
        if ((A->TagNo!=0)&&(B->TagNo!=0)) cout << "Coll Test Error 1\n";
        if (probftn(1, distance_sq, A, B))
        { 
            NoCollisions++;
        }
    }
    else if (probftn(GetBindingStrength(Alink,Blink), distance_sq, A, B))
    {
        if ((Alink->CanBind(Blink)==false) || (Blink->CanBind(Alink)==false))
        {
            cout << "naughty dimers = " << A->Model->Name << " & " << Alink->Type->Name << " linking " << B->Model->Name << " & " << Blink->Type->Name<<"\n";
            if (Alink->CanBind(Blink)==false) cout << " First can't link second ";
            if (Blink->CanBind(Alink)==false) cout << " Second can't link first";
        }

        assert (Alink->CanBind(Blink));
        assert (Blink->CanBind(Alink));
        if (LinkUp(A, B, Alink, Blink) == OK) // link the dimers up
        {
            if (TheProgram.verbose)  
                cout << " linking dimers " << A->TagNo << "," << B->TagNo << " after collision\n ";
            A->CheckStateOnBind(Alink, Blink);
            B->CheckStateOnBind(Blink, Alink);
        }
        else // unable to link dimers  - may be physically overlapping something else etc.
        {
            if (TheProgram.verbose)  
                cout << " unable to link dimers after collision\n ";

            return DONOTHING;
        }

        return(BOUND);
    }
    else
            if (TheProgram.verbose)  
                cout << " dimers not bound \n";
    // otherwise...
    // the dice say "no binding today"...

    return(DONOTHING);
}


/*
 *   Calculate the minimum distance vector required to move two objects
 *   away so that they are only just touching.
 *
 *   This is a simplification, introduced to avoid the computational overhead
 *   of strictly modelling velocities and collisions (which would be tricky,
 *   due to the brownian nature of motion at this scale).
 *
 *   so what we do, is when objects overlap, we try to move them apart so
 *   that they're just touching.
 *
 *   It's possible that instead it might be better to give them a bit of a
 *   push away from each other.  Hmmm.
 *
 *   A further problem:  This algorithm breaks down for large aggregates
 *   
 *
 *
 *   CONSIDER:  This is very slow.  Maybe a quick hack could be done here.
 *              (Such as using the bounding envelopes only...) (DONE 4/12/97)
 */

pt Dimer::RepulsionVector(Dimer *A, Dimer *B)
{
    pt cent_diff = B->center - A->center;

    float sum_rads = B->BoundRad + A->BoundRad; 

    pt Rep_Vect(0);

    float cent_diff_mag = pt::abs(cent_diff);

    if (cent_diff_mag > sum_rads) return Rep_Vect;  // shouldn't have been called!

    Rep_Vect = ~cent_diff * (cent_diff_mag - sum_rads)/2; 

//    cout << " Seperating "<<B->TagNo<<" & "<<A->TagNo<<" " << B->center << "/" <<B->BoundRad<<" from "<<A->center << "/" <<A->BoundRad<<"\n      dist = " << cent_diff << " diff vect = " << Rep_Vect*2<<"\n";


    return Rep_Vect;
}


/*
 * If two dimers collide, but don't bind, they bounce away a bit...
 */

void Dimer::Repulse(Dimer *A, Dimer *B)
{
    pt DebugA = A->center;
    pt DebugB = B->center;

    static count = 0;
    static countbad = 0;


    if (A->GetBound() && A->Owner==NULL) {cerr<<"Fuckup with Dimer"<<A->TagNo<<"\n"; A->clear(); return;}
    if (B->GetBound() && B->Owner==NULL) {cerr<<"Fuckup with Dimer"<<B->TagNo<<"\n"; B->clear(); return;}

    pt Repulsion = RepulsionVector(A, B);

    assert(isnan(Repulsion.x)==FALSE);

    if (Repulsion==0) {cout << " Unneccessary call for Repulse " << A->TagNo<<" "<<B->TagNo<<"\n";return;}

    if (A->GetBound()==FALSE && B->GetBound()==FALSE )   // move both dimers equally
    {
#ifdef VEM_HACK
        if (A->rms_dp==0 && B->rms_dp==0)
        {
            cout << " Error - Repulse called for static objects\n";
        }
        else if (A->rms_dp==0)
        {
            B->add(Repulsion*-2);
        }
        else if (B->rms_dp==0)
        {
            A->add(Repulsion*2);
        }
#endif
        A->add(Repulsion);
        B->add(Repulsion*-1);
    }
    else if (A->GetBound()==FALSE)                      // assume only the free dimer moves
        A->add(Repulsion*2);
    else if (B->GetBound()==FALSE)                      // assume only the free dimer moves 
        B->add(Repulsion*-2); 
    else                                                //move m.t.s...
    {
        if (A->Owner != B->Owner)
        {
        //Move the whole microtubule.  Shouldn't actually happen very often.
            //float totalmass = A->Owner->Mass + B->Owner->Mass;
            if (A->Owner->Mass == 0) cerr << " MT: " << A->Owner->TagNo << " has no mass???\n";
            if (B->Owner->Mass == 0) cerr << " MT: " << B->Owner->TagNo << " has no mass???\n";
            
            assert(isnan(Repulsion.x)==FALSE);

            if (pt::abs(Repulsion)>100)
                cerr<< "large repulsion in dual MT coll. : " << Repulsion <<"
n";

            pt Avect = Repulsion * 1;// 2  * (B->Owner->Mass/totalmass);
            pt Bvect = Repulsion * -1;//2  * (A->Owner->Mass/totalmass);
           
            assert(isnan(Avect.x)==FALSE);
            assert(isnan(Bvect.x)==FALSE);

            A->Owner->MoveLinearly(Avect);
            B->Owner->MoveLinearly(Bvect);
        }
        else  //OOPS - Error: why are two dimers from the same m.t. "colliding"???
        {
            
            Link *Alink, *Blink; 
            float Dist;
            GetLinksAndDistSq(A, B, &Alink, &Blink, Dist);
 
            float bloopII = min_contact_dist_sq(A,B);

            cout << "Wierdness - internal collision: Distance = " << bloopII;
            cout << " link dist: " << sqrt(Dist) << "\n"; 
            cout << "oh oh - internal collision... aggregate" << A->Owner->TagNo << "( ";
            cout << A->Owner->NoObjects << " objects)\n"; 
            cout << "Wierdly deleteing dimer " << A->TagNo << " from " << A->Owner->TagNo << "\n";
        cerr << " Dimer A = " << A << "\nDimer B = " << B << "\n";

            delete(A->Owner);

/* EMERGENCY HACK
            A->unbind();
            A->Owner->DeleteDimer(A);
            A->center = pt(RpmS, RpmS, RpmS);
            A->RecalculateDetails();
*/
        }
    }

    /*
     *   (apparently) It's possible, although not very likely, that the dimers will *still* 
     *   be touching, so we recurse for a while continuing to try.
     */

    float bloop = min_contact_dist_sq(A,B);
    if (bloop<-0.001)
    {
#ifdef VEM_HACK
        if (A->rms_dp==0 || B->rms_dp==0) return; // we don't mind things overlapping with statics
#endif
        countbad++;
        cerr << "\n";
        cerr << TheProgram.NoCyclesCompleted << "Error -> Dimers " <<A->TagNo<<" & "<<B->TagNo<<"still Touching after Repulse: dist= " << bloop<< " \n";
        cerr << "Repulsion = " << Repulsion << " center A "<<DebugA<<" "<< DebugB << "\n";
        cerr << "Now : A " << A->center << " and B " << B->center << "\n";
        pt dist = A->center - B->center;
        cerr << " curr dist : " <<  dist << " ( = " << pt::abs(dist) << ")\n";
        cerr << " proportion : " << countbad << " /  " << count << "\n";
        if (A->GetBound()) cerr << " A Bound ";
        if (B->GetBound()) cerr << " B Bound ";
        cerr << " Dimer A = " << A << "\nDimer B = " << B << "\n";
        assert (FALSE);

    }
    count++;
if (pt::abs(B->center)>TheProgram.field_size)
    cout << "\nWierdness in repulse - Dimer " << B << "\nRepulsion = " << Repulsion ; 
if (pt::abs(A->center)>TheProgram.field_size)
    cout << "\nWierdness in repulse - Dimer " << A << "\nRepulsion = " << Repulsion ; 
}


/*
 *   Gets the index of a dimer in the global dimer array.
 *   Only used for debugging.  Way Slow.
 */

int Dimer::Get_pos(Dimer* P)
{
    for (int i=0; i<NoObjects; i++)
        if (&Dimers[i] == P) return(i);
    return(-1);
}


/*
 *  Give a gigantic data dump of all the dimers.
 *  Only used for debugging.
 */

void Dimer::Dump_dimers()
{
    if (TheProgram.silent) return;

    ERRORLOG << "Dimer Dump\n";
    int i;
    Dimer* P = NULL;
    for (i=0; i<NoObjects; i++)
    {
        P = &Dimers[i];
        if (P->GetBound())
        {
            ERRORLOG << "\nDimer["<<i<<"] : " << P;
        }
    }
    ERRORLOG << "\nUnbound Dimers\n";
    for (i=0; i<NoObjects; i++)
    {
        if (!P->GetBound())
        {
            ERRORLOG << "\nDimer["<<i<<"]   -  unbound " << P;
        }
    }
    ERRORLOG << "\nEnd DimerDump\n";
}


/*
 *  What is the chance that this dimer will break away from its 
 *  owning aggregate.
 *
 *  HEURISTIC:  Multiply Bound Links have a total breakage chance
 *              equal to the strongest link (i.e. lowest break prob)
 *              divided by 10 for each extra link...
 *
 *
 *
 *  CONSIDER:  getting this working so as to include state info
 *             properly
 */

float Dimer::BreakageChance()
{
    float BreakChance = 0;
    int NoBindingLinks = 0;
    
    for (int i=0; i<NoSites; i++)
    {
       if (LinkNo[i].Bound())
       {
           NoBindingLinks++;
           if ((BreakChance==0) || (BreakChance > LinkNo[i].breakprob)) 
               BreakChance = LinkNo[i].breakprob;
       }
    }
    

    if (BreakChance == 1)  // Shouldn't happen - recoverable error...
    {
        cout << "\n" << TheProgram.NoCyclesCompleted << ": "
             << "warning: certain breakage in dimer " << TagNo << "\n state: "
             << Model->GetStateName(State) << "\n";

        for (int i=0; i<NoSites; i++)
        {
            if (LinkNo[i].Bound())
            {
                cout << "link site: (" << Model->GetSiteName(i) << ")  break = " 
                     << LinkNo[i].breakprob << "\n"; 
                cout << "bound to dimer " << LinkNo[i].to->TagNo << " state: " 
                     << Model->GetStateName(LinkNo[i].to->State) << "\n" ;

                for (int j=0; j<NoSites; j++)
                {
                    if (LinkNo[i].to->LinkNo[j].to == this)
                        cout << " link site: (" << Model->GetSiteName(j) 
                             << ") " << LinkNo[i].to->LinkNo[j].breakprob << "\n";
                }
            }
        }
    }
//assert (BreakChance != 1);

    if (NoBindingLinks > 1)
    { 
       if (TheProgram.NoMultLinkBreakage==TRUE)
           BreakChance = 0; 
       else  
           for (int i=0; i<NoBindingLinks; i++)
           {
               BreakChance = BreakChance/10.0;
           }
    }

    return BreakChance;
}


/*
 *  Bound dimers may break away from the microtubule...
 *
 */

void Dimer::Check_binding()
{
    float Breakage;

    //TODO: There's some nifty way to iterate a pointer though a list.
    // find it and do it.

    Dimer* D;

    for (int i=0; i<NoObjects; i++)
    {
        if (Dimers[i].GetBound() == TRUE)
        {
            D = &Dimers[i];

            if (D->Owner==NULL)     // shouldn't happen.  But we is a cautious cyberdweeb... 
            {
                cerr << "Found bad dimer " << D << " -> unbinding ";
                D->unbind();
            }
            else
            {
                Breakage = D->BreakageChance();

                if (R1<Breakage) 
                {
                    if (!TheProgram.silent) ERRORLOG << "\nUnbinding Dimer ["<<i<<"] "<<D;
                    if (!TheProgram.silent) ERRORLOG << " with prob of " << Breakage<<"\n";
                    if (TheProgram.verbose) 
                       cout << "breaking off "<<D->TagNo<<" with prob: "<<Breakage<<"\n";

                    D->unbind();
                    assert (D->Owner == NULL);
                }
            }
        }
    }
}


/*
 *    Free, unphosphorelated dimers get a chance to rephosphorelate (we assume lots of 
 *    nice enzymes around).
 */

void Dimer::Random_events()
{
    for (int i=0; i<NoObjects; i++)
    {
        Dimer *D = &Dimers[i];
        if (D->RandomEventFlag == TRUE)
        {
            D->CheckStateOnRandomEvent(); 
        }
    }
}


/*
 *     This *should* be where all the data integrity checks are done
 */

void Dimer::Integrity_check()
{
    //This is time consuming. We only do it every 100 cycles.

#ifdef EVIL_VEM_HACK
    for (int i=0; i<NoObjects; i++)
    {
        if (Dimer::Dimers[i].rms_dp !=0)
        {
            if (Dimer::Dimers[i].center.z < 0)
            {
                Dimer::Dimers[i].center.z = - Dimer::Dimers[i].center.z;  // le hack brutal...
                Dimer::Dimers[i].RecalculateDetails();
            }
   
            if (Dimer::Dimers[i].center.z > (Half_Field_Size - 10))
            {
                Dimer::Dimers[i].center.z = Dimer::Dimers[i].center.z - 10; 
                Dimer::Dimers[i].RecalculateDetails();
            }
        }
    }
#endif

    Dimer *D;
    for (int i=0; i<NoObjects; i++)
    {
        D = &Dimer::Dimers[i];
        if (TheProgram.OutOfBounds(D->center))
        {
/*
            cout << "INTEGRITY ERROR: cycle: " << TheProgram.NoCyclesCompleted << "\n"; 
            cout << "Dimer ";
            cout << D->TagNo << " in bad position:  " << D->center;
            if (D->GetBound() == TRUE)
                cout << " bound to " << D->Owner->TagNo;
            else
                cout << " unbound "; 
            cout << "\n";
*/
        }

        D->CheckLinkTypes("int. check");

        D->CheckLinks();

        if (D->GetBound() == TRUE && D->Owner == NULL)
        {
            cout << "INTEGRITY ERROR: cycle: " << TheProgram.NoCyclesCompleted << "\n"; 
            cout << "Bound dimer " << D->TagNo << " is unowned\n";
        }
        else if (D->Owner != NULL && D->GetBound() == FALSE)
        {
            cout << "INTEGRITY ERROR: cycle: " << TheProgram.NoCyclesCompleted << "\n"; 
            cout << "Owned dimer " << D->TagNo << " is unbound\n";
        }
    }

    if (TheProgram.NoCyclesCompleted%1000==0)
    {
        DATALOG5 << "\n\n\nDATA DUMP CYCLE " << TheProgram.NoCyclesCompleted <<"\n\n";         
        for (int i=0; i<NoObjects; i++)
        {
            D = &Dimer::Dimers[i];
            DATALOG5 << " #" <<  D->TagNo << " " << D->Model->GetStateName(D->State);
            if (D->Owner != NULL) DATALOG5 << "   owner = " << D->Owner->TagNo;
            DATALOG5 << "\n";
        }     
    } 
}


/*
 *  This static function goes through all unbound dimers, "bumping" them randomly.
 */

void Dimer::Agitate()
{
    for (int i=0; i<NoObjects; i++)
    {
        if (Dimers[i].GetBound() == FALSE) Dimers[i].bump();
    }
}


/*
 *  No idea.
 */

pt Dimer::Sproing(pt A, pt B, float spring_const)
{
    return ((A-B)*spring_const);  
}


/*
 *  This repositions a dimer that is part of a mt. lattice.
 *
 *  Centroid: The pt around which the dimer is rotated.
 *  AngVel:   "Angular Velocity" - in practice, the ang.displacement vector
 *  MTLocalZ: the "+" direction of the MT, to which the dimers are aligned
 */

void Dimer::rotate_bound_dimer(pt Centroid, pt AngVel)
{
    rotate_bound_dimer(Centroid, ~AngVel, pt::abs(AngVel));
}

void Dimer::rotate_bound_dimer(pt Centroid, pt axis, float angle)
{
   assert(isnan(Centroid.x)==FALSE);
   assert(isnan(axis.x)==FALSE);
   assert(isnan(angle)==FALSE); 

   assert(isnan(center.x)==FALSE);
   center -= Centroid;
   center.Rotate(axis, angle);
   assert(isnan(center.x)==FALSE);
   center += Centroid;

//Owning MT already has this info.  should be passed through...

   Basis = CBMatrix::RotationMatrix(axis, angle) * Basis;

   RecalculateDetails();
}


/*
 *  This is called by the microtubule movement routine
 *  if the solid microtubule approximation is *not* used.
 */

void Dimer::bump_bound_dimer()
{
    if (Owner!=NULL)
    {
        assert(isnan(Owner->velocity.x)==FALSE);
        add(Owner->velocity);
    }
    else
        cerr<<"ERROR - Dimer " <<TagNo << " Has no parent in bump_bound_dimer : bound = " << GetBound()<<"\n";

}


/*
 *  Load up the Binding Matrix.
 *  Hard coded for the nonce.
 *  Generalise later.
 */

void Dimer::CalculateBindingMatrix()
{
cout << "No Binding Matrix";
//Still a good idea.  Should do this.

/*     BindMatrix.Resize(2,4,4);

     float topbind = (bindtop+bindbottom)/2;

     BindMatrix(TOP,BOTTOM) = topbind;
     BindMatrix(BOTTOM,TOP) = topbind; 
     BindMatrix(LEFT,RIGHT) = bindside;
     BindMatrix(RIGHT,LEFT) = bindside; 
*/
}


int Dimer::CheckFieldSize(int Field_Size)
{
    if ((Field_Size*Field_Size*Field_Size/16)<=NoObjects)
    {
        cerr << "\nERROR - field size ("<<Field_Size<<" microns**3) is";
        cerr << " too small to contain " << NoObjects << " dimers.\n";
        cerr <<"min field size would be approx. ";
        cerr << int (1+pow((NoObjects*16),0.33333)) << "\n";
        return -1; 
    }
    return 0;
}

/*
 *  Initialises the global dimer variables, such as the scale, the
 *  size of the simulation field (probably unecessary here), the
 *  total number of dimers - and finally creates all the dimers
 */

TubParser* Dimer::Init(int free_objects, int FieldSize, char *PDDFStr, char* StructureString)
{
    int StructObjects = 0;

    NoCollisions = 0;
    if (PDDFStr==NULL) PDDFStr = "info.pddf";

    TubParser* Parser = new TubParser();
    Parser->parse(PDDFStr);

    Field_Size = FieldSize;
    Half_Field_Size = FieldSize/2;

    if (CheckFieldSize(Field_Size)) return NULL;

    NoObjects = free_objects;

    cout << "No Objects = " << NoObjects << "\n";

    if (StructureString!=NULL)
    {
        StructObjects = Parser->ParseStructureFile(StructureString);
        NoObjects += StructObjects;
    }
    cout << "Total objects of all kinds: " << NoObjects << " \n";

    Dimers = new Dimer[NoObjects];

    int GroupStart = 0;

    for (int m=0; m<Parser->NoModels; m++)
    {
        int NoModelObjects=0;
        int NoStructureObjects = (StructureString==NULL)?0:Parser->Structure->NoType(m);

        /*
         *  Find how many free objects of this type we will create.
         *  If this is the last type, use a different calculation to make
         *  up for any rounding error, and wind up with the right number
         *  of total dimers.
         */

        if ( m == (Parser->NoModels-1) )  // last model - set to all remaining dimers.
        {
           NoModelObjects = NoObjects - GroupStart - NoStructureObjects; 
        }
        else
        {
           NoModelObjects  = (int) ( ((float)free_objects) * Parser->ModelNo[m]->Freq );

        }

        // set the data for the newly created objects to random initial values.
        for (int n=0; n<NoModelObjects; n++)        
        {                                         // Because C++ init.ing of array 
            assert (Parser->ModelNo[m]!=NULL);
            Dimers[GroupStart+n].SetData(Parser->ModelNo[m]);     // objects sucks.
        }
       
        // Handles (optional) pre-built structures. 
        if (StructureString != NULL)
        {
            NoStructureObjects =  Parser->Structure->NoType(m);
cout << "Initing " << NoStructureObjects << " Structure Objects\n";
            for (int n=0; n<NoStructureObjects; n++)
            {
                 int index = GroupStart+NoModelObjects+n;
                 assert (Parser->ModelNo[m]!=NULL);
                 Dimers[index].SetData(Parser->ModelNo[m],Parser->Structure->GetData(m,n));
            }
        }
        GroupStart += NoModelObjects + NoStructureObjects;
    }
    
    return Parser;
}


/*
 *  break all bindings and assign random new position  - used when running 
 *  multiple test runs, or when a dimer gets itself into a difficult-to-
 *  extricate-from error condition.
 */

void Dimer::clear()    
{
    if (Owner!=NULL && Owner->NoObjects>2)
        cout << " clearing dimer from " << Owner->TagNo; 
    cout << " (clearing dimer " << TagNo <<")\n";

    free();

    center = pt(RpmS, RpmS, RpmS);

    assert(isnan(center.x)==FALSE);

    spin();

    RecalculateDetails();

    State = 0;  // set back to initial state.
    SetState(0);
}

/*
 *   Break all bindings.
 */

void Dimer::free()
{
   if (Owner!=NULL)
    {

cout << "freeing dimer " << TagNo << " from mt: " << Owner->TagNo << "\n";
        unbind();
        assert(Owner==NULL);
    }
    else 
    {
cout << "freeing unowned dimer " << TagNo << "\n";
        unbind();
    }
    State = 0;  // set back to initial state.

}

/*
 *  Checks no dimers have gone bonkers. 
 */

void Dimer::CheckAll(char* Msg)
{
    for (int i=0; i<NoObjects; i++)
    {
       if (isnan(Dimers[i].center.x)==TRUE)
       {
           cout << "\n\n" << Msg << " Dimer " << i << " stuffed ";
           if (Dimers[i].Owner != NULL) cout << " owned by: " << Dimers[i].Owner->TagNo;
           Dimers[i].clear();
           cout << "\n";
       }
           
       assert (isnan(Dimers[i].center.x)==FALSE); 
    }
}

/*
 *   Resets all dimers back to a general starting state. 
 */

void Dimer::ClearAll()
{
    for (int i=0; i<NoObjects; i++)
    {
        Dimers[i].clear();
    }
}

void Dimer::Report()
{
    int state[MAXSTATES];
    for (int i=0;i<MAXSTATES;i++) state[i]=0;

    for (int i=0; i<NoObjects; i++)
    {
        state[Dimers[i].State]++; 
    }

    DATALOG4 << "States: ";
    
    for (int i=0; i<MAXSTATES; i++)
    {
        DATALOG4 << i << " " << state[i] << ", "; 
    }
    DATALOG4 << "\n";
}

/*
 *  Creates random bind strengths, for the purposes of testing 
 *  lots of different values.  (more pretentiously, "randomly
 *  sampling the binding-strength space").
 */

void Dimer::InitRandomBindStrengths()
{
/*
*/
}


