#include "mathftn.h"


// *** Calculate Diffusion Coefficient ***
// (Based on P.W.Atkins, "Physical Chemistry", Oxford Uni. Press, 5th edn, pp C26
//
// Arbitrarily approximates things as "oblate ellipsoids" or "prolate ellipsoids"...
//  a bit rough and ready but, we hope, good enough...
//
//  axes are (*cough*) determined by reversing the m.o.i. equation for an ellipsoids
//  ( M.O.I. = 1/5 * (a*a + b*b) * Mass ; a & b are orthogonal axes to M.O.I. axis )
//  -> first finding one of the two, (assumed!) identical minor axes, and then
//  the major axis.
//
float MathFtn::CalculateDiffusion(pt MOI, float Viscosity, float Mass, float Temperature)
{
    float Diffusion = -1;
    float frictional_const = -1;

    //cout << "CD: MOI = " << MOI << "\n";

    int   count = 0;
    double f = 0;

    if (MOI.x!=MOI.y || MOI.x != MOI.z)
    {
        // determine if oblate or prolate, and get the major axis and its value.

        double AvMOI = (MOI.x + MOI.y + MOI.z) / 3;
        double largest = MOI.x;
        double smallest = MOI.x;

        if (MOI.x > AvMOI) count++;  // this is just to work out
        if (MOI.y > AvMOI) count++;  // whether it is oblate or
        if (MOI.z > AvMOI) count++;  // prolate.

        if (largest < MOI.y) { largest = MOI.y; }
        if (largest < MOI.z) { largest = MOI.z; }

        if (smallest > MOI.y) { smallest = MOI.y; }
        if (smallest > MOI.z) { smallest = MOI.z; }

        if (count == 2) // ( = Prolate Ellipsoid)
        {
           double b = sqrt(smallest*2.5/Mass);    // estimate of (a) minor semi-axis
           double a = sqrt((largest*5/Mass)-b*b); // estimate of (the) major semi-axis

           f = pow((a*b*b),0.33333);
           f = f * sqrt(1-((b*b)/(a*a))) / (pow(b/a,0.66666));
           f = f / log((1+sqrt(1-(b*b)/(a*a)))/(b/a));
           //cout << " effective prolate radius = " << f << "\n";
        }
        else            // ( = Oblate Ellipsoid)
        {
           double b = sqrt(smallest*2.5/Mass);     // estimate of (a) minor semi-axis
           double a = sqrt((largest*5/Mass)-b*b); // estimate of (the) major semi-axis

           f = pow((a*a*b),0.33333);
           f = f * sqrt(((a*a)/(b*b))-1);
           f = f / (pow(a/b,0.66666));
           f = f / atan(sqrt(((a*a)/(b*b))-1));
           //cout << " effective oblate radius = " << f << "\n";
        }
    }
    else // we have a perfect sphere
    {
       f = sqrt(MOI.x*2.5/Mass); // recover the radius of the sphere
    }
    frictional_const = 6 * PI * Viscosity * f * 1e-9; 

    // Stokes' Einstein equation
    Diffusion = BOLTZMANN_CONST * Temperature / frictional_const;
    //cout << " frictional coefficient = " << frictional_const << "\n";
    //cout << "Calculated Diffusion Const = " << Diffusion << "\n";

    if (isnan(Diffusion))
    {
        cout << "ERROR calculating diffusion: " << MOI << "\n";
        cout << " count = " << count << "\n";
        cout << "Mass: " << Mass << " visc: " << Viscosity << " f = " << f << "\n";
        cout << " BOLTZ " << BOLTZMANN_CONST << " frict : " << frictional_const << "\n";
        Diffusion = 0;
    }

    return Diffusion;
} // end Diffusion calculation
   

float MathFtn::CalculateDiffusion(float Radius, float Viscosity, float Temperature)
{
    // Stokes' eqn, unmodified.
    float frictional_const = 6 * PI * Viscosity * Radius * 1e-9;

    // Stokes' Einstein equation
    float Diffusion = BOLTZMANN_CONST * Temperature / frictional_const;
    //cout << " frictional coefficient = " << frictional_const << "\n";
    //cout << "Calculated Diffusion Const = " << Diffusion << "\n";

    return Diffusion;
}
