#ifndef MATRIX
#define MATRIX

#include "pt.h"

/*
 *   An extremally NON-GENERIC matrix class!
 *
 *   3*3 matrices only, for 3D vector manipulation
 *
 *   CBMatrices are aware of the Identity Matrix, and will take
 *   shortcuts if a Matrix is the Identity in order to save
 *   computation.
 *
 *   The Matrix format is:
 *
 *   [a b c]
 *   [d e f]
 *   [g h i]
 */

class CBMatrix
{

public:

    double a, b, c, d, e, f, g, h, i;

    CBMatrix()
    {a=1;b=0;c=0;d=0;e=1;f=0;g=0;h=0;i=1;is_identity=TRUE;}

    int is_identity;

    CBMatrix(double* M)
    {
        a = M[0]; b = M[1]; c = M[2];
        d = M[3]; e = M[4]; f = M[5];
        g = M[6]; h = M[7]; i = M[8];
        is_identity = (a==1&&b==0&&c==0&&d==0&&e==1&&f==0&&g==0&&h==0&&i==1)?TRUE:FALSE; 
    }

    CBMatrix(double** M)
    {
        a = M[0][0]; b = M[0][1]; c = M[0][2];
        d = M[1][0]; e = M[1][1]; f = M[1][2];
        g = M[2][0]; h = M[2][1]; i = M[2][2];
        is_identity = (a==1&&b==0&&c==0&&d==0&&e==1&&f==0&&g==0&&h==0&&i==1)?TRUE:FALSE; 
    }


    CBMatrix(double A1, double A2, double A3, double B1, double B2,
           double B3, double C1, double C2, double C3)
    {
        a = A1; b = A2; c = A3; 
        d = B1; e = B2; f = B3;
        g = C1; h = C2; i = C3;
        is_identity = (a==1&&b==0&&c==0&&d==0&&e==1&&f==0&&
                       g==0&&h==0&&i==1)?TRUE:FALSE; 
    }

    CBMatrix(pt X, pt Y, pt Z)   // X, Y, Z (may be) orthogonal basis vectors
    {
       a = X.x; b = Y.x; c = Z.x; 
       d = X.y; e = Y.y; f = Z.y;
       g = X.z; h = Y.z; i = Z.z;
       is_identity = (a==1&&b==0&&c==0&&d==0&&e==1&&f==0&&
                      g==0&&h==0&&i==1)?TRUE:FALSE;
  
    }

    pt operator* (pt p) 
    {
        if (is_identity) return p;
        if (p.x==0&&p.y==0) return pt(c*p.z,f*p.z,i*p.z);
        if (p.x==0&&p.z==0) return pt(b*p.y,e*p.y,h*p.y);
        if (p.y==0&&p.z==0) return pt(a*p.x,d*p.x,g*p.x);
        return pt((a*p.x+b*p.y+c*p.z),
                  (d*p.x+e*p.y+f*p.z),
                  (g*p.x+h*p.y+i*p.z));
    }
    
    pt operator* (pt *p)
    {
        return (*this*(*p));   // is that enough "*"s? :-)
    }
       
    CBMatrix operator= (CBMatrix M)
    {
         a=M.a;b=M.b;c=M.c;
         d=M.d;e=M.e;f=M.f;
         g=M.g;h=M.h;i=M.i;

         is_identity=M.is_identity;

         return *this;
    }
   
    CBMatrix operator* (CBMatrix M)
    {
        if (is_identity==TRUE) return M;
        if (M.is_identity==TRUE) return *this;
        return CBMatrix
        (a*M.a+b*M.d+c*M.g,  a*M.b+b*M.e+c*M.h,  a*M.c+b*M.f+c*M.i,
         d*M.a+e*M.d+f*M.g,  d*M.b+e*M.e+f*M.h,  d*M.c+e*M.f+f*M.i,
         g*M.a+h*M.d+i*M.g,  g*M.b+h*M.e+i*M.h,  g*M.c+h*M.f+i*M.i);
    } 
     
    CBMatrix operator* (CBMatrix *M) {return (*this * (*M));}

    void SetX(pt X) {a = X.x; d = X.y; g = X.z;}   // set X basis vector
    void SetY(pt Y) {b = Y.x; e = Y.y; h = Y.z;}   // set Y basis vector
    void SetZ(pt Z) {c = Z.x; f = Z.y; i = Z.z;}   // set Z basis vector

    pt X() {return pt(a,d,g);}
    pt Y() {return pt(b,e,h);}
    pt Z() {return pt(c,f,i);}

    void Clear() {a=0;b=0;c=0;d=0;e=0;f=0;g=0;h=0;i=0;is_identity=false;}

    int CheckIdentity()
    {
       is_identity = (a==1&&b==0&&c==0&&d==0&&e==1&&f==0&&
                      g==0&&h==0&&i==1)?TRUE:FALSE;
       return (is_identity);
    } 

    double det() {return (a*(e*i-f*h)-b*(d*i-f*g)+c*(d*h-e*g));}

    CBMatrix normalise()
    {
        pt x = ~X();
        pt y = ~Y();
        pt z = ~Z();
        a=x.x;b=y.x;c=z.x;d=x.y;e=y.y;f=z.y;g=x.z;h=y.z;i=z.z;
        return (*this);
    }

    CBMatrix inv()
    {
        double D = det(); 
        if (D==0)
        {
            cout << "Error - matrix has no inverse\n";
            return Identity;
        }
        return CBMatrix((e*i-f*h)/D, (c*h-b*i)/D, (b*f-c*e)/D,
                        (f*g-d*i)/D, (a*i-c*g)/D, (c*d-a*f)/D,
                        (d*h-e*g)/D, (b*g-a*h)/D, (a*e-b*d)/D); 
    }

    static CBMatrix Identity;   // The identity matrix, always useful... :-)

    static CBMatrix RotationMatrix(pt Axis, double radians);
 
    friend ostream& operator << (ostream &o, CBMatrix &M);
    friend ostream& operator << (ostream &o, CBMatrix *M);
};

ostream& operator << (ostream &o, CBMatrix &M);
ostream& operator << (ostream &o, CBMatrix *M);

#endif 
