#include "microtubule.h"
#include "assert.h"
#include "mathftn.h"

int DebugRotation = FALSE;

#define DEBUGMICRO FALSE 
//#define NOROTATION                     // cancels rotational movement
//#define NOMOVEMENT                     // stops all movement

Microtubule* Microtubule::First = NULL;
Microtubule* Microtubule::Last  = NULL;

int   Microtubule::NoMicroTs = 0;

long  Microtubule::turbidity;
long  Microtubule::count;
long  Microtubule::totalmass;
float Microtubule::avweight;


ostream& operator << (ostream &o, Microtubule *T)
{
   o<<"*****\nMicrotubule " << T->TagNo << "\n"
     << "    Centroid " << T->Centroid << "\n"
     << "    M.O.I. " << T->LocalMomentOfInertia <<"\n   no dimers: " 
     << T->NoObjects << "\n    Mass = " << T->Mass << "    \n";

   o<<"dimers:\n";
   for (int i=0; i<T->NoObjects; i++)
   {
      Dimer* D = T->DimerPtArray[i];
      if (D!=NULL) 
      {
          o << setw(4)<<i<<": (" << setw(4) << D->TagNo<<") ";
          for (int j=0; j<D->NoSites; j++)
          {
              if (D->LinkNo[j].to != NULL)
                  o << setw(5) << D->LinkNo[j].to->TagNo; 
              else
                  o << " -- ";
          }

          o << " |\n  "<< setprecision(6);
          for (int j=0; j<D->NoSites; j++)
          {
              o << D->LinkNo[j].true_pos << "  "; 
          }
      }
      else 
          o << i<<":-1";
//      o << "        " << D->center << "\n";
      o << "\n";

   }  
   o << "\n    Back " << (T->Back==NULL?-1:T->Back->TagNo) << "\n";
   o << "    Next " << (T->Next==NULL?-1:T->Next->TagNo) << "\n\n";
   return o;   
}

ostream& operator << (ostream &o, Microtubule &T)
{
    o << &T;
    return o;
}


/*
 *  Microtubule Class
 *
 *  - handles aggregations of dimers.  These come and go very rapidly,
 *  hence all the arrays and stuff have to be handled carefully, ditto
 *  for cleaning up.
 *
 *  The dimer array is a little special:
 *
 *   THE DIMER ARRAY:
 *       CurrentArraySize : the size of the array in use -> new dimers will
 *                          be added at this location
 *       NoObjects         : Actual real number of dimers in the micro tubule
 *
 *       => These two are *not* usually identical. To save time, when we
 *       delete a dimer, its location is just set to null - we don't shrink the
 *       array down.  When the array has reached its full size, a new 
 *       array is created, and *it* is shrunk.  This may lead to a lot
 *       of unneccessary array creation, and may need to be changed...
 *
 */

Microtubule::Microtubule(Dimer *A)  // Dimer provides rotational data - no more.
{
    Back = Microtubule::Last;
    if (Microtubule::Last == NULL)  
        Microtubule::First = this;    //empty list - initialise
    else
        Microtubule::Last->Next = this;
    Microtubule::Last = this; 
    Next = NULL;

    //CurrentArraySize = 0;
    velocity = pt(0);

    NoObjects = 0;
    Mass = 0;
    rms_dp = 0;
    ArraySizeMax = 8;
    DimerPtArray = new (Dimer *[ArraySizeMax]);
    assert(DimerPtArray != NULL);
    
    for (int i=0; i<ArraySizeMax; i++) DimerPtArray[i] = NULL;

    TagNo = ++NoMicroTs;
    if (!TheProgram.silent)
        ERRORLOG << "Creating New Microtubule " << this;

    InitRotationalData(A);
}


    /*
     *  Dimers are treated as constant density cylinders for the 
     *  purpose of calculating the Moment Of Inertia (M.O.I).
     *  -> treated as 40kD, 8nm (major axis) * 4nm (diameter)
     *
     *  In general, if cylinder is radius r and length l,
     *  taking z as major axis, Iz = (1/2) * r*r*Mass
     *  Ix = Iy = (1/4) *r*r*Mass + (1/12) * l*l*Mass
     *
     *  we calc. microtubule M.O.I. using parallel axis theorem.
     *
     *  This ftn however simply initialises are variables.  The
     *  real work is done by AddMass()
     *
     *  CONSIDER: Changing this to generic sphere calculations.
     */

void Microtubule::InitRotationalData(Dimer *D)
{
    Centroid = pt(0);          // These will be set when the initial dimers are 

    // To avoid recalculating M.O.I. every time mt. rotates, we use 
    // a local reference frame.  Local Z is in the +ve protofilament dirtn.
    // arbitrarily we set the basis to that of the first dimer. 

    Basis = D->Basis;
    InvBasis = Basis.inv();

    LocalMomentOfInertia = pt(0);  
}


/*
 *  Delete the microtubule, release all the dimers, 
 *  and free up the array space
 */

Microtubule::~Microtubule()
{

    if (TheProgram.verbose)
    {
        cout << "Awooga - Destroying Microtubule " << TagNo << "\n"; 
        for (int i=0; i<NoObjects; i++)
        {
           if (DimerPtArray[i]!=NULL) 
               cout << "  "<<i<<": " << DimerPtArray[i]->TagNo;
           else
              cout << "  "<<i<<": (NULL)";
        }
        cout << "\n";
    }

    if (Back!=NULL) 
        Back->Next = Next;
    else
    {
        assert(this == Microtubule::First); // ERROR in mt. linked list : Bad First Pointer
        Microtubule::First = Next;   //Which = Null, if this is the last mt
    }

    if (Next!=NULL) 
        Next->Back = Back;
    else 
    {
        assert (this == Microtubule::Last); 

        Microtubule::Last = Back;   //Which = Null, if this is the last mt
    }

    /*
     *   By setting all the "Owner" pointers to null, we avoid any
     *   problems with the objects being removed recursively calling
     *   this destructor. 
     */

    for (int i=0; i<NoObjects; i++)
    {
        if (DimerPtArray[i]!=NULL)
            DimerPtArray[i]->FreeBySelf(); 
        DimerPtArray[i] = NULL;
    }

    delete [] DimerPtArray;

    if (!TheProgram.silent)
        ERRORLOG << "It's dead Jim\n";
}


/*
 *  This static function creates a new microtubule from the first
 *  two dimers.
 */

void Microtubule::CreateNewMicrotubule(Dimer *A, Dimer *B)
{
    // TempPointer only used here - mainly, mt. is added to linked list in 
    // Static Microtubule variable.

    Microtubule* TempPointer = new Microtubule(A);

    assert (TempPointer != NULL);

    if (TempPointer->Add(A, NULL) == ERROR)
        {cerr << "error adding A to mt \n";}   // very rare
    else
        if (TempPointer->Add(B, A) == ERROR)
        {
            cerr << "error B in mt add - deleting m.t.\n";
            cerr << A << " && " << B <<"\n";
            delete TempPointer;
        }
     else
     {
         assert( TempPointer->DimerPtArray[0] == A );
         assert( TempPointer->DimerPtArray[1] == B );
     }
}


/*
 * This simply checks if a valid dimer pointer and position is being passed to 
 * the microtubule add() function, and screams if it isn't.
 */

int Microtubule::CheckDimerPos(Dimer *D)
{
    assert (D!=NULL);

    MoveDimersToCommonRegion();

    for (int i=0; i<NoObjects; i++)
    {
        Dimer* test = DimerPtArray[i];
        assert (test!=NULL);

        if (test!=D)
        {
      
            float dist_sep = Dimer::min_contact_dist_sq(D, test);
//cout << " test " << i << ": "<< D->TagNo << " + " << DimerPtArray[i]->TagNo << " = " << dist_sep << " ... " << D->center << ", " << DimerPtArray[i]->center << "\n";
            if (dist_sep < -0.01)
            {
                if (TheProgram.verbose )
                {
                    cout << "dimer overlap! "<<D << "\n in bad position - overlaps \n";
                    cout << DimerPtArray[i] << "\n by " << dist_sep << "\n"; 
                    cout << "No Objects = " << NoObjects << " i= " << i << "\n";

                }
                MoveDimersToTruePositions();
                return ERROR; 
            }
        }
    }

    MoveDimersToTruePositions();
    return OK;
}


/*
 * This handles the variable size dimer array in the mt.  
 */

int Microtubule::CheckArrayInAdd()
{
    assert (NoObjects<=ArraySizeMax);
        
    if (NoObjects==ArraySizeMax) // Create new array, copy in contents...
    {
        int i;
        int NewArraySize = ArraySizeMax*2;
   

        /*
         *  Set up new array to hold the old dimer array.
         */

        Dimer **temp = new Dimer *[NewArraySize];

        for (i=0; i<NoObjects; i++)
            temp[i] = DimerPtArray[i]; 

        for (i=NoObjects; i<NewArraySize; i++)
            temp[i] = NULL;

        /*
         *  Double the array size
         */

        for (i=0; i<ArraySizeMax; i++)
            DimerPtArray[i]=NULL;       // to prevent objects also being deleted (?)

        ArraySizeMax *= 2;              // reset ArraySizeMax object variable

        delete [] DimerPtArray; // .. being careful not to delete dimers

        DimerPtArray = temp;
    }
    return OK;
}


/*
 *  Adds a dimer to a microtubule.  Calculates new rotational data.
 */

int Microtubule::Add(Dimer *D, Dimer *local_dimer)
{
    if (TheProgram.verbose)
    {
        cout << "MT::Add("<<D->TagNo<<",";
        if (local_dimer==NULL) cout << "null)\n"; else cout << local_dimer->TagNo<<")\n";
    }

    assert(D->Owner == NULL);

#ifdef SAFE
    assert(IntegrityCheck("Add Start") == OK);  // Make sure mt is o.k. before we add more trouble.
#endif

    /* 
     * NoObjects is the number of actual dimers in the array...
     * ArraySizeMax is the size allocated to the array in memory. 
     */

    if (CheckDimerPos(D)==ERROR)
    {
        cout << " Dimer " <<D->TagNo << "in impossible position in mt add. Aborting add.\n";
        Dimer::UnLink(D, local_dimer);
        return ERROR;
    }

    if (TheProgram.verbose)
        cout << "dimer " << D->TagNo << " past CheckPos in MT::Add()\n";

    if (AddDimerToArray(D)==ERROR) 
    {
        cerr << "unable to add dimer " << D->TagNo << " to array in mt: " << TagNo << " - unusual\n"; 
        Dimer::UnLink(D, local_dimer);
        return ERROR;
    }

    AddMass(D);
    D->SetOwner(this);

//TEMP
    assert (CheckDimerPos(D)!=ERROR);
    
    if (TheProgram.verbose)
        cout << "dimer " << D->TagNo << " about to check cross links in MT::Add()\n";

    if (CheckCrossLinks(D)==ERROR)
    {
       cerr << "unable to add dimer: " << D->TagNo << " due to link error\n"; 
       DeleteDimer(D);
       return ERROR;
    }

    if (TheProgram.verbose) 
        cout <<"Added dimer "<<D->TagNo<<" to microtubule "<<TagNo<<"\n";

#ifdef SAFE
    assert(IntegrityCheck("Add End") == OK);  // Make sure mt is still o.k. .
#endif
 
    return OK;
}

int Microtubule::AddDimerToArray(Dimer *D)
{
    if (CheckArrayInAdd() == ERROR) return ERROR;
    DimerPtArray[NoObjects] = D;
    NoObjects++;
    return OK;
}

/*
 *   This makes sure that all the dimers in a microtubule are correctly
 *   cross linked.  For example, when a new dimer "slots into" a gap
 *   in a microtubule, it needs to be linked with *all* of its neighbours.
 *
 *   This is only called when adding a new dimer, 'D'.  But for the moment
 *   we check *all* links as well, to reassure ourselves of the soundness of the MT.
 */

int Microtubule::CheckCrossLinks(Dimer *D)
{
    assert (D!=NULL);
    assert (CheckLinks() != ERROR);
    assert (IntegrityCheck("Check Cross Links ") != ERROR);

    Dimer *A, *B;
    int i,j;

    MoveDimersToCommonRegion();

    for (i=0; i<NoObjects; i++) 
    {
        A = DimerPtArray[i];
        assert(A != NULL);

        for (j=i+1; j<NoObjects; j++) 
        {
            B = DimerPtArray[j];
            assert (B!=NULL); 
            assert (A != B);

            float dist = pt::dist(A->center,B->center);

            if ( dist < (A->BoundRad+B->BoundRad) )
            {
                if (CheckLinkBetween(A,B) == ERROR)  // Check Link actually links,
                {                                     // if appropriate
                    if (A==D || B==D) 
                    {
                        cout << "checkcrosslinks bad for " << D->TagNo << "\n";
                    }
                    else              
                    {
                        cerr << "existing error in mt: " << TagNo << " - destroying \n";
                        delete(this); 
                        return ERROR;
                    } 
                    MoveDimersToTruePositions();
                    return ERROR;
                }
            }
        }
    }

    MoveDimersToTruePositions();
    return OK;
}

/*
 *  Checks whether the microtubule's links are correctly aligned
 */

int Microtubule::CheckLinks()
{
    for (int i=0; i<NoObjects; i++)
    {
        Dimer *D = DimerPtArray[i]; 
        for (int j=0; j<D->NoSites; j++)
        {
             if (D->LinkNo[j].to != NULL)
             {
                 Link* LinkA = &(D->LinkNo[j]);            // forward link
                 Link* LinkB = LinkA->to->FindLink(D);  // get back link 
                 if (LinkB==NULL) return ERROR;
                 if (CheckLinkPair(LinkA, LinkB)==ERROR) return ERROR; 
             }
        }
    }
    return OK;
}

int Microtubule::CheckLinkPair(Link A, Link B)
{
    // check two links are sane...
    // assert (TheProgram.EqualWithWrap(A.true_pos, B.true_pos));

    if (TheProgram.EqualWithWrap(A.true_pos, B.true_pos)==false)
    {
        cout << " Equal With Wrap False: " << A << "\n" << B << "\n";
//        exit(-1);
        return ERROR;
    }
    return OK;
}

/*
 *   Checks whether A and B should be cross-linked within a microtubule 
 *   -> and links if appropriate
 */

int Microtubule::CheckLinkBetween(Dimer *A, Dimer *B)
{
    int i,j;

    assert (CheckDimerPos(A) != ERROR);
    assert (CheckDimerPos(B) != ERROR);

    for (i=0; i<A->NoSites; i++)
    {
         if (A->LinkNo[i].to == NULL)
         {
             for (j=0; j<B->NoSites; j++)
             {
                  if (TheProgram.EqualWithWrap(A->LinkNo[i].true_pos,B->LinkNo[j].true_pos))
                  {
                      if (B->LinkNo[j].to==NULL) 
                      {
                          if (A->LinkNo[i].CanBind(&(B->LinkNo[j])))
                          {
                              if (TheProgram.verbose)
                                  cout << " putting extra links between "<< A->TagNo <<" and "<<B->TagNo << "\n"; 
                              assert( A->LinkNo[i].CanBind(&(B->LinkNo[j])) == TRUE );
                              Dimer::LinkUp(A, B, &A->LinkNo[i], &B->LinkNo[j]);
                          }
                          else
                          {
/*                              cerr << " MT  possible structural fault - overlapping links:\n";
                              cerr << A->TagNo << ": (" << A->Model->Name << ") " << A->LinkNo[i].Type->Name;
                              cerr << "\n   at same pos as\n ";
                              cerr << B->TagNo << ": (" << B->Model->Name << ") " << B->LinkNo[j].Type->Name<<"\n";
*/
                              float overlap = Dimer::min_contact_dist_sq(A,B);
                              if (overlap < (-GROSSERROR))
                              {
                                  cerr << "*** SERIOUS ERROR ***\n  Dimers overlapping.\n";
                                  cerr << "centers: " << A->center << " overlaps " << B->center;
                                  cerr << " by " << overlap << "\n";
                                  cerr << "   ... attempting to jettison dimer " << B->TagNo<<"\n";
                                  B->free();
                                  B->clear();
                                  IntegrityCheck("Dimer jettison in CheckLinkBetween()"); 
                              }

                          }
                      }
                      else
                      {
                          cerr <<" Poss MT structural fault: Dimer " << A->TagNo;
                          cerr <<" occupying same pos as Dimer "<<B->LinkNo[j].to->TagNo<<"\n";
                          cerr <<" A = " << A->center << "\n B = " << B->center << "\n";
                          return ERROR;
                      }

                      return OK;  // BREAK OUT OF LOOP - Job done.
                  }
             } 
         }
    } 
    return OK;
}


/*
 *   This recalculates data when a new dimer is added.  For robustness
 *   this currently recalculates the entire mt from scratch.  This ftn.
 *   is identical to DeleteMass, at the moment.
 *
 *   This could be sped up considerably.
 */

void Microtubule::AddMass(Dimer *D)
{
    // SPECIAL INTERATION WITH STATIONARY OBJECTS:
    // if (D->rms_dp == 0 || (Mass>0&&rms_dp==0)) 
    //   rms_dp = 0;

    RecalculateMassDetails();
}

// sum masses of components.  If all components are the same, this would be
// simply NoObjects * ...->Model->Mass, but we can't be certain.

float Microtubule::CalculateMass()
{
    float Mass = 0;
    for (int i=0; i<NoObjects; i++)
    {
        assert (isnan(DimerPtArray[i]->Model->Mass) == FALSE);
        Mass += DimerPtArray[i]->Model->Mass;
    }
//if (NoObjects>3) cout << "MT: " << TagNo << " size " << NoObjects << " Mass = " << Mass << "\n";

    return Mass;
}

pt Microtubule::CalculateCentroid(float Mass)
{
    // this is very slow.  So life sucks.

    /*  starting test pt for horrible offset stuff...
     *  we choose this test pt to 'pin' the mt on a 
     *  particular side of the simulation field. 
     */

    assert (DimerPtArray[0] != NULL);
    pt test = DimerPtArray[0]->center; 

    assert (isnan(test.x) == FALSE);

    pt Centroid = pt(0,0,0); 
    Dimer *D;
    for (int i=0; i<NoObjects; i++)
    {
       D = DimerPtArray[i];    

       pt diff = D->center;
       diff += TheProgram.BoundsOffset(diff - test);

       assert (isnan(diff.x) == FALSE);       

       Centroid += diff * D->Model->Mass;
    }

    // The New centre of mass
    Centroid /= Mass;

    pt offset = TheProgram.BoundsOffset(Centroid);
    if (offset != 0)
    {
        Centroid += offset;
    }
//if (NoObjects>3) cout << "MT: " << TagNo << " size " << NoObjects << " Centroid = " << Centroid << "\n";

    return Centroid;    
}

pt Microtubule::CalculateMOI(pt Centroid)
{
    pt MOI = pt(0,0,0); 

    Dimer *D;
    for (int i=0; i<NoObjects; i++)
    {
       D = DimerPtArray[i];

       pt diff = D->center - Centroid;    

       pt offset = TheProgram.BoundsOffset(diff);

       if (offset!=0) diff += offset;

       // change diff to local MT co-ordinates (math and debugging convenience)

       diff = InvBasis*diff; 

       assert (isnan(diff.x) == FALSE);

       // get square of co-ordinate distances -> (deltaX^2,deltaY^2,deltaZ^2)

       float diff_sq = pt::sqr(diff);
       diff = diff.element_mult(diff);
       pt diff_coord_sq = pt(diff_sq,diff_sq,diff_sq) - diff;

       MOI += D->LocalMomentOfInertia + diff_coord_sq * D->Model->Mass; 

   } 


//if (NoObjects>3) cout << "MT: " << TagNo << " size " << NoObjects << " MOI = " << MOI << "\n";

   assert(isnan(MOI.x)==FALSE);

   return MOI;
}

// calculate the standard deviation of movement, based on simple minded
// summation of component object standard deviations...

float Microtubule::CalculateRMS_DP(pt MOI)
{
    float Viscosity = TheProgram.Viscosity;
    float Temperature = TheProgram.Temperature;
    float TimeStep    = TheProgram.TimeScale;

    assert(isnan(Viscosity)==FALSE);
    assert(isnan(Temperature)==FALSE);
    assert(isnan(TimeStep)==FALSE);
    assert(isnan(MOI.x)==FALSE);


    float Diffusion = MathFtn::CalculateDiffusion(MOI, Viscosity, Mass, Temperature);
//cout << " Diff: " << Diffusion << " rand nums: " << MOI << ","<<Viscosity<<","<<Mass<<","<<Temperature<<"\n";
    assert(isnan(Diffusion)==FALSE);

    float rms_dp = sqrt(2*Diffusion*TimeStep);
    assert(isnan(rms_dp)==FALSE);

//if (NoObjects>3) cout << "MT: " << TagNo << " size " << NoObjects << " rms_dp = " << rms_dp << "\n";

    return rms_dp;
} 

/*
 *  Delete a Dimer's mass, and update centroid and M.O.I.
 *
 *  TODO: update for generic masses and shapes
 *
 *  Hmmm... can we just add a negative mass to AddMass() instead?
 */

void Microtubule::DeleteMass(Dimer *D)
{
    RecalculateMassDetails();
}

void Microtubule::RecalculateMassDetails()
{
    // must be done in order

    Mass                      = CalculateMass();
    Centroid                  = CalculateCentroid(Mass);
    LocalMomentOfInertia      = CalculateMOI(Centroid);
    rms_dp                    = CalculateRMS_DP(LocalMomentOfInertia) * 1e9;
}

/* This is for when we dissassemble a mt and give all its
 * dimers to another mt.
 */

void Microtubule::StripDimerFromOwner(Dimer *D)
{
    assert (D != NULL);

    int i;
    for (i=ArraySizeMax-1; i>=0; i--)
    {
        if (DimerPtArray[i]==D)
        {
            DimerPtArray[i]=NULL;  
            D->SetOwner(NULL); 
            NoObjects--;           
            break; 
        }
    }

    if (i<0)
    {
         cerr << "Unable to find and Strip Dimer "<< D->TagNo << " from mt ";
         cerr << TagNo<<"\n";
    }

}


/*
 *  Take a dimer from a mt.
 *
 *  nb - links are handled independently, as they are attributes of dimers.
 *       -> think of 'Microtubules' as just being clever bags to hold dimers in.
 */

void Microtubule::DeleteDimer(Dimer *D)
{
    assert (D!=NULL);
    assert (D->Owner == this); 

//    IntegrityCheck("Delete Dimer Start");

    if (TheProgram.verbose) 
    {
        cout << "deleting dimer "<<D->TagNo<<" from microtubule " << TagNo; 
        for (int i=0; i<NoObjects; i++)
        {
           if (DimerPtArray[i]!=NULL) 
               cout << "  "<<i<<": " << DimerPtArray[i]->TagNo;
           else
              cout << "  "<<i<<": (NULL)";
        }
        cout << "\n";
    }

    if (RemoveDimerFromArray(D)==ERROR)
    {
         cerr << "Unable to RemoveDimerFromArray "<<D->TagNo<<" mt: "<<TagNo<<"\n" ;
         cerr << "MT = " << this << "\n";

         delete(this);
         return;
    } 

    D->SetOwner(NULL);   // the dimer is now officially free (as long as its bonds
                         // have been seperately broken).

    if (CheckMT())       //If Microtubule has less than two dimers...
    {
        delete(this);    // ... delete this microtubule.       
    }
    else
    {
        DeleteMass(D);  
        BreakCrossLinks(D);  // almost always these have already been broken...
    }


    assert (D->Owner == NULL);
    if (D->GetBound() == TRUE)
    {
        cout << "bastard dimer: " << D->TagNo << " still hanging on!\n";
        exit(-1); 
    }

    if (TheProgram.verbose) 
        cout << "Dimer: " << D->TagNo << " deleted from Microtubule " << TagNo << "\n"; 

//    IntegrityCheck("Delete Dimer End");
}

/*
 *  BreakCrossLinks takes a dimer that has just been removed from an
 *  aggregate, and makes sure that it has no remaining linkages to
 *  the aggregate.
 */

void Microtubule::BreakCrossLinks(Dimer *D)
{
    for (int i=0; i<NoObjects; i++)
    {
        Dimer *A = DimerPtArray[i]; 
        assert (A != NULL);
        assert (A != D);

        for (int j=0; j<A->NoSites; j++)
        {
            if ( A->LinkNo[j].to == D )
            {
                A->LinkNo[j].to = NULL;
                cout << "BreakCrossLinks Breaking Link from " << A->TagNo;
                cout << "to " << D->TagNo << " in mt: " << TagNo << " \n";
            }
        }
    }
}

int Microtubule::RemoveDimerFromArray(Dimer *D)
{
    //try from most recent to least recent -> should catch them
    //quickly, since recently added ones are more likely to break away...

    assert (D!=NULL);

    int i;

    for (i=NoObjects-1; i>=0; i--)
    {
        if (DimerPtArray[i]==D)
        {
            // reshuffle
            for (int j=i; j<NoObjects-1; j++)
                DimerPtArray[j] = DimerPtArray[j+1];

            NoObjects--;     
            DimerPtArray[NoObjects] = NULL;

            break; 
        }
    }

    if (i<0)
    {
        cerr << "can't find dimer " << D->TagNo << "in mt: " << TagNo << "\n";
        return FALSE;
    }
    else
    {
        return OK;
    }
} 

/*
 *  See if mt is still viable
 */

int Microtubule::CheckMT()
{
    return (NoObjects == 1);  //Microtubules with one dimer deserve death
                             //(Microtubules with zero dimers are already dying)
}


/*
 *  Checks whether the microtubule is sane, and whether all
 *  its bound dimers really belong to it.  In an ideal world,
 *  this is all unneccessary, and it should never flag nothin'...
 */

int Microtubule::IntegrityCheck()
{
    return Microtubule::IntegrityCheck("null msg ");
}

int Microtubule::IntegrityCheck(char *Msg)
{
#ifdef SAFE
    Dimer *D;

    if (NoObjects <= 1) return OK;  // the mt. is being destroyed (we hope!)


    /*
     *  Check Each Dimer for Gross Errors
     */
    for (int i=0; i< NoObjects; i++)
    {
        D = DimerPtArray[i];
        if (D!=NULL)
        {
            // Check that dimer is owned.
            if (D->Owner != this)
            {
                cerr <<Msg<<"ERROR mt int. check.: "<<TagNo<<" does not own "<<D<<"\n";
                return ERROR;
            }

            // info only.  This actually happens (but rarely).
            if (D->GetBound()== FALSE && NoObjects != 1) 
            {
                if (TheProgram.verbose)
                    cout<<Msg<<" INFO mt int. check.: orphan "<<D<<" - pos due to internal escape ";
            }

            // Check Dimer Links.
            if (D->CheckLinks()==ERROR)                 // Check Orientation
                return ERROR;

            // Check Dimer Position.
            if (CheckDimerPos(D)==ERROR)
            {
                cerr<<Msg<<" ERROR mt int. check. bad dimer pos for " << D << "\n";
                return ERROR;
            }

            for (int j=i+1; j < NoObjects; j++) 
            {
                if (DimerPtArray[j]==D) 
                {
                    cerr<<Msg<<" ERROR mt int. check. dimer " << D->TagNo << " entered twice\n";
                    return ERROR;
                }
            }
        }
        else
        {
            cerr << Msg << " Null hole in dimer array for mt " << TagNo << "\n" << this; 
            return ERROR;
        }
    }
    
    // Check no stray dimers in array...

    for (int i=NoObjects; i<ArraySizeMax; i++)
    {
        if (DimerPtArray[i]!=NULL)
	{
            cerr << Msg << "ERROR -> Dimer existant in array above noObjects limit\n";
            return ERROR;  
        }
    }

    // Check Links O.K.

    if (CheckLinks()==ERROR) 
        return ERROR;

#endif
    return OK;
}


/*
 *   static ftn:  go through all the aggregates, moving them appropriately
 */

void Microtubule::MoveAll()
{
    if (TheProgram.freeze_mts) return;

    Microtubule* Iterator = First;

    while (Iterator != NULL)
    {
        Iterator->Move();
        Iterator=Iterator->Next;
    }

}


/*
 *   static ftn:  run an integrity check on *all* mts.  time consuming.
 */

void Microtubule::CheckAll(char *Msg)
{
    Microtubule* Iterator = First;

    while (Iterator != NULL)
    {
        Microtubule* temp = Iterator;
        Iterator=Iterator->Next;
        if (temp->IntegrityCheck(Msg)==ERROR)
        {
            cerr << "*** ERROR ***\nBad Aggregate ("<<temp->TagNo<<") found.  Deleting.\n";
            delete(temp);
        }
    }
}

/*
 *    move the individual aggregates
 */

int Microtubule::Move()
{
    //Model the 1st - simple gaussian motion for rigid MT
/*    
    velocity = RAND_DU_VECT/sqrt((float)NoObjects);

    for (int i=0; i< CurrentArraySize; i++)
    {
        if (DimerPtArray[i]!=NULL)
        {
            DimerPtArray[i]->bump_bound_dimer();
        }
    } 
*/
    //Model the 2nd - stokes dodad for motion, rotation via
    //moments of inertia and a stab at the equipartition principle.

    //TODO: Work out Rotation Properly

    // AHEM: this is a "heuristic" hack... :-) 
    // (rotation, as long as it occurs a bit, shouldn't grossly affect the
    // models accuracy...)

    // SPECIAL HACK FOR STATIONARY OBJECTS

    if (rms_dp==0) return 0;

    assert(isnan(rms_dp)==FALSE);

    velocity = TheProgram.RandomNormalPt() * rms_dp ;

    assert(isnan(velocity.x)==FALSE);

    pt MOIinv = LocalMomentOfInertia.inv(); 
    pt MagAngVel = pt(sqrt(MOIinv.x),sqrt(MOIinv.y),sqrt(MOIinv.z));
    MagAngVel = MagAngVel.element_mult(RAND_U_VECT) * 50;


    // Fools rush in... add angular velocity vectors
    // -> and convert to global co-ordinates

    pt AngVel = Basis*MagAngVel;    

    if (DebugRotation) 
    {
        cout << " MOIinv " << MOIinv << "  MOI: " << LocalMomentOfInertia << "\n";
        cout << "X ang vel: " << new pt(Basis.X())<<" * " <<MagAngVel.x <<"\n";
        cout << "Y ang vel: " << new pt(Basis.Y())<<" * " <<MagAngVel.y <<"\n";
        cout << "Z ang vel: " << new pt(Basis.Z())<<" * " <<MagAngVel.z <<"\n";
        cout << "Ang Vel:   " << AngVel << "\n\n";
    }

    AngVel = ~AngVel;
    if (isnan(AngVel.x)==FALSE)  // small but finite chance AngVel is undefined... 
        Rotate(AngVel, pt::abs(AngVel));        

    assert(isnan(velocity.x)==FALSE);

    MoveLinearly(velocity);

    return 0;
}

void Microtubule::MoveLinearly(pt velocity)
{
    assert (isnan(velocity.x)==FALSE);

    for (int i=0; i<NoObjects; i++)
    {
        assert (DimerPtArray[i]!=NULL);
        DimerPtArray[i]->add(velocity);
    }
    Centroid += velocity;
    Centroid += TheProgram.BoundsOffset(Centroid);
}



/*
 *   Quicky probability of melding function.  Modify later.
 *
 *   TODO: need better generalised collision function
 */

int Microtubule::ProbMeld(Dimer *A, Dimer *B)
{
//   cout<<"Prob Meld of Agg.: "<<A->Owner->TagNo<<" & "<<B->Owner->TagNo << "\n";

   // If either of the binding sites is full, there is no
   // chance of binding at all.

   if (TheProgram.NoMerge == TRUE) return FALSE;

   if (A->GetFullyBound()||B->GetFullyBound()) return FALSE;

   pt A_dirtn = A->Basis.Z();
   pt B_dirtn = B->Basis.Z();

   float Alignment = A_dirtn^B_dirtn; 

//   cout << "Prob Meld: align = " << Alignment; 
            
   int Return_Val =  (Alignment>R1)?TRUE:FALSE;

//   cout << " ret val : " << Return_Val << "\n";

//   if (Return_Val == TRUE) cout << TheProgram.NoCyclesCompleted << " MICROTUBULES " << A->Owner->TagNo << " & " << B->Owner->TagNo << " ARE GO FOR MERGE \n";
 
   return Return_Val; 

}

 
/*
 *  If polymers collide, then how do we tell if they link up with each
 *  other?  
 *
 *  When this function is called, we have already established that a 
 *  collision has occured, and if the dimers were free would have resulted
 *  in binding.  This function (fairly arbitrarily) calls a further
 *  probabilistic ftn that must be satisfied for binding to occur, based
 *  on how closely aligned the polymers are. 
 *
 *  Finally the two microtubules are merged to create a single mt.
 *
 */

int  Microtubule::Meld(Dimer *A, Dimer *B)
{
//    cout << " Melding " << A->Owner->TagNo << " & " << B->Owner->TagNo << "\n";

    assert(A->Owner->IntegrityCheck("Start Meld") == OK);

    Link *Alink = A->FindLink(B);   // Could pass these as parameters
    Link *Blink = B->FindLink(A);   // but this seems less effort...

    if (ProbMeld(A,B)) 
    {
cout << "MELDING!!\n";
assert (Alink != NULL && Blink != NULL);
assert (Alink->to == B && Blink->to == A);
assert (A->Owner->IntegrityCheck("in Meld")==OK);

        PhysicallyJoin(A,B,Alink,Blink);       
    }
    else
    {
        Dimer::UnLink(A,B);
        return (FAIL);
    }
    return (OK);    
}

/*
 *  This rotates a microtubule in absolute co-ordinates
 *  (rather than the microtubule's local co-ordinates)
 */

void Microtubule::Rotate(pt Axis, float Angle)
{
#ifndef NOROTATION

    if (TheProgram.NoRotation == TRUE) return;

assert (isnan(Axis.x)==FALSE);
    Dimer *D;
 
    Basis = CBMatrix::RotationMatrix(Axis, Angle) * Basis;

#ifdef SAFE
//Check Basis still normalised
    float Det = Basis.det();

//DEBUG - can remove, or make rarer... (floating point error may slowly skew
// matrix from det = 1

    if (Det > 1.01 || Det < .99)
    {
        cout << "Basis vector skewy: " << Basis;
        Basis.normalise();
    } 
#endif

    InvBasis = Basis.inv();

    MoveDimersToCommonRegion();

assert (isnan(Centroid.x)==FALSE);
assert (isnan(Axis.x)==FALSE);
assert (isnan(Angle)==FALSE);

    for (int i=0; i<NoObjects; i++)
    {
        D = DimerPtArray[i];
        D->rotate_bound_dimer(Centroid, Axis, Angle);
        assert (isnan(D->center.x)==FALSE);
    }

    MoveDimersToTruePositions();

#endif
}

/*
 *  A mt may span the boundary of the simulation field, with dimers "wrapping- 
 *  around".  This complicates calculations greatly, so this ftn is intended to
 *  temporarily move all dimers to the side that the majority of dimers currently
 *  are on (found using the centroid).
 *
 *  This ftn. should be undone immediately after use using MoveDimersToTruePositions() 
 */

void Microtubule::MoveDimersToCommonRegion()
{
    pt offset(0);
    for (int i=0; i<NoObjects; i++)
    {
        Dimer* D = DimerPtArray[i];
        assert (D!=NULL);

        offset = TheProgram.BoundsOffset(D->center - Centroid);

        if (offset!=0) 
            D->FalsifyPosition(offset);
    }
}

void Microtubule::MoveDimersToTruePositions()
{
    for (int i=0; i<NoObjects; i++)
    {
        Dimer* D = DimerPtArray[i];
        D->DeFalsifyPosition();
    }
}

void Microtubule::FullPrint(int PrintAllLinks)
{
    cout << "\nFull Print Of Microtubule " << TagNo << "\n\n";
    cout << "Centroid : " << Centroid << "\n";
    cout << "Basis : " << Basis << "\n";
    if (PrintAllLinks) cout << " ... links print code to be implemented ... \n";
    PrintDimerList();
}

void Microtubule::PrintDimerList()
{
    for (int i=0; i<NoObjects; i++)
    {
assert(DimerPtArray[i]!=NULL);
            cout << "i : "<< DimerPtArray[i]->TagNo <<"              "<<DimerPtArray[i]->center<<"\n";
            //DimerPtArray[i]->printlinks(PrintAllLinks);
    }
}


/*
 *   Report on the state of all microtubules...
 */

void Microtubule::Report()
{
    int NUMBINS = 20;
    int BINSIZE = TheProgram.number_of_dimers/(NUMBINS*10);
 
    int total_bound = 0;
    if (TheProgram.silent) return;

    if (BINSIZE > 16) BINSIZE=16;  // should adjust above to be non-linear instead...

    Microtubule *Current = Microtubule::First;
    int polymerised = 0;

    float time = TheProgram.NoCyclesCompleted*TheProgram.TimeScale ;
    int hist[NUMBINS];

    for (int i=0;i<NUMBINS;i++) hist[i]=0;

    cout << "\n\n ------------ AGGREGATED POLYMER REPORT ------------\n";
    cout << " time = " << time <<" seconds";
    cout << "    cycle = " << TheProgram.NoCyclesCompleted << "\n"; 


    while (Current != NULL)
    {
        if (Current->NoObjects<(NUMBINS*BINSIZE))
        {
           hist[(Current->NoObjects/BINSIZE)]++;
        }
        else
            hist[NUMBINS-1]++;   // tack extras on end... (happens *very* rarely, if ever!)

        total_bound += Current->NoObjects;

        polymerised += Current->NoObjects;
        
        // print out small sample for indiv. analysis...

        if (((Current->TagNo-1)%50)==0)
            DATALOG2 << "mt: " << Current->TagNo << " size: " << Current->NoObjects << "\n";

        Current = Current->Next;
    }

    /*
     * Console Output
     */

    cout << "\n 2+ ";
    for (int i=1; i<10; i++) { cout << setw(4) << (BINSIZE*i); }
    cout << "\n";
    for (int i=0; i<10; i++) { if (i<NUMBINS) cout << setw(4) << hist[i]; }

    cout << "\ntotal objects: " << Dimer::NoObjects << "  bound: " << total_bound;
    cout << " free: " << Dimer::NoObjects - total_bound;
    cout << " molarity: " << TheProgram.molarity << "\n";

    /*
     *  Datalog output
     */

    int dispWidth = 5;
    if (Dimer::NoObjects<10000) dispWidth = 4;
    if (Dimer::NoObjects<1000) dispWidth = 3;

    if (time == 0)  // first report
    {
        DATALOG1 << "\n# ***** AGGREGATED POLYMER REPORT *****\n\n";
        DATALOG1 << "\n# " <<  NUMBINS << " histogram bins, each of size " << BINSIZE << "\n#";
        DATALOG1 << "\n# molarity: " << TheProgram.molarity;
        DATALOG1 << "\n# no objects " << Dimer::NoObjects;
        DATALOG1 << "\n# field size: " << TheProgram.field_size;
        DATALOG1 << "\n# time step: " << TheProgram.TimeScale << "#\n#\n";
        DATALOG1 << "#          [bound] - "; 
        for (int i=0; i<NUMBINS; i++)
        {
            DATALOG1 << setw(dispWidth-1) << (i*BINSIZE) << "+";
        }
        DATALOG1 << "\n#\n";
    }

    DATALOG1 << "t: " << setw(6) << time << " b: " << setw(dispWidth+1) << total_bound << " - ";
    for (int i=0; i<NUMBINS; i++)
        DATALOG1<< setw(dispWidth) << hist[i];
    DATALOG1 << "\n"; 
    
}

void Microtubule::FinalReport()
{
    int NUMBUCKETS = 200;
 
    int total_bound = 0;
    if (TheProgram.silent) return;

    Microtubule *Current = Microtubule::First;
    turbidity = 0;
    count     = 0;

    float time = TheProgram.NoCyclesCompleted*TheProgram.TimeScale ;
    int hist[NUMBUCKETS];

    for (int i=0;i<NUMBUCKETS;i++) hist[i]=0;

    while (Current != NULL)
    {
        hist[Current->NoObjects]++;
        total_bound += Current->NoObjects;
        turbidity += (Current->NoObjects>600)?Current->NoObjects:0;
        Current = Current->Next;
    }

    hist[1] = Dimer::NoObjects - total_bound;  

    DATALOG3<< "#Final Histogram Report @ time ";
    DATALOG3<< time << " for " << Dimer::NoObjects << " objects \n"; 

    for (int i=1; i<NUMBUCKETS; i++)
        DATALOG3<< i << setw(6) << (hist[i]*i) << setw(6) << hist[i] << "\n"; 
    
    if (turbidity > 0)
        DATALOG1<< "#---- Final turbidity ---- \n";
        DATALOG1<< " ... Turbidity = " << turbidity << "( " << (turbidity*100/Dimer::NoObjects) << " %)\n";

}

void Microtubule::PhysicallyJoin(Dimer *A, Dimer *B, Link *Alink, Link *Blink)
{
assert (A->Owner->IntegrityCheck("in Phys Join I")==OK);

cout << "Joining MTs " << A->Owner->TagNo << " & " << B->Owner->TagNo << " #objects:";
cout << A->Owner->NoObjects << " & " << B->Owner->NoObjects << "\n";

    Dimer *JoiningDimer = B;

    Microtubule *MA = A->Owner;
    Microtubule *MB = B->Owner;

    // Modify to cope with wrap around problem:

cout <<"Attempting to join MT "<<A->Owner->TagNo << " with " << B->Owner->TagNo;
cout << "  size A = " << MA->NoObjects << " size B " << MB->NoObjects << "\n";

assert (A->Owner->IntegrityCheck("in Phys Join II")==OK);
    
    if (MA->CheckBeforeAdding(A, B, Alink, Blink)!=OK) return;

    MA->RecursivelyAddDimer(A, B, Alink, Blink);

    MA->RecalculateMassDetails();


//reality check

    if (JoiningDimer->Owner != A->Owner)
        cout << "It didn't take, Jim \n";

cout << "Joined MTs " << MA->TagNo << " & " << MB->TagNo ;
cout << " Size of MA: " << MA->NoObjects <<"\n";

}

int Microtubule::CheckBeforeAdding(Dimer *A, Dimer *B, Link *Alink, Link *Blink)
{
    // with complex structures, it may be necessary to check that space exists
    // for a dimer, even if their is a link free.  At the moment though, the
    // existence of a free link implies sufficient space for a binding dimer.
    return OK;
}

/*
 *  when merging two mts, we may find that part of the second mt cannot
 *  be inserted into the new - In which case the old mt is disassembled.
 *
 *  (WARNING: This may be the wrong answer.  Keep an eye on it.)
 */

void Microtubule::RecursivelyFreeDimer( Dimer *B, Microtubule* OwnerB)
{
    // Actually, this is now simply "free dimer" - no need to be recursive.

    B->FreeBySelf();

}

/*
 *  Add all the dimers attached to, and including, Dimer B to this aggregate
 *
 *  P.S. rewritten, there is now nothing recursive about it...
 *
 *  P.P.S. My Brain Hurts.
 */

void Microtubule::RecursivelyAddDimer( Dimer *A, Dimer *B, Link *Alink, Link *Blink)
{
    Microtubule* OwnerA = A->Owner;
    Microtubule* OwnerB = B->Owner;

    assert(OwnerA == this);
    assert(OwnerA->IntegrityCheck("Rec Add Start") == OK);  // Make sure mt is o.k. before we add more trouble.

    int finished = FALSE;

cout << "MERGING MTS: Linking "<< A->TagNo << " from mt: " << OwnerA->TagNo << " with " << B->TagNo << " from mt: " << OwnerB->TagNo << "\n";

    // Copy info from old microtubule (OwnerB)

    int NoDimersInB = OwnerB->NoObjects;
    Dimer* TempArray[NoDimersInB];
    Link*  TempLinks[NoDimersInB][MAXSITES];  

    /*
     *  Copy all the data about dimers in the to-be-assimilated microtubule B
     *  into the TempArray and TempLinks buffers.
     */

    for (int i=0; i<NoDimersInB; i++)   
    {
        Dimer* CopiedDimer = OwnerB->DimerPtArray[i];
        assert(CopiedDimer!=NULL);
        TempArray[i] = CopiedDimer;
//TEMP
        cout << i << ": " << CopiedDimer->TagNo << "  ";

        TempArray[i]->tempFlag = FALSE;
        for (int j=0;j<CopiedDimer->NoSites;j++)
        {
            TempLinks[i][j] = new Link(CopiedDimer->LinkNo[j]);
            if (TempLinks[i][j]->to == A)                 // we're not interested in the link 
                TempLinks[i][j]->to = NULL;               // to the new mt. (should only be one) 
//TEMP
            if (TempLinks[i][j]->to != NULL)              // just some debugging output... 
                cout << " " << TempLinks[i][j]->to->TagNo;
            else 
                cout << "  -  ";
        }

        for (int j=CopiedDimer->NoSites; j<MAXSITES; j++) 
            TempLinks[i][j] = new Link();
//TEMP
        cout << "\n";
    }

    /*
     *  Set the "special" flag for the one dimer that has initiated the link-up
     */

    B->tempFlag = TRUE;        // Dimer B is base reference pt.

    /*
     * Get Rid of the old Microtubule; set dimer owners, but not links, to null.
     */
    for (int i=NoDimersInB-1; i>=0; i--) 
    {
        TempArray[i]->unbind();
        assert(TempArray[i]->Owner == NULL);
        assert(TempArray[i]->GetBound() == FALSE);
    }

    


    /* 
     * Start linking the dimers to their new owner, mt A
     */

//    TheProgram.verbose = TRUE;
    // Make initial link.

    if (Dimer::LinkUp(A,B,Alink,Blink)==ERROR) 
    {
        cout << "WARNING: Unable to make first linkage in MT join... ";
        cout << "destroying 2nd mt and aborting\n";
        return; 
    }

    while (finished == FALSE)
    {
        finished = TRUE;
        for (int i=0; i<NoDimersInB; i++)
        {
           Dimer* Test = TempArray[i];
           assert (Test != NULL);

           if (Test->tempFlag==TRUE)       // this one is correctly positioned
           {
               for (int j=0; j<Test->NoSites; j++)
               {
                   Dimer* Neighbour = TempLinks[i][j]->to;
                   if (Neighbour != NULL) // now position its links.
                   {
                       if (Neighbour->tempFlag == FALSE )
                       {
                           // Find Neighbour in array
cout << "looking for " << Neighbour->TagNo << "\n";
                           int N_pos = 0;
                           for (N_pos=0; N_pos<NoDimersInB; N_pos++)
                           {
                               if (TempArray[N_pos]==Neighbour) 
                                   break;
                           }
                           assert (N_pos < NoDimersInB);
cout << "found it at " << N_pos << "\n";
                           // Find Neighbour -> Test dimer link in link array

cout << "looking for link at row " << N_pos  <<" to " << Test->TagNo << " from " << Neighbour->TagNo << "\n";

                           int k = 0;
                           for (k=0; k<Neighbour->NoSites; k++)
                           {
if (TempLinks[N_pos][k]->to != NULL)
cout << " link to : " << TempLinks[N_pos][k]->to->TagNo << " vs test " << Test->TagNo << "\n";
                               if (TempLinks[N_pos][k]->to == Test) break;
                           }
                           assert(k != Neighbour->NoSites);
cout << "found it at " << k << "\n"; 
                           Link* FrontLink = &(Test->LinkNo[j]);
                           Link* BackLink  = &(Neighbour->LinkNo[k]);  
//                           Link* BackLink = Neighbour->FindLink(Test);

                           if (BackLink == NULL) cout << "Error - null backlink in MTRecAddDim()\n";

cout << "adding dimer: " << Neighbour->TagNo << " to already added dimer " << Test->TagNo << " using links: \n" << FrontLink << "\n" << BackLink << "\n\n"; 
 

                           if (Dimer::LinkUp(Test,Neighbour,FrontLink,BackLink) == ERROR)
                           {
                               cout << "can't add dimer " << Neighbour->TagNo << " to mt: " << OwnerA->TagNo << "\n";
//                               Neighbour->unbind();
                               Neighbour->jump();
                           }
                           else
                           {
                               Test->CheckStateOnBind(FrontLink, BackLink);
                               Neighbour->CheckStateOnBind(BackLink, FrontLink);

                               cout << " successful add of " << Neighbour->TagNo << " to mt: " << OwnerA->TagNo << "\n";
                           }
                           finished = FALSE;
                           Neighbour->tempFlag = TRUE;
                       }    
                   }
               }
           }
        }
    }

    if (TheProgram.verbose)
        cout << "MT " << TagNo << " now size: " << NoObjects << "\n"; 

    if (OwnerA->CheckCrossLinks(B) == ERROR)
    {
        cout << " Error in Check Cross Links mt: "<<OwnerA->TagNo<<" after merge magic\n";
    };

    if (OwnerA->IntegrityCheck("merge check ") == ERROR)
    {
        cout << " Error in Integrity Check mt: "<<OwnerA->TagNo<<" after merge magic\n";

    };

    cout << " end rec add dimer\n ";

//    TheProgram.verbose = FALSE;
}

/**
 *   Sometimes floating point errors accumulate to such an extent that
 *   Dimers start to "drift" out of alignment.  This recalculates *all*
 *   Dimer positions using the "first" dimer as a starting reference
 *   point.  The actual positional movements should be very, very
 *   small under normal circumstances.
 *
 *   nb. "tempFlag" is  simply a marker flag showing whether the dimer
 *       has been processed yet.
 */

void Microtubule::RecalculateDimerPositions()
{
    int finished = FALSE;

    for (int i=1; i<NoObjects; i++)   
        DimerPtArray[i]->tempFlag = FALSE;
    DimerPtArray[0]->tempFlag = TRUE;        // Dimer 0 is base reference pt.

    while (finished == FALSE)
    {
        finished = TRUE;
        for (int i=0; i<NoObjects; i++)
        {
           Dimer *A = DimerPtArray[i];
           if (A->tempFlag==TRUE)       // this one is correctly positioned
           {
               for (int j=0; j<A->NoSites; j++)
               {
                   Dimer* B = A->LinkNo[j].to;
                   if (B != NULL) // now position its links.
                   {
                       if (B->tempFlag == FALSE )
                       {
                           finished = FALSE;
                           Link* BackLink = B->FindLink(A);
                           Dimer::Align(A,B,&(A->LinkNo[j]),BackLink);
                           B->tempFlag = TRUE;
                       }    
                   }
               }
           }
        }
    }
    cout << "finished re-align of mt: " << TagNo << "\n";
}

