/*
 *   This is a general user interface class.  It handles graphical
 *   and file i/o.  Most of it is taken up with sgi-specific code
 *   to do fast display of the dimers, represented as sphere pairs,
 *   but there are also routines to save data in povray (a popular
 *   raytracer) format, as well as saving data files for 'virtual
 *   turbidity" and frequency counts for different sized microtubules. 
 */

#include "surfette.h"
#include "tubular.h" 
#include "dimer.h"
#include "gridcell.h"
#include "microtubule.h" 


#ifndef NOSGI
#include "unistd.h"  // for "sleep" call...

Matrix Identity =
    { {1, 0, 0, 0}, 
      {0, 1, 0, 0}, 
      {0, 0, 1, 0}, 
      {0, 0, 0, 1}};

float ColourMatrix[MAXCOLOURS][15];


float black[] = 
{
    AMBIENT, .0, .0, .0,
    DIFFUSE, 0, 0, 0,
    SPECULAR,.5, .5, .5,
    SHININESS, 10,
    LMNULL,
};
/*
float green[] = 
{
    AMBIENT, .1, .1, .1,
    DIFFUSE, 0, .9, 0,
    SPECULAR,.5, .5, .5,
    SHININESS, 10,
    LMNULL,
};

float red[] = 
{
    AMBIENT, .1, .1, .1,
    DIFFUSE, 0.9, 0, 0,
    SPECULAR,.5, .5, .5,
    SHININESS, 10,
    LMNULL,
};

float red2[] = 
{
    AMBIENT, .1, .1, .1,
    DIFFUSE, 0.8, 0.8, 0,
    SPECULAR,.5, .5, .5,
    SHININESS, 10,
    LMNULL,
};

float red3[] = 
{
    AMBIENT, .1, .1, .1,
    DIFFUSE, 0.2, 0.2, 0,
    SPECULAR,.5, .5, .5,
    SHININESS, 10,
    LMNULL,
};

float red4[] = 
{
    AMBIENT, .1, .1, .1,
    DIFFUSE, 0.0, 0.0, 0.0,
    SPECULAR,.5, .5, .5,
    SHININESS, 10,
    LMNULL,
};
*/

float lm[] =
{
    AMBIENT, .1, .1, .1,
    LOCALVIEWER, 1,
    LMNULL
};

float lt[] =
{
    LCOLOR, 1, 1, 1,
    POSITION, 0, 0, 1, 0,
    LMNULL
};
#endif

static int SHOW_UNBOUND = TRUE;

static int ShowLinks = FALSE;

static int PauseMode = FALSE;

int Surf::CheckUserInput()
{
#ifdef NOSGI
#else
  if (!TheProgram.NoGraphics)
  {
    if (getbutton(RKEY)) 
    {    
       return view();  // WARNING; a bit of a hack: this *can* cause 
                           // recursion, but shouldn't be serious... :-)
    }

    if (getbutton(QKEY)) 
    { 
        CloseWindow(); 
cout << " user initiated quit - returning TRUE " << TRUE << "\n";
        return (TRUE); 
    }
        
    if (getbutton(BKEY)) {SHOW_UNBOUND = !SHOW_UNBOUND; }

    if (getbutton(SKEY)) Write_next_rgb_file(); 
        
    if (getbutton(VKEY)) Write_next_pov_file();

    if (getbutton(LKEY)) {ShowLinks = !ShowLinks; }

    if (getbutton(PKEY)) {PauseMode = !PauseMode; sleep(1);} 
    
    if (getbutton(HKEY)) WriteHelp(); 
  }
#endif
    return FALSE;
} 

void Surf::WriteHelp()
{
#ifndef NOSGI
  if (TheProgram.NoGraphics) return;
  
        color(BLACK);

        cmov2i(50,140);
        charstr("hit [R] to refresh the view");
        cmov2i(50,120);
        charstr("hit [F] to toggle free dimer display");  
        cmov2i(50,100);
        charstr("hit [S] to save the image in sgi format"); 
        cmov2i(50,80);
        charstr("hit [V] to save the image in povray format");
        cmov2i(50,60);
        charstr("hit [P] to pause"); 
        cmov2i(50,40);
        charstr("hit [C] to continue after pausing"); 
        cmov2i(50,20);
        charstr("hit [Q] to quit"); 
#endif
}

int Surf::view()
{
#ifndef NOSGI
    if (TheProgram.NoGraphics) return 0;

    if (!TheProgram.silent)
        ERRORLOG << "surf\n";
    
    long xorigin, yorigin, xsize, ysize;
    getorigin(&xorigin, &yorigin);
    getsize(&xsize, &ysize);

    assert (Dimer::Dimers[0].Model != NULL);
    assert (Dimer::Dimers[0].Model->Parser != NULL);
    TubParser *Parser = Dimer::Dimers[0].Model->Parser;

    do 
    {
        displayScene(xorigin, yorigin, xsize, ysize, Parser);
        if (CheckUserInput()==TRUE)
        {
cout << "second return at TRUE " << TRUE << "\n";
             return TRUE;  // this may reset 'PauseMode'
        }
    } while ((PauseMode==TRUE)&&(!getbutton(CKEY)) );  // 'C' Key = emergency bail out ...
#endif
    return 0;
}

/* 
 *  Draws a scene by iterating throught the dimer objects and displaying the
 *  component spheres as icosahedrons
 */

void Surf::displayScene(long xorigin, long yorigin, long xsize, long ysize, TubParser *Parser)
{
    float ry = 300 * (2.0*(getvaluator(MOUSEX)-xorigin)/xsize-1.0);
    float rx = -300 * (2.0*(getvaluator(MOUSEY)-yorigin)/ysize-1.0);
    czclear(0xFFFFFF, getgdesc(GD_ZMAX));

/*
 *   This could be "go-fastered" muchly if visual display speed becomes important.
 */

    Dimer *D; 
    for (int i=0; i<Parser->NoColours; i++)
    {
        lmbind(MATERIAL, i+1);
        pushmatrix();
        rot(ry, 'y');
        rot(rx, 'x');
        for (int j=0; j<Dimer::NoObjects; j++)
        {
           D = &Dimer::Dimers[j];               
           if (SHOW_UNBOUND || (D->Owner != NULL) )
           {
               for (int s=0; s<D->NoSpheres; s++)
               {
                   int Col = D->Model->GetColour(s, D->State);  
                   if (Parser->GetColourIndex(Col) == i) 
                       DrawSphere(D->SpherePos[s], D->SphereRad[s]); 
               }
           }
        } 
        if (ShowLinks == TRUE)
        {
            color(BLACK);
            for (int j=0; j<Dimer::NoObjects; j++)
            {
                D = &Dimer::Dimers[j];               
                for (int s=0; s<D->NoSites; s++)
                {
                    DrawSphere(D->LinkNo[s].true_pos, 0.3);                   
                }
            }
        }
        popmatrix();
    } 
    color(BLACK);
    int xpos = (int)(((TheProgram.width>512)?-TheProgram.width/6:-TheProgram.width/4)*((float)TheProgram.field_size/200.0));
         
    int ypos = (int)(((TheProgram.height>512)?-TheProgram.height/6:-TheProgram.height/4)*((float)TheProgram.field_size/200.0));
    cmov2i(xpos,ypos);
    char temp[80];
    sprintf(temp, "FRAMES: %ld             ", TheProgram.NoCyclesCompleted);
    charstr(temp);

    swapbuffers();
}

void Surf::Write_next_rgb_file()
{
#ifndef NOSGI
    if (TheProgram.NoGraphics) return;

    char Buf[30];
    char* Numb ="00000";
    static int rgbno = 0;
   
    long X1,Y1,X2, Y2;
    getorigin(&X1,&Y1);
    getsize(&X2,&Y2);

    int len = (rgbno==0)?1:1 + (int) (log10(rgbno)); //length of rgbno in chars 
    Numb[5-len]='\0';
    sprintf(Buf,"scrsave tub%s%i.rgb %ld %ld %ld %ld",Numb,rgbno++,X1,X1+X2,Y1,Y1+Y2);
    system(Buf);
#endif
}

void Surf::Write_next_pov_file()
{
    char Buf[30];
    char* Numb = "00000";
    static int povno = 0;

    int len = (povno==0)?1:1 + (int) (log10(povno)); //length of povno in chars
    Numb[5-len]='\0';
    sprintf(Buf,"tub%s%i.pov",Numb,povno++);

    ofstream OutFile;
    OutFile.open(Buf); 

    if (!OutFile)
        cerr <<"unable to open " << Buf <<"\n";
    else
        cout << "saving povray frame "<<povno << " saved \n";

    Write_pov_file(&OutFile); 

    OutFile.close();
}

void Surf::Write_pov_colour(ofstream *F, int C)
{
   float R = (float)(C&0xFF0000)/((float)0xFF0000);
   float G = (float)(C&0xFF00)/((float)0xFF00);
   float B = (float)(C&0xFF)/((float)0xFF);

   if ( (C&0xFF000000) > 0)   // there is an alpha component
   {
       float A = (float)(C&0xFF000000)/((float)0xFF000000);
       *F << "colour rgbf <" << R << " "<<G<<" "<<B<<" "<<A<<">"; 
   }
   else
       *F << "colour rgb <" << R << " " << G << " " << B << ">";
}

void Surf::Write_pov_file(ofstream* F)
{
    Write_pov_header(F);

    Dimer* D;

    // Draw Alpha Tubulin as blue
    // diphosphorelated Beta Tubulin as Green,
    // triphosphorelated Beta Tubulin as Red

    for (int i=0; i<Dimer::NoObjects; i++) 
    {
        D = &Dimer::Dimers[i];

        if (D->NoSpheres == 2) 
        {
            int ColA = D->Model->GetColour(0, D->State);
            int ColB = D->Model->GetColour(1, D->State);
            Write_pov_diblob(F, D, ColA, ColB); 
        }
        else
        { 
            for (int s=0; s<D->NoSpheres; s++)
                Write_pov_sphere(F, D->SpherePos[s], D->Model->GetColour(s,D->State), D->SphereRad[s]);
        }

        if (ShowLinks == TRUE)
        {
            for (int j=0; j<D->NoSites; j++)
            {
                pt dirtn = D->Basis * D->LinkNo[j].Type->dirtn;
                Write_pov_link(F, D->LinkNo[j].true_pos, dirtn, 0x66FF66);
            } 
        }
    }
}

void Surf::Write_pov_header(ofstream* F)
{
    *F << HEADER << "\n\n";
//    *F << CAMERA << "\n\n";
    *F << "\ncamera {\n  location <0, 0, "<< (GridCell::Field_Size) << ">\n  look_at <0,0,0>\n}";
    *F << POVBACKGROUND<<"\n";
    *F << LIGHTS << "\n\n"<<TEXTURE <<"\n"<<START<<"\n";
}

void Surf::Write_pov_sphere(ofstream* F, pt C, int Colour, float Radius, char* Finish)
{
    *F << "\n\nsphere { " << C << "," << Radius;  
    WriteCurrentTexture(F, Colour, Finish);
    *F << "  no_shadow\n";
    *F << "}\n";
}

void Surf::Write_pov_link(ofstream* F, pt pos, pt dir, int Colour, char* Finish)
{
    pt end = pos+dir;
    pt tip = end + dir;
  
    *F << "\nunion {\n";
    *F << "\n  cylinder { " << pos << ","      << end << ",0.2 }";
    *F << "\n  cone     { " << end << ",0.4, " << tip << ",0 }";
    WriteCurrentTexture(F, Colour, Finish);
    *F << " no_shadow\n";
    *F << "}\n";

}

void Surf::Write_pov_diblob(ofstream *F, Dimer *D, int ColourA, int ColourB)
{
    static const float SQRT2 = sqrt(2);

    pt SpherePosA = D->SpherePos[0];
    pt SpherePosB = D->SpherePos[1];
    // Sphere Radii are "increased" by sqrt(2) to make the blob stuff
    // draw the radii at their true pos when the threshold = 0.25.  trust me.
    float SphereRadA = D->SphereRad[0] * SQRT2;
    float SphereRadB = D->SphereRad[1] * SQRT2;


    *F << "\n";
    *F << "blob {\n";
    *F << "  threshold 0.25\n";
    *F << "  component 1.0," << SphereRadA <<", " << SpherePosA << "\n";
    *F << "  component 1.0," << SphereRadB <<", " << SpherePosB << "\n";
    *F << "  texture {\n";
    *F << "    pigment {\n";
//    pt temp = ((SpherePosB-SpherePosA)/16);
    pt temp = (SpherePosB-SpherePosA);
    float dist = SphereRadA + SphereRadB + pt::abs(temp);
    temp = ~temp * (dist/90) ;
    *F << "      gradient " << temp << "\n";
    *F << "      color_map {\n";
    *F << "        [0.0 "; Write_pov_colour(F,ColourA); *F <<"]\n";
    *F << "        [0.20 "; Write_pov_colour(F,ColourA); *F <<"]\n"; 
    *F << "        [0.30 "; Write_pov_colour(F,ColourB); *F <<"]\n"; 
    *F << "        [0.70 "; Write_pov_colour(F,ColourB); *F <<"]\n"; 
    *F << "        [0.80 "; Write_pov_colour(F,ColourA); *F <<"]\n"; 
    *F << "        [1.00 "; Write_pov_colour(F,ColourA); *F <<"]\n";
    *F << "      }\n";
    *F << "      translate "<<SpherePosA<<"\n";
    *F << "    }\n  }\n}\n";
}

/*
void Surf::Write_pov_diblob(ofstream *F, pt A, pt B, int ColourA, int ColourB)
{
    *F << "\n";
    *F << "blob {\n";
    *F << "  threshold 0.3\n";
    *F << "  component 1.0, 1.7, " << A << "\n";
    *F << "  component 1.0, 1.7, " << B << "\n";
    *F << "  texture {\n";
    *F << "    pigment {\n";
    pt temp = ((B-A)/16);
    *F << "      gradient " << temp << "\n";
    *F << "      color_map {\n";
    *F << "        [0.1 "; Write_pov_colour(F,ColourA); *F <<"]\n";
    *F << "        [0.15 "; Write_pov_colour(F,ColourB); *F <<"]\n"; 
    *F << "        [0.75 "; Write_pov_colour(F,ColourB); *F <<"]\n"; 
    *F << "        [0.75 "; Write_pov_colour(F,ColourA); *F <<"]\n";
    *F << "      }\n";
    *F << "      translate "<<A<<"\n";
    *F << "    }\n  }\n}\n";
}
*/

void Surf::WriteCurrentTexture(ofstream* F, int Colour, char* Finish)
{
    *F << "\n  texture {";

    *F<<"\n    pigment { "; Write_pov_colour(F, Colour); *F << " }";
    if (Finish!=NULL)
        *F<<"\n    finish  { "<<Finish<< " }";    

    *F << "\n    }\n";
}

/* 
 *
 *
 */

void Surf::WriteTurbidity(int Step)
{
    long result = 0;

    ofstream OutFile;
    OutFile.open("turbidity,ios::app"); 
    OutFile << "# data for step : "<<Step<<"\n";

    Microtubule* current = Microtubule::First;
    while (current != NULL)
    {
        if (current->NoObjects > 300)    // we treat turbidity as a linear ftn
        {                              // of microtubules with more than 600
           result += current->NoObjects; // monomers.
        }
        current = current->Next;
    }

    OutFile << Step << " " << result << "\n";
    OutFile.close();  //File closed to preserve data in event of termination
}

/* 
 *
 *
 */

void Surf::WriteFrequencyCount(int Step)
{
    const int ArraySize = 50; 
    int Small[ArraySize];
    int Large[ArraySize];
 

    ofstream OutFileSmall, OutFileLarge;
    OutFileSmall.open("freqcountS",ios::app);
    OutFileLarge.open("freqcountL",ios::app);
    OutFileSmall << "# data for step : "<<Step<<"\n";
    OutFileLarge << "# data for step : "<<Step<<" -> each increment = " << ArraySize<< " dimers\n";

    Microtubule* current = Microtubule::First;
    while (current != NULL)
    {
        if (current->NoObjects < ArraySize)
            Small[current->NoObjects]++;
        Large[current->NoObjects/ArraySize]++;
        current = current->Next;
    }
    for (int i=0; i<ArraySize; i++)
    {
        OutFileSmall << Small[i] << " ";
        OutFileLarge << Large[i] << " ";
    }
    OutFileSmall<<"\n"; //File closed to preserve data in event of termination
    OutFileLarge<<"\n"; //File closed to preserve data in event of termination
    OutFileSmall.close();
    OutFileLarge.close();
}


/*
 *   Draws a sphere using SGI illumination.
 *
 *   Draws Radius 1 and 2 spheres slightly faster
 */


void Surf::DrawSphere(pt P, float R)
{
#ifndef NOSGI
    if (TheProgram.NoGraphics) return;

static int i;

static float M[12][3];  //array of modified vertices

/* 
 *  The following gives vertex, normal, and face information for
 *  an arbitrary "standard" dodecahedron  
 */

// "radius" 1

static float V1 [12] [3] =
{{0.0,       1.0,            0.618      },
{ 0.0,       1.0,           -0.618      },
{ 1.0,       0.618,          0.0        },
{ 1.0,      -0.618,          0.0        },
{ 0.0,      -1.0,           -0.618      },
{ 0.0,      -1.0,            0.618      },
{ 0.618,     0.0,            1.0        },
{ -0.618,    0.0,            1.0        },
{ 0.618,     0.0,           -1.0        },
{ -0.618,    0.0,           -1.0        },
{ -1.0,      0.618,          0.0        },
{ -1.0,     -0.618,          0.0        }};

// "radius" 2

static float V2 [12] [3] =
{{0.0,       2.0,            1.236      },
{ 0.0,       2.0,           -1.236      },
{ 2.0,       1.236,          0.0        },
{ 2.0,      -1.236,          0.0        },
{ 0.0,      -2.0,           -1.236      },
{ 0.0,      -2.0,            1.236      },
{ 1.236,     0.0,            2.0        },
{ -1.236,    0.0,            2.0        },
{ 1.236,     0.0,           -2.0        },
{ -1.236,    0.0,           -2.0        },
{ -2.0,      1.236,          0.0        },
{ -2.0,     -1.236,          0.0        }};


static float N [12] [3] =
{{0.0,       0.8507,          0.5257    },
{ 0.0,       0.8507,         -0.5257    },
{ 0.8507,    0.5257,          0.0       },
{ 0.8507,   -0.5257,          0.0       },
{ 0.0,      -0.8507,         -0.5257    },
{ 0.0,      -0.8507,          0.5257    },
{ 0.5257,     0.0,            0.8507    },
{ -0.5257,    0.0,            0.8507    },
{ 0.5257,     0.0,           -0.8507    },
{ -0.5257,    0.0,           -0.8507    },
{ -0.8507,    0.5257,         0.0       },
{ -0.8507,   -0.5257,         0.0       }};

/**************************************************************
         Here starteth the procedure
 **************************************************************/

    if (R==1)  // fast drawing for radius 1 spheres 
    {
        for (i=0; i<12; i++)  
        {
            M[i][0] = V1[i][0] + P.x;
            M[i][1] = V1[i][1] + P.y;
            M[i][2] = V1[i][2] + P.z;
        }
    }
    else if (R==2)  // fastish drawing for radius 2 spheres
    {
        for (i=0; i<12; i++)  
        {
            M[i][0] = V2[i][0] + P.x;
            M[i][1] = V2[i][1] + P.y;
            M[i][2] = V2[i][2] + P.z;
        }
    }
    else   // everyday drawing for other radius spheres
    {
        for (i=0; i<12; i++)  
        {
            M[i][0] = R*V1[i][0] + P.x;
            M[i][1] = R*V1[i][1] + P.y;
            M[i][2] = R*V1[i][2] + P.z;
        }
    }

//This *does* obey the right hand rule.  Not as fast, but.

    bgntmesh();
    n3f(N[0]); v3f(M[0]); 
    n3f(N[1]); v3f(M[1]); 
    n3f(N[2]); v3f(M[2]); 
    swaptmesh();
    n3f(N[8]); v3f(M[8]); 
    swaptmesh();
    n3f(N[9]); v3f(M[9]); 
    swaptmesh();
    n3f(N[10]); v3f(M[10]); 
    swaptmesh();
    n3f(N[0]); v3f(M[0]); 
    endtmesh();

    bgntmesh();   
    n3f(N[4]); v3f(M[4]); 
    n3f(N[5]); v3f(M[5]); 
    n3f(N[3]); v3f(M[3]); 
    swaptmesh();
    n3f(N[6]); v3f(M[6]); 
    swaptmesh();
    n3f(N[7]); v3f(M[7]); 
    swaptmesh();
    n3f(N[11]); v3f(M[11]); 
    swaptmesh();
    n3f(N[4]); v3f(M[4]); 
    endtmesh();
 

    bgntmesh();
    n3f(N[6]); v3f(M[6]); 
    n3f(N[2]); v3f(M[2]); 
    n3f(N[3]); v3f(M[3]); 
    swaptmesh();
    n3f(N[8]); v3f(M[8]); 
    endtmesh();

    bgntmesh();
    n3f(N[3]); v3f(M[3]); 
    n3f(N[8]); v3f(M[8]); 
    n3f(N[4]); v3f(M[4]); 
    swaptmesh();
    n3f(N[9]); v3f(M[9]); 
    endtmesh();

    bgntmesh();
    n3f(N[4]); v3f(M[4]); 
    n3f(N[9]); v3f(M[9]); 
    n3f(N[11]); v3f(M[11]); 
    swaptmesh();
    n3f(N[10]); v3f(M[10]); 
    endtmesh();

    bgntmesh();
    n3f(N[11]); v3f(M[11]); 
    n3f(N[10]); v3f(M[10]); 
    n3f(N[7]); v3f(M[7]); 
    swaptmesh();
    n3f(N[0]); v3f(M[0]); 
    endtmesh();

    bgntmesh();
    n3f(N[7]); v3f(M[7]); 
    n3f(N[0]); v3f(M[0]); 
    n3f(N[6]); v3f(M[6]); 
    swaptmesh();
    n3f(N[2]); v3f(M[2]); 
    endtmesh();

#endif
}

void Surf::InitWindow(int width, int height, int FieldSize, TubParser* Parser)
{	
#ifndef NOSGI
    if (TheProgram.NoGraphics) return;

    if (!TheProgram.silent);
	ERRORLOG << "Init Window " << width << " x " << height << " \n";
    prefposition(1, width-1, 1, height-1);
    winopen("Sim Tubulin [R]-Refresh View  [H]-Help"); 
    winconstraints();
    RGBmode();
    doublebuffer();
    gconfig();
    //lsetdepth(getgdesc(GD_ZMIN), getgdesc(GD_ZMAX));
    zbuffer(TRUE);
    mmode(MVIEWING);
    loadmatrix(Identity);
    //         field aspect  near        far
    float aspect = (float)width/(float)height;
cout <<"... " << aspect << "\n";
    perspective(int(600/aspect), aspect, FieldSize/2, FieldSize*6.0); 
    //         dist    azimuth inc twist
    polarview(FieldSize*1.2,0, 0, 0);

    SetUpColours(Parser);
/*
    lmdef(DEFMATERIAL, 1, 0, blue);
    lmdef(DEFMATERIAL, 2, 0, green);
    lmdef(DEFMATERIAL, 3, 0, red);
    lmdef(DEFMATERIAL, 4, 0, red2);
    lmdef(DEFMATERIAL, 5, 0, red3);
    lmdef(DEFMATERIAL, 6, 0, red4);
*/
    lmdef(DEFLIGHT, 1, 0, lt);
    lmdef(DEFLMODEL, 1, 0, lm);
    lmbind(MATERIAL, 1);
    lmbind(LMODEL, 1);
    lmbind(LIGHT0, 1);  
    //cpack(BACKGROUND); 	
    clear();
    clear();
    ERRORLOG << "Finished Init Window\n";
#endif
}

void Surf::SetUpColours(TubParser *Parser)
{
#ifndef NOSGI
    for (int i=0; i<Parser->NoColours; i++)
    {
        // this seems clumsy.  I wonder if there's a better way...
        float temp[] = 
        {
            AMBIENT, .1, .1, .1,
            DIFFUSE, Parser->Red(i), Parser->Green(i), Parser->Blue(i),
            SPECULAR,.5, .5, .5,
            SHININESS, 10,
            LMNULL,
        };
        for (int j=0; j<15; j++) 
            ColourMatrix[i][j] = temp[j];
        lmdef(DEFMATERIAL, i+1, 0, ColourMatrix[i]);
    }
//    lmdef(DEFMATERIAL, Parser->NoColours, 0, black); // give ourselves black as an extra...
#endif
}

void Surf::CloseWindow()
{
#ifndef NOSGI
    if (TheProgram.NoGraphics) return;
    {
        gexit();
    }
#endif
}




