
/*
 *   This parses the P.D.D.F. file, and loads up a set of data structures
 *   to be used by the main part of the program to create the appropriate
 *   virtual proteins. 
 *
 *   There is a menagerie of data classes (see the header file)
 *   class ModelData  - holds data for an entire protein, comprising:
 *   class SiteData   - a name for a binding site
 *   class SphereData - position and size of a sphere
 *   class StateData  - a particular state of the protein
 *   class EventData  - an event that may change the state of the protein
 *
 *   also imporant is
 *   class LinkData   - a link from a site to other link(s)
 *      ... the links are held seperately from the models, and define
 *   interactions between different sites in different states
 *
 *
 *   Most of this code is simply loading up these data classes.
 *   The usual metaphor is:
 *   
 *   GetNextWord();  if (Word == "a_test_word") then...
 *   ... this works because "Word" is global to the parser class, and is
 *   obtained by GetNextWord().  The current "Word" is used so widely, and
 *   by so many functions, that is seemed redundant to pass it around
 *   explicitly.
 *
 *   nb. This class can be run stand alone for development and testing:
 *   simply uncomment the hash define STANDALONE...  or compile with something like:
 *
 *       g++ -DSTANDALONE tubparser.cc -o tubparser
 *
 *
 *   BTW: apologies - this isn't the world's neatest code.
 */

// #define STANDALONE

#include <assert.h>
#include <iomanip.h>

#include "dataclasses.h"
#include "tubparser.h"

char ShortString::InputBuffer[255];    // This is used to read a word from file.

ParseWord* ParseWord::Last  = NULL;    // Used in the Parse Word linked list of
ParseWord* ParseWord::First = NULL;    // key words.

char *ParseWord::Model   = "Model";    // The actual list of Parse Word key words
char *ParseWord::State   = "State";
char *ParseWord::Site    = "Site";
char *ParseWord::Event   = "Event";
char *ParseWord::Binding = "Binding";
char *ParseWord::Average = "Average";
char *ParseWord::Mix     = "Mix";
char *ParseWord::Sphere  = "Sphere";
char *ParseWord::Diffusion = "Diffusion";
char *ParseWord::TimeScale = "TimeScale";
char *ParseWord::Viscosity = "Viscosity";
char *ParseWord::Temperature = "Temperature";
char *ParseWord::Start   = "{";
char *ParseWord::End     = "}";


// Some I/O operators for the ShortString mini-class

#ifdef STANDALONE    

ostream& operator << (ostream &o, ShortString &S)
{
    return o << S.text;
}

ostream& operator << (ostream &o, ShortString *S)
{
    return o << *S; 
}

istream& operator >> (istream &i, ShortString &S)
{
    i >>  ShortString::InputBuffer;
    S = ShortString::InputBuffer;
    return i;
}

istream& operator >> (istream &i, ShortString *S)
{
    return i >> *S;
}


// a repeat of standard pt class i/o code

ostream& operator << (ostream &o, pt &p)
{   
    if (p.x<0.0001 && p.x>-0.0001) p.x = 0; 
    if (p.y<0.0001 && p.y>-0.0001) p.y = 0; 
    if (p.z<0.0001 && p.z>-0.0001) p.z = 0; 
    o<<"<"<<p.x<<","<<p.y<<","<<p.z<<"> "; 
    
    return o;
}   

ostream& operator << (ostream &o, pt *p)
{
    if (p==NULL)
        o << "<NULL>";
    else
        o << "*"<<(*p);
    return o;
}                                    
#endif

/* A Blow in from dataclasses.h */
/* or not.
LinkData* ModelData::GetLink(int stateno, int siteno)
{
    return Parser->GetLinkNo(StateNo[stateno]->SiteIndex[siteno]);
}
*/


/*
 *    Given a string, this returns the type of a word (which may be
 *    "Unknown"). 
 *
 *    e.g. this converts strings such as "Binding" to members of
 *    the ParseType enumeration. 
 */

ParseWord::ParseType ParseWord::FindType(char *S)   // won't win any speed awards, but who cares!
{
    assert ( First!=NULL );

    if (strchr(S, '#')) { return Comment; }

    if (strchr(S, '<')) { return Point; }
 
    // Meeh... clumsy, but what the heck... 
    if (strchr(S, '0')) { return Number; } 
    if (strchr(S, '1')) { return Number; } 
    if (strchr(S, '2')) { return Number; } 
    if (strchr(S, '3')) { return Number; } 
    if (strchr(S, '4')) { return Number; } 
    if (strchr(S, '5')) { return Number; } 
    if (strchr(S, '6')) { return Number; } 
    if (strchr(S, '7')) { return Number; } 
    if (strchr(S, '8')) { return Number; } 
    if (strchr(S, '9')) { return Number; } 

    ParseWord* Iterator = First;

    while (Iterator != NULL)
    {
        if (*Iterator==S) return Iterator->IsA;    
        Iterator = Iterator->Next;
    }
    return Unknown;

}

/*
 *   returns TRUE if the string matches a key word (a sub class of ParseType)
 */

int ParseWord::IsKeyWord(ShortString S)
{
    if (S==ParseWord::Model) return TRUE; 
    if (S==ParseWord::State) return TRUE; 
    if (S==ParseWord::Site) return TRUE; 
    if (S==ParseWord::Sphere) return TRUE; 
    if (S==ParseWord::Event) return TRUE; 
    if (S==ParseWord::Binding) return TRUE; 
    if (S==ParseWord::Average) return TRUE; 
    if (S==ParseWord::Mix) return TRUE; 
    if (S==ParseWord::Diffusion) return TRUE; 
    if (S==ParseWord::TimeScale) return TRUE; 
    if (S==ParseWord::Temperature) return TRUE; 
    if (S==ParseWord::Viscosity) return TRUE; 
    if (S==ParseWord::Start) return TRUE; 
    if (S==ParseWord::End) return TRUE; 

    return FALSE;
}

// Debugging tool

void ParseWord::PrintList()
{
    assert ( First != NULL );
    ParseWord* Iterator = First;
    while (Iterator != NULL) 
    {
        cout << "ParseWord " << Iterator->text << " " << Iterator->IsA << "\n";
        Iterator = Iterator->Next;
    } 

}

// Initialises the list of KeyWords

void ParseWord::Init()
{   
    ParseWord* dummy;
    dummy = new ParseWord(Model,   ModelKeyWord);
    dummy = new ParseWord(State,   StateKeyWord);
    dummy = new ParseWord(Site,    SiteKeyWord); 
    dummy = new ParseWord(Event,   EventKeyWord);
    dummy = new ParseWord(Binding, BindingKeyWord);
    dummy = new ParseWord(Average, AverageKeyWord);
    dummy = new ParseWord(Mix,     MixKeyWord);
    dummy = new ParseWord(Sphere,  SphereKeyWord);
    dummy = new ParseWord(Diffusion,  DiffusionKeyWord);
    dummy = new ParseWord(TimeScale,  TimeScaleKeyWord);
    dummy = new ParseWord(Viscosity,  ViscosityKeyWord);
    dummy = new ParseWord(Temperature,  TemperatureKeyWord);
    dummy = new ParseWord(Start,   StartBlockKeyWord);
    dummy = new ParseWord(End,     EndBlockKeyWord);
}

/*
 *  Registers a new colour into the colour index, if it doesn't 
 *  already exist.
 */

void TubParser::RegisterNewColour(int NewColour)
{
    if (GetColourIndex(NewColour)==-1)
    {
        ColourNo[NoColours++] = NewColour;
    }
}


/*
 *   Gets the index value of a colour.  
 */

int  TubParser::GetColourIndex(int Value)
{
    for (int i=0; i<NoColours; i++)
        if (ColourNo[i] == Value) return i;
    return -1;
}


/*
 *   Gets the rgb value of a colour index.
 */

int  TubParser::GetColourValue(int Index)
{
    if (Index>=0 & Index<NoColours) return ColourNo[Index];
    cerr << "Unknown Colour " << Index << "\n";
    return -1;
}


/*
 *  Return the appropriate components of an indexed colour
 */

float TubParser::Red(int Index)
{
    return ( (float) (ColourNo[Index]&0xFF0000) ) / (float) 0xFF0000;
}

float TubParser::Green(int Index)
{
    return ( (float) (ColourNo[Index]&0xFF00) ) / (float) 0xFF00;
}

float TubParser::Blue(int Index)
{
    return ( (float) (ColourNo[Index]&0xFF) ) / (float) 0xFF;
}


/*
 * returns a modeldata pointer, by looking up a model name in 
 * the TubParser list of models.
 */

ModelData* TubParser::GetModelByName(ShortString S)
{
    for (int i=0; i<NoModels; i++)
        if (ModelNo[i]->Name==S) return ModelNo[i];
    return NULL;
}

/*
 * Prints a parser error
 */

void TubParser::error(char *Text)
{
    cout << "Error Parsing " << Text << "\n";
}

/*
 *    Gets the next real word (i.e. not a comment)
 * 
 *    (Nb. "Word" is a data member of the TubParser object. 
 *     - i.e. global in this class)
 */

int TubParser::GetNextWord()
{
    while (InFile >> Word && Word(0)=='#')   // Skip commented line portions,
       InFile.getline(Line, LINELEN);        // marked by a word starting with '#'

    return (InFile)?TRUE:FALSE;     
}

/*
 *   The way things are modelled: Proteins have binding sites.  There are
 *   links from these binding sites.  Links are not mentioned explicitly
 *   in the PDDF data files, but every site has at least one link.  If the
 *   behaviour of a binding site changes (due to dephosphorelation, say),
 *   a new link is created for that site and that state. A link belongs
 *   to one site, and lists the binding/breaking probabilities to all the
 *   other links in the simulation, including itself. 
 *
 *   To summarise:  A link is the binding info for a site/state combination.
 *
 *   A link implicitly holds for all later states which don't explicitly
 *   modify it.
 *
 *   This ftn parses the Binding info between two site/state combinations.
 *   If links currently exists for the two sites and states (note: a link
 *   is automatically created for each site and the base state) this ftn
 *   simply updates the bind/break values for those links.
 *
 *   If link(s) don't exist for the site/state pairs, new link(s) are
 *   created.
 *
 *   ... This is probably the most complex ftn in the parser, so don't
 *   dispair - everything else is pretty straight forward :-) 
 */

void TubParser::ParseBinding()
{
    assert ( InFile );

  /***********************************************************************
   *   First, grab the info from the line of text and store it in temporary
   *   variables: Site?No, State?No, bindprob, breakprob, Model?
   **********************************************************************/

    int SiteANo  = -1;   // the (model specific) no of the first site
    int StateANo = -1;   // the (model specific) no of the first state
    int SiteBNo  = -1;   // the (model specific) no of the second site
    int StateBNo = -1;   // the (model specific) no of the second state

    ModelData *ModelA, *ModelB;  // pointers to the two protein models
    float breakprob = 0;
    float bindprob = 0;  // the probabilities of breaking or binding


    GetNextWord();
    ModelA = GetModelByName(&Word); 
    if (ModelA==NULL) {cerr << "Unknown Model in Binding: " <<Word<<"\n"; return ;}

    GetNextWord();
    ShortString SiteAName = Word;
    if (SiteAName != "all")
    {
        SiteANo = ModelA->GetSiteNo(&SiteAName);
        if (SiteANo==-1) {cerr << "Unknown site in Binding" <<Word<<"\n"; return ;}

        GetNextWord();
        StateANo = ModelA->GetStateNo(&Word);
        if (StateANo==-1) {cerr << "Unknown state in Binding" <<Word<<"\n"; return ;}
    }

    GetNextWord();
    if (Word!="to") {cerr<<"Expected 'to' in Binding, instead of "<<Word<<"\n"; return ;}

    GetNextWord();
    ModelB = GetModelByName(&Word); 
    if (ModelB==NULL) {cerr << "Unknown Model in Binding: " <<Word<<"\n"; return ;}

    GetNextWord();
    ShortString SiteBName = Word;
    if (SiteBName != "all")
    {
        SiteBNo = ModelB->GetSiteNo(&SiteBName);
        if (SiteBNo==-1) {cerr << "Unknown site in Binding" <<Word<<"\n"; return ;}

        GetNextWord();
        StateBNo = ModelB->GetStateNo(&Word);
        if (StateBNo==-1) {cerr << "Unknown state in Binding" <<Word<<"\n"; return ;}
    }

    GetNextWord();
    if (Word!="=") {cerr<<"Expected '=' in Binding, instead of "<<Word<<"\n"; return ;}

    GetNextWord();

    if (Word.MatchStart("break("))         // handle variable syntax:  Either
        breakprob = atof(&Word.text[6]);   // 0.4 0.01     # bindprob breakprob
    else if (Word.MatchStart("bind("))     // or 
        bindprob = atof(&Word.text[5]);    // bind(0.4) break(0.01)
    else
        bindprob = Word.asFloat(); 
  
    GetNextWord();

    if (Word.MatchStart("break(")) 
        breakprob = atof(&Word.text[6]);
    else if (Word.MatchStart("bind(")) 
        bindprob = atof(&Word.text[5]);
    else 
        breakprob = Word.asFloat(); 


/**********************************************************************
*   End Parsing...    begin updating/creating links
*
*   at this stage we should have:
*   First Model, State and Site          (ModelA, StateANo, SiteANo)
*   Second Model, State and Site         (ModelB, StateBNo, SiteBNo)
*   binding probability                  (bindprob)
*   breaking probability                 (breakprob)
*
**********************************************************************/

    LinkData* Alink = NULL;
    LinkData* Blink = NULL;

    if (SiteAName != "all") 
    {
        Alink = GetExactLink(ModelA, SiteANo, StateANo);

        if (Alink==NULL)
        /*
         *   We can't find a pre-existing link for siteA and stateA, so 
         *   we need to create a New link for a new State
         */
        {
            // First, find the links predecessor (there must be at least *one*, since
            // one is created when the site is created)

            LinkData* TempLink = GetLink(ModelA, SiteANo, StateANo);
            if (TempLink==NULL) 
            {
                cerr << "error - can't find link "<<ModelA->Name<< " " << SiteAName<<"\n";
                return;
            }

            // Create the new link, having all the break/bind properties of the old one,
            // and the same site, but having a different state

            // THIS IS WHERE THE DEFAULT-TO-OLD-LINK-STRENGTH STUFF LIVES. 
            // MAYBE INCLUDE A "NO ASSUMPTIONS, MUST DEFINE ALL COMBINATIONS" OPTION

            Alink = new LinkData(TempLink, StateANo);  // create
            AddLink(Alink);                            // add to the parsers list of links
                                                       // (also updates other links) 

            // This link is now the default for all later states
 
    if (!TheProgram.silent) 
            cout << "setting model " << ModelA->Name << " state " << StateANo << " site " << Alink->SiteNo << " to link " << Alink->LinkNo << "\n"; 
//            for (int i=StateANo; i<ModelA->NoStates; i++)
//                ModelA->GetStateByNo(i)->SetLink(Alink->SiteNo, Alink->LinkNo); 
            ModelA->GetStateByNo(StateANo)->SetLink(Alink->SiteNo, Alink->LinkNo); 
            
        }
    }
   
    // this code is identical to the block above, but done for the second site
 
    if (SiteBName != "all") 
    {
        Blink = GetExactLink(ModelB, SiteBNo, StateBNo);

        if (Blink==NULL)
        {
            LinkData* TempLink = GetLink(ModelB, SiteBNo, StateBNo);
            if (TempLink==NULL) 
            {
                cerr << "error - can't find link "<<ModelB->Name<< " " << SiteBName<<"\n";
                return;
            }

// THIS IS WHERE THE DEFAULT-TO-OLD-LINK-STRENGTH STUFF LIVES. MAY NEED TO CHANGE, OR
// INCLUDE A "NO ASSUMPTIONS, MUST DEFINE ALL COMBINATIONS" OPTION

            Blink = new LinkData(TempLink, StateBNo);
            AddLink(Blink);

    if (!TheProgram.silent) 
            cout << "setting model " << ModelB->Name << " state " << StateBNo << " site " << Blink->SiteNo << " to link " << Blink->LinkNo << "\n"; 
//            for (int i=StateBNo; i<ModelB->NoStates; i++)
//                ModelB->GetStateByNo(i)->SetLink(Blink->SiteNo, Blink->LinkNo); 

              ModelB->GetStateByNo(StateBNo)->SetLink(Blink->SiteNo, Blink->LinkNo); 
        }
    }


/*****************************************************************
 *   Finally, set the actual link values.
 *****************************************************************/


// We handle the special case "all" for link values, which mean "all currently
// existing links".  We just iterate through them as necessary.

    if ((SiteAName == "all") && (SiteBName=="all"))
    {
        for (int i=0; i<TotalLinks; i++)
            for (int j=0; j<TotalLinks; j++)
                if ((LinkNo[i]->Owner==ModelA)&&(LinkNo[j]->Owner==ModelB))
                    SetLinkValues(LinkNo[i], LinkNo[j], bindprob, breakprob);

    }
    else if (SiteAName == "all")
    {
        for (int i=0; i<< TotalLinks; i++)
            if (LinkNo[i]->Owner==ModelA)
                SetLinkValues(LinkNo[i], Blink, bindprob, breakprob);
    }
    else if (SiteBName == "all")
    {
         for (int j=0; j<< TotalLinks; j++)
                 if (LinkNo[j]->Owner==ModelB)
                     SetLinkValues(Alink, LinkNo[j], bindprob, breakprob);
    }
    else
/*
 *      Otherwise, there's no need to iterate - just set the values
 */
        SetLinkValues(Alink, Blink, bindprob, breakprob);

// now aren't you glad all that's over?
}


/*
 *   Sets the bind / break values for a pair of links.  See "ParseBinding()" for
 *   a full description of the horror of Links. 
 */

void TubParser::SetLinkValues(LinkData* A, LinkData* B, float bindprob, float breakprob)
{
    A->SetBindProb(B->LinkNo,  bindprob);
    A->SetBreakProb(B->LinkNo, breakprob);
    B->SetBindProb(A->LinkNo,  bindprob);
    B->SetBreakProb(A->LinkNo, breakprob);
}


/*
 *  We allow spheres to change colour, reflecting different states.  
 *  To make life nice, we allow for a number of different pre-defined
 *  colours, as text strings.
 */

int TubParser::ParseColour(ShortString ColString)
{
    int ReturnVal;
    ReturnVal = ColString.asInt();
    if (ReturnVal==0)
    {
        if (ColString=="red")             ReturnVal=0xFF0000; 
        else if (ColString=="green")      ReturnVal=0x00FF00;
        else if (ColString=="blue")       ReturnVal=0x0000FF;
        else if (ColString=="yellow")     ReturnVal=0xFFFF00;
        else if (ColString=="magenta")    ReturnVal=0xFF00FF;
        else if (ColString=="cyan")       ReturnVal=0x00FFFF;
        else if (ColString=="white")      ReturnVal=0xFFFFFF;
        else if (ColString=="pink")       ReturnVal=0xFF9999;
        else if (ColString=="lightgreen") ReturnVal=0x99FF99;
        else if (ColString=="lightblue")  ReturnVal=0x9999FF;
        else if (ColString=="lightred")   ReturnVal=0xFF9999;  // = pink
        else if (ColString=="white")      ReturnVal=0xFFFFFF;
        else if (ColString=="black")      ReturnVal=0x000000;
        else if (ColString=="gray")       ReturnVal=0x999999;
        else if (ColString=="lightgray")  ReturnVal=0xCCCCCC;
        else if (ColString=="darkgray")   ReturnVal=0x666666;
        else
        {
            cerr << "\n ERROR: Unknown colour in data file - "<<ColString << "\n\n"; 
            ReturnVal=0x666666;
        }
    }
    return ReturnVal;
}


/*
 *  Parses a Sphere, which is simply a name, a radius, a center point and a colour,
 *  and creates that sphere.
 */

SphereData* TubParser::ParseSphere(ModelData* ThisModel)
{
     GetNextWord();
     ShortString SphereName = Word; // Name - necessary for use in States that change colours
     GetNextWord();
     float Rad = Word.asFloat();    // Radius
     GetNextWord();
     pt Cent = Word.asPoint();      // Center point 
     GetNextWord();
     int Col = ParseColour(Word);   // Colour                         

     RegisterNewColour(Col);

     return new SphereData(ThisModel, SphereName, Cent, Rad, Col); 
}


/*
 *   Parses a site, and creates a new site. 
 */

SiteData* TubParser::ParseSite(ModelData* ThisModel)
{
     GetNextWord();
     ShortString SiteName = Word;    // Name - used in Bindings 

     return new SiteData(ThisModel, SiteName);
}


/*
 *   Adds a new link to the parsers list of links, and updates any 
 *   other links that might be affected.
 */

void TubParser::AddLink( LinkData* L)
{
    if (L==NULL) {cerr<< "Error adding null link - skipping\n"; return;}
    
    if (TotalLinks<MAXLINKS) 
    {
        L->SetNo(TotalLinks); 
        LinkNo[TotalLinks] = L;

    /*
     *   The link knows all a lot about itself - it's model, state, site etc.
     *   Now we inform the model fully about the link:
     */

         ModelData* Model = L->Owner;    // the owning model.
         Model->AddLink(L, TotalLinks);  // set the link in the model, with it's global link pos.  
                                         // (a bit of a hack - should neaten up some time... 

    /*
     *   The current model assumes that all later state links have all the
     *   properties of the earlier link at a given site, except for those
     *   explicitly changed.  This reflects a "later stage" link across all
     *   existing links with the same values as its predeccessor at a site.
     *
     *   For the first links, created along with the sites, this code won't
     *   do anything since all the break/bind probs are set to be <0.
     */

        for (int i=0; i<TotalLinks; i++)
        {
             if (L->BindProb[i]>=0)
                 LinkNo[i]->SetBindProb(TotalLinks, L->BindProb[i]);
             if (L->BreakProb[i]>=0)
                 LinkNo[i]->SetBreakProb(TotalLinks, L->BreakProb[i]);
        }
        TotalLinks++;
    } 
    else
    {
        cerr << "Parse Error: Maximum no Links ("<<MAXLINKS<<") exceeded. Change";
        cerr << " value in hashdefines.h and recompile\n";
    }
}


/*
 *   Calculates the boolean values required for the "go-faster" shortcuts
 *   for links in the program.
 */

void TubParser::FinaliseLinks()
{
    // assign each link a unique boolean identifier, a power of two
    // (e.g. [binary] 000100)
 
    for (long i=0; i<TotalLinks; i++)
    {
//        template = 1;
        if (i<32)
            LinkNo[i]->TypeId = (1 << i);
        else if (i<64)
            LinkNo[i]->TypeIdHigh = (1 << (i-32)); 
        else
            cerr << "ERROR: Max Links (64) Exceeded!\n";
    }

    // give each link a boolean template to compare against possible binding
    // partners.  (e.g. [binary] 010110, would match 000100 above as a possible
    // binding partner)

    for (int i=0; i<TotalLinks; i++)
    {
        LinkNo[i]->BindsTo = 0;
        LinkNo[i]->BindsToHigh = 0;
        for (int j=0; j<TotalLinks; j++)
        {
            if ((LinkNo[i]->BindProb[j]>=0) || (LinkNo[i]->BreakProb[j]>=0))
            {
                LinkNo[i]->SetBinding(LinkNo[j]);
//                LinkNo[i]->BindsTo += LinkNo[j]->TypeId;
            }
        }
    }

    
}

void TubParser::FinaliseModels()
{
    for (int i=0; i<NoModels; i++)
        ModelNo[i]->CalculateDetails(Viscosity, Temperature, TimeScale);
    for (int i=0; i<NoModels; i++)
    {
        ModelNo[i]->CalculateProbMatrix();
    }
}

/*
 *   Searches for a link *exactly* matching the model, site and state specified
 */

LinkData*  TubParser::GetExactLink(ModelData* M, int Si, int St)
{
    for (int i=0; i<TotalLinks; i++)
       if ((LinkNo[i]->Owner==M)&&(LinkNo[i]->SiteNo==Si)&&(LinkNo[i]->StateNo==St))                 return LinkNo[i];
    return NULL; 
}


/*
 *   Searches for a link exactly matching the specified model and site,
 *   and having the given state, or the nearest lower state. 
 */

LinkData*  TubParser::GetLink(ModelData* M, int Si, int St)
{
    LinkData* ReturnVal;
    
    for (int i=St; i>=0; i--)
    {
        ReturnVal = GetExactLink(M,Si,i);        
        if (ReturnVal != NULL) return ReturnVal;
    }
   
    return NULL;   // Couldn't find this link, or any link with a smaller state no.
}


/*
 *   parses and creates the first link for each site.
 *   This is called when the parser gets a line starting with "site..." -
 *   first it creates the site, then the initial link
 */

void TubParser::ParseInitialLink(ModelData* ThisModel, ShortString SiteName)
{
    pt pos, dir, twistVect;
    float twist=0;

    int SiteNo = ThisModel->GetSiteNo(&SiteName);
    if (SiteNo<0)
    {
        cerr << "Error parsing site name " << SiteName << " in link text\n";
        return ;
    }
    GetNextWord();
    pos = Word.asPoint();           // site position
    dir = ~pos;                     // direction default = normalised pos
    GetNextWord();                 

    int StateNo = 0;   // This is the first link created for this site, so it
                       // must also be the first state.
                       // nb : this state may not be created yet, but a default
                       //      state always is, so there *will* be a state 0

    if (!ParseWord::IsKeyWord(Word)) // get a specific direction direction
    {
        dir = Word.asPoint();       // direction of initial link
        GetNextWord();

        if (!ParseWord::IsKeyWord(Word)) // get a twist as well...
        {
            if (ParseWord::FindType(Word)  == ParseWord::Point)
            {
                 twistVect = Word.asPoint(); 
cout << "Read twistVect as: " << twistVect << "\n";
                 double dot_prod = twistVect^dir; // check twist vect is orthogonal...
                 if (dot_prod>0.0001 || dot_prod < -0.0001)
                 {
                     cerr << "Error parsing site " << SiteName << ": non-orthogonal twist vector. dot product was: " << dot_prod << "\n"; 
                 }    
                 else
                     AddLink(new LinkData(ThisModel, SiteName, SiteNo, StateNo, pos, dir, twistVect));

            }
            else if (ParseWord::FindType(Word)  == ParseWord::Number)
            {
                twist = Word.asFloat(); // twist of initial link
cout << "Read twist as: " << twist << "\n";
                AddLink(new LinkData(ThisModel, SiteName, SiteNo, StateNo, pos, dir, twist));
 
            }
            else
            {
                 cerr << "Error parsing site " << SiteName << " bad twist '";
                 cerr << Word << "'\n";
                 return;
            }
        }
//     else ... AddLink....
    }
    AddLink(new LinkData(ThisModel, SiteName, SiteNo, StateNo, pos, dir, 0));

}


/*
 *   Used by ParseEvent(), this takes a list of zero, one or two sites
 *   enclosed by curly braces {}.
 */

int TubParser::ParseSiteList( ModelData *ThisModel, int *Site, int *LinkedSite)
{
        GetNextWord();

        if (Word != "{") {cerr<<" '{' missing in Event\n"; return -1;}
        GetNextWord();
        if (Word == "}") 
        {
            *Site = ALL; *LinkedSite = ALL; return 0;
        }
        *Site = ThisModel->GetSiteNo(&Word);
        GetNextWord();
        if (Word == "}")
        {
            *LinkedSite = ALL; return 0;
        }
        *LinkedSite = ThisModel->GetSiteNo(&Word);
        GetNextWord();
        if (Word != "}")
        {
            cerr << "error - missing '}' in Event\n"; return -1;
        }

        return 0;
}


/*
 *   parses an "event", which can be a binding, a breaking, or a random
 *   event.  An event changes one particular state to another particular
 *   state.
 */

EventData* TubParser::ParseEvent(ModelData* ThisModel)
{
    int EventType;
    int Site       =  UNKNOWN;
    int LinkedSite =  UNKNOWN;
    float RandVal  =  0;
    int OldState   = -1;
    int NewState   = -1;

    GetNextWord();

    if (Word=="random")
    {
        EventType = RANDOMEVENT;
        GetNextWord();
        RandVal = Word.asFloat(); 
    }
    else if (Word=="bind")
    {
        EventType = BINDEVENT;
        if (ParseSiteList( ThisModel, &Site, &LinkedSite)) return NULL;
    }
    else if (Word=="break")
    {
        EventType = BREAKEVENT;
        if (ParseSiteList( ThisModel, &Site, &LinkedSite)) return NULL;
    }
    else
    {
        cout << "Unknown Event: " << Word << "\n";
        return NULL;
    }
    GetNextWord();
    OldState = ThisModel->GetStateNo(&Word); 
    GetNextWord();
    if (Word != "->") { cerr<< "Error - expected '->' in Link, got: "<<Word<<"\n"; return NULL; }
    GetNextWord();
    NewState = ThisModel->GetStateNo(&Word);

    EventData* NewEvent;
    if (EventType==RANDOMEVENT)
        NewEvent = new EventData(ThisModel, EventType, OldState, NewState, RandVal);
    else
        NewEvent = new EventData(ThisModel, EventType, OldState, NewState, Site, LinkedSite);

    return NewEvent; 
}


/*
 *    The very first state of a virtual protein defaults to having the
 *    colours the spheres were defined with, and the first links created
 *    as well.  
 *
 *    nb. this "first state" gets created even if no states are explicitly
 *    specified in the PDDF file.
 */

void TubParser::InitFirstState(ModelData* ThisModel, StateData* FirstState)
{
     // nb. The First State of a given model is always numbered 0
     const int FirstStateNo = 0;

     for (int i=0; i<ThisModel->NoSpheres; i++)
     {
         FirstState->SetColour(i, ThisModel->SphereNo[i]->colour);
     }
     for (int i=0; i<ThisModel->NoSites; i++)
     {
//cout << "link " << i << "\n";
//cout << "link no = " << GetLink(ThisModel, i, FirstStateNo) << "\n";
         FirstState->SetLink(i, GetLink(ThisModel, i, FirstStateNo)->LinkNo); 
     }
}


/*
 *   A State is simply a name, and possibly a colour change for a sphere.
 *   Changes in linkage strengths are done with Bindings, which are aware
 *   of states, but not vica-versa.
 *
 *   FUTURE WORK:  Possibly states could be made to change the positions,
 *                 directions and twists of links - the design sorta allows
 *                 for this to happen.
 */

StateData* TubParser::ParseState(ModelData* ThisModel)
{
    GetNextWord();
    ShortString StateName = Word;
    StateData* NewState; 

    if (ThisModel->NoStates>0)  // check if this is the first state
    {
        NewState = new StateData(ThisModel, StateName);

        // The new states inherits the properties of its immediate predecessor
        NewState->SetPropertiesTheSameAs(*ThisModel->StateNo[ThisModel->NoStates-1]);
    }
    else
    {
         NewState = new StateData(ThisModel, StateName, TRUE);
         InitFirstState(ThisModel, NewState);
    }

    GetNextWord();

    while (Word == "colour" || Word == "changelink")
    {
        if (Word=="colour")  // change a sphere colour when this state is active
        {
            GetNextWord();
            ShortString SphereName = Word; 
            int SphereNo = ThisModel->GetSphereNo(&SphereName);
            GetNextWord();
            if (Word != "=")
            {
                 cerr << "Syntax error in State " << StateName << " expected '='\n";
                 cerr << ", got " << Word << "\n";
                 return NULL;
            }
            GetNextWord();
            int Colour = ParseColour(Word);

            RegisterNewColour(Colour);

            NewState->SetColour(SphereNo, Colour);
        }
        // this might be where that moving link stuff goes... :-)
        else if (Word=="changelink")   
        {
            cerr << "Sorry, changelink not yet implemented\n";
        }
        GetNextWord();
    }
    if (ParseWord::IsKeyWord(Word) == FALSE) 
    cout << "Unexpected word parsing state " << StateName <<"\n";
    
    return NewState;
}


/* 
 *   The top level model parsing function.  Mainly it checks for parser
 *   keywords, and calls the appropriate keyword-parsing function.
 */

ModelData* TubParser::ParseModel()
{
    assert (InFile != NULL);

    GetNextWord();
    ShortString Name = Word;   // The name of the model

    ModelData*  ThisModel = new ModelData(this, Name.asString());

    int NotFinished = GetNextWord();   // Finished if stream ends, or on "}"
    int NextWordAlreadyGot;            // Sometimes subfunctions get the next word before
                                       // returning to this function. A tad messy.

    while (NotFinished)
    {
        NextWordAlreadyGot = FALSE;

        ParseWord* term = new ParseWord(Word); // in hindsight, a list of string compares
                                               // might have been less complex

        switch(term->IsA) 
        { 
            case ParseWord::EndBlockKeyWord:
        
                NotFinished = FALSE; break;
       
            case ParseWord::StartBlockKeyWord:
        
                break;                         // do nothing

            case ParseWord::DiffusionKeyWord:

                GetNextWord();        
cout << "\nFound Diffusion as " << Word.asFloat() << "\n\n";
                ThisModel->SetDiffusionConst(Word.asFloat()*1e-11); break; 

            case ParseWord::SphereKeyWord:

                 ThisModel->AddSphere( ParseSphere(ThisModel)); break;
         
            case ParseWord::SiteKeyWord:
            { 
                 SiteData *ThisSite = ParseSite(ThisModel); 

                 ThisModel->AddSite(ThisSite); 
                 ParseInitialLink(ThisModel, ThisSite->Name); 

                 if (ParseWord::IsKeyWord(Word)==TRUE) // if the current word is a keyword, it 
                     NextWordAlreadyGot = TRUE;        // must have been prematurely obtained 
                                                       // by ParseInitialLink()
                 break;
            } 

            case ParseWord::StateKeyWord:
        
                 ThisModel->AddState(ParseState(ThisModel));
                 if (ParseWord::IsKeyWord(Word)==TRUE)        // see immediately above
                     NextWordAlreadyGot = TRUE;
                 break;

            case ParseWord::EventKeyWord:
        
                ThisModel->AddEvent(ParseEvent(ThisModel));      
                break;

            default:
                cout << "Unknown Parse Term in Model: " << Word << "\n"; 
        } 

        if (NotFinished && (NextWordAlreadyGot == FALSE))
        {
            NotFinished = GetNextWord();
        }

    }
 
    // Sanity Checks

    if (ThisModel->NoSpheres == 0)
        ThisModel->AddSphere(new SphereData(ThisModel, "default", pt(0,0,0), 1, 0xFFFFFF));
 
    // Set Defaults, if neccessary

    if (ThisModel->NoStates == 0)   // no state explicitly set, so use a default
    {
        StateData *NewState = new StateData(ThisModel, "Default");
        InitFirstState(ThisModel, NewState);
        ThisModel->AddState(NewState);
    }

    return ThisModel; 
}


/*
 *   Parse the proportional mix of different types of proteins
 *
 *   nb. - Word.asFloat() will cope with both .45 and 45% as valid
 *         proportions.
 */

void TubParser::ParseMix(ModelData* ModelNo[], int NoModels)
{
    int i,j;

    for (i=0; i<NoModels; i++)
    {
        GetNextWord();
        j=0;
        if (ModelNo[j]==NULL ) 
        {
             cout << "Error - Model Null in ParseMix()! "; 
             break;
        }

        while (ModelNo[j]->Name != Word)
        {
            if ( ++j >= NoModels ) 
            {
                cout <<"Parse Error: Unknown Model '"<<Word<<"' in Mix statement\n";
                ErrorStatus = TRUE;
                Word.text = ModelNo[0]->Name;
                break;
            }
        }
        GetNextWord(); 
        ModelNo[j]->SetFrequency( Word.asFloat() ); 
    }
}


/*
 *   The top level parse function: goes looking for "Model"s, "Binding"s, or "Mix"s.
 */

void TubParser::ParseFtn()
{
    assert (InFile != NULL);

    for (int i=0; i<MAXMODELS; i++) ModelNo[i]=NULL;

    while (GetNextWord())
    {
        ParseWord* term = new ParseWord(Word);
        switch(term->IsA)
        {
            case ParseWord::ModelKeyWord:
        
                AddModel(ParseModel());
                break; 
        
          case ParseWord::BindingKeyWord:
        
                ParseBinding();
                break; 

            case ParseWord::MixKeyWord:

                ParseMix(ModelNo, NoModels); 
                break;
       
            case ParseWord::ViscosityKeyWord:
 
                GetNextWord();
                Viscosity = Word.asFloat()/1000;
                break; 

            case ParseWord::TemperatureKeyWord:
 
                GetNextWord();
                Temperature = Word.asFloat();
                break; 

            case ParseWord::TimeScaleKeyWord:

                GetNextWord();        
                TimeScale = Word.asFloat()*1e-6; 
                break; 

            case ParseWord::Unknown: 
        
                cout << "Unknown Word: " << term->text << "\n";
                break;

            default:  
 
                cerr << "Parser Error in ParseFtn()\n"<<Word<<"\n";
 
        }
    }
    ParseCollisionDataFile();
    FinaliseLinks();
    FinaliseModels();
}    


/*
 *  Sanity check - prints out what it thinks the models are.
 */

void TubParser::PrintModels()
{
    for (int q=0; q<NoModels; q++)
        ModelNo[q]->print(q);
}

void TubParser::PrintColours()
{
    cout <<"\n\n **** Colours ****\n";

    for (int c=0; c<NoColours; c++)
        cout << "Colour ["<<hex<<c<<"] = "<<ColourNo[c] <<"\n";
    cout <<dec<<"\n";
}

/*
 *  Sanity check - prints out what it thinks the links are.
 */

void TubParser::PrintLinks()
{
    cout << "Print Links:  - "<<TotalLinks<<" links \n";
    for (int i=0; i<TotalLinks; i++)
    {
        cout << "\nLink " << LinkNo[i]->LinkNo << "\n";
        cout << "  model = " << LinkNo[i]->Owner->Name << "   site: ";
        cout << LinkNo[i]->Owner->GetSiteByNo(LinkNo[i]->SiteNo)->Name<<" state: ";
        cout << LinkNo[i]->Owner->GetStateByNo(LinkNo[i]->StateNo)->Name<<"\n";
        LinkNo[i]->print(i);
    }

    /*
     *   Print out binding matrix
     */
    cout << "\n\n All Links [bind] Probabilities \n           ";
    for (int i=0; i<TotalLinks; i++)
        cout << setw(11) << i; 
    cout << "\n            ";
    for (int i=0; i<TotalLinks; i++)
        cout << setw(10) << LinkNo[i]->Name  << " "; 
    cout << "\n            ";
    for (int i=0; i<TotalLinks; i++)
        cout << setw(10) << LinkNo[i]->Owner->GetStateByNo(LinkNo[i]->StateNo)->Name << " "; 
    cout << "\n";
    for (int i=0; i<TotalLinks; i++)
    {
        cout << "\n"<< setw(10) << LinkNo[i]->Name <<" ";
        for (int j=0; j<TotalLinks; j++)
            if (j>=i)
                cout << setw(11) << LinkNo[i]->BindProb[j];
            else
                cout << setw(11) << " ";
    }
    cout << "\n\n";

    /*
     *     Print out breakage matrix
     */

    cout << "\n\n All Links [break] Probabilities \n           ";
    for (int i=0; i<TotalLinks; i++)
        cout << setw(11) << i ;
    cout << "\n            ";
    for (int i=0; i<TotalLinks; i++)
        cout << setw(10) << LinkNo[i]->Name << " " ; 
    cout << "\n            ";
    for (int i=0; i<TotalLinks; i++)
        cout << setw(10) << LinkNo[i]->Owner->GetStateByNo(LinkNo[i]->StateNo)->Name << " "; 
    cout << "\n";
    for (int i=0; i<TotalLinks; i++)
    {
        cout << "\n"<< setw(10) << LinkNo[i]->Name <<" ";
        for (int j=0; j<TotalLinks; j++)
            if (j>=i)
                cout << setw(11) << LinkNo[i]->BreakProb[j];
            else
                cout << setw(11) << " ";
    }
    cout << "\n\n";
}

void TubParser::PrintEnvironment()
{
    cout << " Viscosity   = " << Viscosity   << " [kg/Ms]\n";
    cout << " Temperature = " << Temperature << " [Kelvin]\n";
    cout << " TimeScale   = " << TimeScale   << " [seconds]\n";
}


void TubParser::PrintMix()
{
    cout << "Model Frequencies\n";
    for (i=0; i<NoModels; i++)
    {
        cout << "Model " << ModelNo[i]->Name << " freq = " << ModelNo[i]->GetFrequency() << "\n";
    } 
}

void TubParser::InitDefaults()
{
    Temperature = 298;      // = 25 degrees centigrade
    Viscosity   = 0.891e-3; // = viscosity of water (at 298 kelvin)
    TimeScale   = 1e-6;     // = time steps defaults to 1 microsecond
}

ShortString TubParser::GetNextWord(ifstream* DataFile)
{
    static char buffer[1024];
    ShortString Word;

    int GotWord = FALSE;
    while (GotWord == FALSE && (*DataFile >> Word))
    {
        if (Word(0)=='#') 
            DataFile->getline(buffer,1024);
        else
            GotWord = TRUE;
    }
    return Word;
} 

int TubParser::ParseStructureFile(char *FileName)
{
cout << "Parsing file " << FileName << "\n";
    ifstream DataFile;
    DataFile.open(FileName);

    ShortString Word;

    Word = GetNextWord(&DataFile);
    int TotalStructObjects = Word.asInt();

    Structure = new StructureData(TotalStructObjects);

    /*
     *  "Model Numbers"
     *
     *  A bunch of variables refer to model numbers at various times, in
     *  slightly different contexts.  If the .struct file lists the models
     *  in the same order as the .pddf file these will all be identical, but
     *  this doesn't have to happen.
     *
     *  M_no - model number according to '.pddf' file, and the parser object 
     *  StructModelNo 
     */

    for (int i=0; i<NoModels; i++)
    {
        Word = GetNextWord(&DataFile);
        ModelData* Model = GetModelByName(Word);
        if (Model==NULL) 
        {
            cout << " Error in Structure File - ignoring file\n";
            cout << " (unable to find model " << Word << " in pddf file \n";
            return 0;
        }
        int M_no = Model->ModelNo; 
        Word = GetNextWord(&DataFile);
        Structure->SetNoObjectsForModel(M_no, Word.asInt());
    }

    Word = GetNextWord(&DataFile);
    int index = 0;
    for (int StructModelNo=0; StructModelNo<NoModels; StructModelNo++)
    {
        for (int j=0; j<Structure->NoType(StructModelNo); j++)
        {
            pt pos=0;
            CBMatrix Matrix;
            int StateNo = 0;

            ShortString ModelName = Word;
            if (GetModelByName(Word)->ModelNo != StructModelNo)
            {
                cout << " Error in Structure File - ignoring file\n";
                cout << " (unexpectedly encountered "<<ModelName<<" instead of correct name)\n";
                return 0;
            } 
            Word = GetNextWord(&DataFile);
            pos = Word.asPoint();
            Word = GetNextWord(&DataFile);
            if (Word(0) == '[') 
            {
                Word = GetNextWord(&DataFile);
                if (Word == "state") 
                {
                    Word = GetNextWord(&DataFile);
                    StateNo = GetModelByName(ModelName)->GetStateNo(&Word);
                    Word = GetNextWord(&DataFile);
                }
            }
// Create new PosData Object

            PosData* NewPD = new PosData(StructModelNo, pos, Matrix, StateNo); 
           
            Structure->SetData(index++, NewPD); 

        } 
    }

//    Structure->print();

    return Structure->NoStructureObjects;
}

/*
 *  A matrix of collision probabilities for different spheres with
 *  combined radii R at distances D is read from a file.  The
 *  file was created by the program "spheretest", which can be
 *  compiled by something like:
 *
 *  g++ spheretest.cc FastNorm.c -o spheretest -lm
 *
 *  (it uses FastNorm.c, a fast normal random number generator 
 *   written by Professor Chris Wallace.)
 */

void TubParser::ParseCollisionDataFile()
{
    ifstream DataFile;
    ShortString Word;
    char buffer[1024];
    
    int NoRadii, NoDists;

    for (int r=0; r<MAXRADS; r++)
        for (int d=0; d<MAXDISTS; d++)
             CollProbMatrix[r][d] = 0;             
 
    DataFile.open("SphereProgData");

    while (DataFile >> Word)
    {
        if (Word(0)=='#')  // read in variable length comment header
            DataFile.getline(buffer,1024);
        else // read in (a very) fixed format data file
        {
            int i;

            NoRadii = Word.asInt();
            for (i=0; i<NoRadii; i++)
                DataFile >> RadVals[i];                        
            for (i=NoRadii; i<MAXRADS; i++)
                RadVals[i] = 0;

            DataFile.getline(buffer,1024);
            DataFile.getline(buffer,1024);
            DataFile >> NoDists;
            for (i=0; i<NoDists; i++)
                DataFile >> DistVals[i];                        
            for (i=NoDists; i<MAXDISTS; i++)
                DistVals[i] = 0;

            DataFile.getline(buffer,1024);
            DataFile.getline(buffer,1024);

            for (int d=0; d<NoDists; d++)
                for (int r=0; r<NoRadii; r++) 
                    DataFile >> CollProbMatrix[r][d]; 
        }
    }
}

/*
 *  This finds the probability of collision of a Sphere of Radius Rad
 *  at a distance Dist from a pt,  moving with a standard deviation of 1. 
 *
 *  This is used to set up fast find arrays for each model type to each
 *  other model type. 
 *
 *  It is also used directly when calculating collision probs for bound
 *  objects.
 *
 */

float TubParser::CollProb(float Dist, float Rad)
{
    float a,b,c,d;
    int R = -1;    // index of Radius Value
    int D = -1;    // index of Distance Value

    if (Dist > DistVals[MAXDISTS-1]) return 0; // Dist too large.  No interaction.

    if (Dist < 0 ) return 1;  // Dist too small.  Certain interaction.

    if (Rad  > RadVals[MAXRADS-1])
    {
        R = MAXRADS-1;                        // Sphere size too large. Set to largest Known.
    }
    else
    {
        for (int i=0; i<MAXRADS; i++)         // Inefficient.  Would a log look up be better?
        {
            if (RadVals[i]>=Rad) {R=i; break;}
        }
    }

    for (int i=0; i<MAXDISTS; i++)            // Inefficient. ditto.
    {
        if (DistVals[i]>=Dist) {D=i; break;}
    }

    if (D==-1||R==-1) {cerr << " warning: Bad Values in CollProb("<<Dist<<","<<Rad<<") "<<DistVals[MAXDISTS-1]<<" "<<RadVals[MAXRADS-1] << " \n"; return 0;} 

    assert(R<MAXRADS);
    assert(D<MAXDISTS);
   
    if (R==0||D==0) return CollProbMatrix[R][D];

    d = CollProbMatrix[R-1][D-1];   // this will always be the biggest value             

    if (d==0) return 0;             // hence if 0, don't bother doing any more.

    a = CollProbMatrix[R][D];
    b = CollProbMatrix[R-1][D];
    c = CollProbMatrix[R][D-1];
    
    // Interpolate results: find weights.           e.g.:
                                                    // dist[1] 7 & [2] 9, dist value 7.5, say
    float Dspan    = DistVals[D]-DistVals[D-1];     // = 2 
    float WeightD1 = (DistVals[D] - Dist)/Dspan;    // = .75
    float WeightD2 = (Dist - DistVals[D-1])/Dspan;  // = .25

    float Rspan    = RadVals[R]-RadVals[R-1];
    float WeightR1 = (RadVals[R] - Rad)/Rspan;
    float WeightR2 = (Rad - RadVals[R-1])/Rspan;

    // linearly interpolate from data values.

    float result = (a*WeightD2+c*WeightD1)*WeightR2+(b*WeightD2+d*WeightD1)*WeightR1;

    return result;
}


#ifdef STANDALONE
int main (int argc, char *argv[])
{
    TubParser MyParser;
    ParseWord::Init();
    InitDefaults();
    //    ParseWord::PrintList();

    char *fake_args[2];
    fake_args[0] = "";
    fake_args[1] = "info.pddf";
    if (argc<=1)
        MyParser.Run(2,fake_args);
    else
        MyParser.Run(argc,argv);  // base class eventually runs ParseFtn()
    MyParser.PrintLinks();
    MyParser.PrintModels();
    MyParser.PrintColours();
    MyParser.PrintEnvironment();
    MyParser.PrintMix();
    return (0);
}
#endif

void TubParser::parse(char *infile)
{
    char *fake_args[2];
    fake_args[0] = "";
    if (infile == NULL)
        fake_args[1] = "info.pddf";
    else
        fake_args[1] = infile;

    cout << " Parsing " << infile << "\n";
    ParseWord::Init();
    InitDefaults();
    Run(2, fake_args); 
    
    if (!TheProgram.silent) PrintModels();
    if (!TheProgram.silent) PrintLinks();
    if (!TheProgram.silent) PrintColours();
    if (!TheProgram.silent) PrintEnvironment();
    if (!TheProgram.silent) PrintMix();
}
			

