/*
 *   A parser class - reads a file in, and extracts yummy useful info...
 *
 */

#ifndef TUBPARSER
#define TUBPARSER

#include "parser.h"
#include "hashdefines.h"
#include "dataclasses.h"
#include "structure.h"

#define DEGTORAD (3.141593/180.0)

/*
 *   "Parser" is a class defined in parser.h - a simple
 *   base class that handles the basics of reading the
 *   file in.
 *
 *   The "TubParser" class reads in the "protein dynamic
 *   description file" and translates it for the program.
 *   It was decided this was simple enough not to require
 *   the overhead of a full STL (or similar) implementation,
 *   instead a small "roll-your-own" implementation was made.
 *
 */

class TubParser : public Parser
{
    public:

    ShortString Word;
    const int LINELEN = 255;          //clean this up?
    char Line[LINELEN];     //...?
    int i;

    ModelData* ModelNo[MAXMODELS];
    LinkData*  LinkNo[MAXLINKS];
    int        ColourNo[MAXCOLOURS];
    int NoModels;
    int TotalLinks;
    int NoColours;

    // Collision Probability Data variables
    float      CollProbMatrix[MAXRADS][MAXDISTS];  // The actual collision probs
    float      DistVals[MAXDISTS];                 // list of effective distances 
    float      RadVals[MAXRADS];                   // list of effective radii

    float Temperature, Viscosity, TimeScale;

    int ErrorStatus;

    StructureData *Structure;                       // optional pre-existing structures 

    TubParser() {// for (int i=0;i<MAXCOLOURS;ColourNo[i++]=0); 
                  ErrorStatus = FALSE; NoModels=0; TotalLinks=0; NoColours = 0; }

    void error( char *Text );

    void        InitDefaults();
    void        ParseFtn();
    void        ParseMix( ModelData* ModelNo[], int NoModels ); 
    void        ParseBinding();
    ModelData*  ParseModel();
    int         ParseColour( ShortString ColString );
    SphereData* ParseSphere( ModelData* ThisModel );
    StateData*  ParseState( ModelData* ThisModel );
    SiteData*   ParseSite( ModelData* ThisModel );
    void        ParseInitialLink( ModelData* ThisModel, ShortString SiteName );
    EventData*  ParseEvent( ModelData* ThisModel );
    int         ParseSiteList( ModelData* ThisModel, int *Site, int *LinkedSite );

    void        InitFirstState( ModelData* ThisModel, StateData* FirstState );
    int         GetNextWord();
    ShortString GetNextWord(ifstream* DataFile);
    void        AddModel( ModelData* M ) {if (NoModels<MAXMODELS) ModelNo[NoModels++]=M;} 
    void        AddLink( LinkData* L); 

    LinkData* GetLinkNo(int No) {return LinkNo[No];}
    LinkData* GetLink(ModelData* M, int Si, int St);
    LinkData* GetExactLink(ModelData* M, int Si, int St);
    void      SetLinkValues(LinkData* A, LinkData* B, float bindprob, float breakprob);


    ModelData* GetModelByName(ShortString S); 

    void PrintLinks(); 
    void PrintModels(); 
    void PrintColours();
    void PrintEnvironment();
    void PrintMix();

    float Red(int Index);
    float Green(int Index);
    float Blue(int Index);

    void FinaliseLinks();
    void FinaliseModels();
 
    static void Init();
   
    void parse(char *infile); 

    void RegisterNewColour(int NewColour);
    int  GetColourIndex(int Value);  // converts 0xFF7730 to, say, '7'
    int  GetColourValue(int Index);  // converts 7 to, say, 0xFF7730

    float CollProb(float Dist, float Rad);
    void  ParseCollisionDataFile();

    int   ParseStructureFile(char *FileName);

};

/*
 *   A little utility ftn - ParseWords provide a map between a text
 *   name and a "ParseType" -> an enumeration.
 */

class ParseWord
{
public:
 
    friend TubParser;

    static char *Model,*State,*Site,*Sphere,*Event,*Binding,*Average;
    static char *Mix,*Start,*End,*Diffusion,*TimeScale,*Temperature,*Viscosity;

    enum ParseType { ModelKeyWord, ModelName, StateKeyWord, StateName, 
                     SiteKeyWord,  SiteName,  SphereKeyWord, EventKeyWord, 
                     BindingKeyWord, EndBlockKeyWord, StartBlockKeyWord,
                     AverageKeyWord, MixKeyWord, DiffusionKeyWord,
                     TimeScaleKeyWord, TemperatureKeyWord, ViscosityKeyWord,
                     Comment, Point, Number, Unknown };

    static ParseWord *Last, *First; 

    char *text;
    ParseType IsA;       
    ParseWord* Next;
    ParseWord* Back;

    ParseWord(char *S) 
    {text = strdup(S); IsA = FindType(text); Next=NULL; AddToList(this);}

    ParseWord(ShortString S) 
    {text = strdup(S.text); IsA = FindType(text); Next=NULL; AddToList(this);}

    ParseWord(char *S, ParseType P) 
    {text = strdup(S); IsA = P; Next=NULL; AddToList(this);}

    ~ParseWord() 
    {
        free(text); 
        if (Back!=NULL) Back->Next=Next; if (Next!=NULL) Next->Back=Back; 
        if (this==Last) Last = Back; if (this==First) First = Next;
    } 

    
    static void AddToList(ParseWord *P)
    {
        if (First==NULL) First = P;
        if (Last!=NULL)  Last->Next = P;
        P->Back = Last;
        Last = P;
    }

    int operator== (const ParseWord *P) {return !(strcasecmp(text, P->text));} 
    int operator== (const ParseType P)  {return (IsA==P);} 
    int operator== (const char *S)      {return !(strcasecmp(text, S));} 

    float asFloat()  { return (strchr(text,'%'))?(atof(text)/100):atof(text); }
    int   asInt()    { return atoi(text); }
    char* asString() { return (text); }

    static ParseType FindType(char *S);

    static int IsKeyWord(ShortString S);

    static void PrintList();

    static void Init();
};

#endif
